#include "addtrackdialog.h"
#include "ui_addtrackdialog.h"

#include <algorithm>
#include <QFileDialog>
#include "config.h"

extern const std::vector<std::string> supported_ffmpeg_extensions;

AddTrackDialog::AddTrackDialog(AcbFile *acb, AwbFile *external_awb, AwbFile *internal_awb, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::AddTrackDialog)
{
    std::vector<uint16_t> commands;
    std::vector<uint32_t> count;
    uint32_t max_value, max_idx;

    ui->setupUi(this);
    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);

    if (external_awb && internal_awb)
    {
        // No op
    }
    else
    {
        ui->awbComboBox->setCurrentIndex((external_awb) ? 0 : 1);
        ui->awbComboBox->setDisabled(true);
    }

    ui->cmd1Edit->setValidator(new QIntValidator(0, 65535, this));

    max_value = 0;
    max_idx = 0;

    for (uint32_t i = 0; i < acb->GetNumTracks(); i++)
    {
        uint16_t cmd_idx;

        if (acb->GetTrackCommandIndex(acb->TrackIndexToCueIndex(i), &cmd_idx))
        {
            size_t index;

            auto it = std::find(commands.begin(), commands.end(), cmd_idx);
            if (it == commands.end())
            {
                index = commands.size();
                commands.push_back(cmd_idx);
                count.push_back(1);
                ui->cmd1ComboBox->addItem(QString("%1 (1)").arg(cmd_idx));
            }
            else
            {
                index = it - commands.begin();
                count[index]++;
                ui->cmd1ComboBox->setItemText((int)index, QString("%1 (%2)").arg(cmd_idx).arg(count[index]));
            }

            if (count[index] > max_value)
            {
                max_value = count[index];
                max_idx = (uint32_t)index;
            }
        }
    }

    ui->cmd1ComboBox->setCurrentIndex(max_idx);

    //
    ui->cmd2Edit->setValidator(new QIntValidator(0, 65535, this));

    max_value = 0;
    max_idx = 0;
    commands.clear();
    count.clear();

    for (uint32_t i = 0; i < acb->GetNumTracks(); i++)
    {
        uint16_t cmd_idx;

        if (acb->GetSequenceCommandIndex(acb->TrackIndexToCueIndex(i), &cmd_idx))
        {
            size_t index;

            auto it = std::find(commands.begin(), commands.end(), cmd_idx);
            if (it == commands.end())
            {
                index = commands.size();
                commands.push_back(cmd_idx);
                count.push_back(1);
                ui->cmd2ComboBox->addItem(QString("%1 (1)").arg(cmd_idx));
            }
            else
            {
                index = it - commands.begin();
                count[index]++;
                ui->cmd2ComboBox->setItemText((int)index, QString("%1 (%2)").arg(cmd_idx).arg(count[index]));
            }

            if (count[index] > max_value)
            {
                max_value = count[index];
                max_idx = (uint32_t)index;
            }
        }
    }

    ui->cmd2ComboBox->setCurrentIndex(max_idx);
}

AddTrackDialog::~AddTrackDialog()
{
    delete ui;
}

QString AddTrackDialog::GetFile()
{
    return ui->fileEdit->text();
}

QString AddTrackDialog::GetName()
{
    return ui->nameEdit->text();
}

bool AddTrackDialog::IsExternalAwb()
{
    return (ui->awbComboBox->currentIndex() == 0);
}

uint16_t AddTrackDialog::GetCmd1()
{
    return ui->cmd1Edit->text().toUInt();
}

uint16_t AddTrackDialog::GetCmd2()
{
    return ui->cmd2Edit->text().toUInt();
}

int AddTrackDialog::GetLoop()
{
    return ui->loopComboBox->currentIndex();
}

void AddTrackDialog::on_cmd1ComboBox_currentIndexChanged(int index)
{
    QString text = ui->cmd1ComboBox->itemText(index);
    int par = text.lastIndexOf('(');
    if (par != -1)
        text = text.mid(0, par-1);

    uint16_t val = text.toUInt();
    ui->cmd1Edit->setText(QString("%1").arg(val));
}

void AddTrackDialog::on_cmd2ComboBox_currentIndexChanged(int index)
{
    QString text = ui->cmd2ComboBox->itemText(index);
    int par = text.lastIndexOf('(');
    if (par != -1)
        text = text.mid(0, par-1);

    uint16_t val = text.toUInt();
    ui->cmd2Edit->setText(QString("%1").arg(val));
}

void AddTrackDialog::on_fileButton_clicked()
{
    QString filter = "Standard audio files (*.wav";

    for (const std::string &ext : supported_ffmpeg_extensions)
    {
        filter += " *" + Utils::StdStringToQString(ext);
    }

    filter += ')';
    filter += ";;HCA files (*.hca)";

    QString file = QFileDialog::getOpenFileName(this, "Open", config.last_directory_import, filter);

    if (file.isNull())
        return;

    ui->fileEdit->setText(file);
}

void AddTrackDialog::on_buttonBox_accepted()
{
    if (ui->nameEdit->text().isEmpty())
    {
        DPRINTF("Name field is empty!\n");
        return;
    }

    if (ui->cmd1Edit->text().isEmpty())
    {
        DPRINTF("Cmd 1 field is empty!\n");
        return;
    }

    if (ui->cmd2Edit->text().isEmpty())
    {
        DPRINTF("Cmd 2 field is empty!\n");
        return;
    }

    if (ui->fileEdit->text().isEmpty())
    {
        DPRINTF("File field is empty!\n");
        return;
    }

    if (!Utils::FileExists(Utils::QStringToStdString(ui->fileEdit->text())))
    {
        DPRINTF("File \"%s\" doesn't exist.\n", Utils::QStringToStdString(ui->fileEdit->text()).c_str());
        return;
    }

    accept();
}
