#include "loopdialog.h"
#include "ui_loopdialog.h"

#include "debug.h"

LoopDialog::LoopDialog(QWidget *parent, QVector<AudioFile *> files):
    QDialog(parent), files(files), ui(new Ui::LoopDialog)
{
    ui->setupUi(this);
    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);

    AudioFile *file = files[0];

    float start, end;
    int count;

    this->max = SecondsToQTime(file->GetDuration());
    this->zero.setHMS(0, 0, 0);

    ui->loopStart->setMaximumTime(this->max);
    ui->loopEnd->setMaximumTime(this->max);

    if (file->GetLoop(&start, &end, &count))
    {
        this->start = SecondsToQTime(start);
        this->end = SecondsToQTime(end);
        this->loop_count = count;

        ui->hasLoop->setChecked(true);

    }
    else
    {
        this->start = zero;
        this->end = max;
        this->loop_count = 0;

        ui->hasLoop->setChecked(false);
        ui->loopStart->setDisabled(true);
        ui->loopEnd->setDisabled(true);
        ui->loopCount->setDisabled(true);
    }

    ui->loopStart->setTime(this->start);
    ui->loopEnd->setTime(this->end);
    ui->loopCount->setValue(this->loop_count);
}

LoopDialog::~LoopDialog()
{
    delete ui;
}

QTime LoopDialog::SecondsToQTime(float time)
{
    int total_seconds = time;
    int milisecs = (int)(time * 1000.0) % 1000;
    int hours = total_seconds / 3600;
    int rem_minutes = (total_seconds % 3600);
    int minutes = rem_minutes / 60;
    int seconds = rem_minutes % 60;

    QTime result;
    result.setHMS(hours, minutes, seconds, milisecs);

    return result;
}

float LoopDialog::QTimeToSeconds(const QTime &time)
{
    float ret = time.hour()*3600.0;
    ret += time.minute()*60.0;
    ret += time.second();
    ret += (float)time.msec()/1000.0;

    return ret;
}

bool LoopDialog::IsModified() const
{
    AudioFile *file = files[0];
    bool has_loop = file->HasLoop();

    if (has_loop != ui->hasLoop->isChecked())
        return true;

    if (!has_loop)
        return false;

    if (ui->loopStart->time() != start)
        return true;

    if (ui->loopEnd->time() != end)
        return true;

    if (ui->loopCount->value() != loop_count)
        return true;

    return false;
}

void LoopDialog::on_hasLoop_toggled(bool checked)
{
    ui->loopStart->setEnabled(checked);
    ui->loopEnd->setEnabled(checked);
    ui->loopCount->setEnabled(checked);
}

void LoopDialog::on_buttonBox_rejected()
{
    done(0);
}

void LoopDialog::on_buttonBox_accepted()
{
    if (!IsModified())
    {
        done(0);
        return;
    }

    for (int i = 0; i < files.size(); i++)
    {
        AudioFile *file = files[i];

        if (!ui->hasLoop->isChecked())
        {
            file->RemoveLoop();
        }
        else
        {
            QTime new_start = ui->loopStart->time();
            QTime new_end = ui->loopEnd->time();
            int new_count = ui->loopCount->value();

            if (new_start >= new_end)
            {
                DPRINTF("Loop start must be smaller than loop end.\n");
                return;
            }

            if (new_start == zero && new_end == max && new_count == 0)
            {
                file->SetMaxLoop();
            }

            else if (!file->SetLoop(QTimeToSeconds(new_start), QTimeToSeconds(new_end), new_count))
            {
                DPRINTF("Cannot set loop, either the gap between loop start and loop end is smaller than internal time resolution, or loop is not implemented for this format.\n");
                return;
            }
        }
    }

    done(1);
}


