#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QFileDialog>
#include <QMessageBox>
#include <QCloseEvent>
#include <QToolButton>
#include <QAudioOutput>
#include "workerdialog.h"
#include "loopquestiondialog.h"
#include "loopdialog.h"
#include "optionsdialog.h"
#include "mixedmodedialog.h"
#include "selectsilencedialog.h"
#include "addtrackdialog.h"
#include "debug.h"

#include "IniFile.h"
#include "Thread.h"
#include "config.h"
#define PSEUDO_FILE_PREFIX  "***\r\nPSEUDOFILE\r\n***"

enum
{
    COLUMN_PLAYING,
    COLUMN_TRACK,
    COLUMN_ID,
    COLUMN_NAME,
    COLUMN_TYPE,
    COLUMN_CHANNELS,
    COLUMN_FREQUENCY,
    COLUMN_DURATION
};

const std::vector<std::string> supported_ffmpeg_extensions =
{
    ".mp3",
    ".ogg",
    ".wma",
    ".aac",
    ".m4a",
    ".mp4",
    ".flac",
    ".fla",
    ".mka",
    ".m4b",
    ".m4r",
    ".oga",
    ".ogx",
    ".mp1",
    ".mp2",
    ".mpc",
    ".mpc8",
    ".mpp",
    ".mp+",
    ".ac3",
    ".wv",
    ".eac3",
    ".dts",
    ".adts",
    ".aif",
    ".aiff",
    ".aifc",
    ".afc",
    ".al",
    ".amr",
    ".apc",
    ".ape",
    ".apl",
    ".au",
    ".daud",
    ".caf",
    ".siff",
    ".snd",
    ".spx",
    ".svx",
    ".8svx",
    ".tta",
    ".w64"
};

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    set_debug_level(2);

    statusLabel = new QLabel(this);
    ui->statusBar->addPermanentWidget(statusLabel);
    statusLabel->setText("No file loaded.");
    //ui->statusBar->showMessage("No file loaded.");

    ui->tracksList->sortByColumn(COLUMN_TRACK, Qt::AscendingOrder);
    ui->tracksList->setSortingEnabled(true);
    ui->tracksList->setCurrentItem(nullptr);

    ui->tracksList->header()->resizeSection(COLUMN_PLAYING, 60);
    ui->tracksList->header()->resizeSection(COLUMN_NAME, 240);

    // Add disabled context menu
    QAction *separator1 = new QAction();
    QAction *separator2 = new QAction();
    QAction *separator3 = new QAction();

    separator1->setSeparator(true);
    separator2->setSeparator(true);
    separator3->setSeparator(true);

    ui->tracksList->addAction(ui->actionPlay);
    ui->tracksList->addAction(ui->actionPause);
    ui->tracksList->addAction(ui->actionStop);
    ui->tracksList->addAction(separator1);

    ui->tracksList->addAction(ui->actionExtract);
    ui->tracksList->addAction(ui->actionExtractJoin);
    ui->tracksList->addAction(ui->actionExtract_all);
    ui->tracksList->addAction(separator2);

    ui->tracksList->addAction(ui->actionReplace);
    ui->tracksList->addAction(ui->actionReplaceSplit);
    ui->tracksList->addAction(ui->actionReplace_all);
    ui->tracksList->addAction(separator3);

    ui->tracksList->addAction(ui->actionLoop);

    ui->actionPlay->setDisabled(true);
    ui->actionPause->setDisabled(true);
    ui->actionStop->setDisabled(true);
    ui->actionExtract->setDisabled(true);
    ui->actionExtractJoin->setDisabled(true);
    ui->actionExtract_all->setDisabled(true);
    ui->actionReplace->setDisabled(true);
    ui->actionReplace_all->setDisabled(true);
    ui->actionLoop->setDisabled(true);
    ui->actionAddTrack->setDisabled(true);

    ui->actionPlay->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_MediaPlay));
    ui->actionPause->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_MediaPause));
    ui->actionStop->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_MediaStop));
    ui->actionExtract->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowDown));
    ui->actionExtractJoin->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowDown));
    ui->actionExtract_all->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowDown));
    ui->actionReplace->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowUp));
    ui->actionReplaceSplit->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowUp));
    ui->actionReplace_all->setIcon(ui->tracksList->style()->standardIcon(QStyle::SP_ArrowUp));

    // Disable save with no file loaded
    ui->actionSave->setDisabled(true);
    ui->actionSave_as->setDisabled(true);

    acb = nullptr;
    external_awb = nullptr;
    internal_awb = nullptr;

    audio_output = nullptr;
    audio_buffer = nullptr;

    track_playing = -1;

    QDir::setCurrent(qApp->applicationDirPath());
    config.Load();

    av_register_all();

    // Handle arguments
    if (qApp->arguments().size() == 2)
    {
        QString file = qApp->arguments()[1];

        SavePath(file, config.last_directory_open, false);
        DoLoadFile(file, false);
    }
}

MainWindow::~MainWindow()
{
    if (acb)
        delete acb;

    if (external_awb)
        delete external_awb;

    if (internal_awb)
        delete internal_awb;

    for (const AudioTrack &track : tracks)
    {
        if (track.file)
            delete track.file;
    }

    delete ui;
}

bool MainWindow::ProcessShutdown()
{
    if (IsFileModified())
    {
        if (QMessageBox::question(this, "Eternity Audio Tool", "Changes on current file haven't been saved. Do you want to exit?",
                                  QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No), QMessageBox::No) == QMessageBox::No)
        {
            return false;
        }
    }

    config.Save();
    return true;
}

void MainWindow::SavePath(const QString &file_path, QString &directory, bool is_directory)
{
    if (is_directory)
    {
        directory = file_path;
        return;
    }

    int last_pos = file_path.lastIndexOf('/');
    int last_pos2 = file_path.lastIndexOf('\\');

    if (last_pos2 > last_pos)
        last_pos = last_pos2;

    if (last_pos > 0)
    {
       directory = file_path.left(last_pos);
    }
}

QString MainWindow::GetAwbPath(const QString &file, bool cpk)
{
    QString awb_path;
    QString ext = (cpk) ? ".cpk" : ".awb";
    int ext_idx = file.lastIndexOf('.');

    if (ext_idx != -1)
    {
        awb_path = file.left(ext_idx) + ext;
    }
    else
    {
        awb_path = file + ext;
    }

    return awb_path;
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    if (ProcessShutdown())
        event->accept();
    else
        event->ignore();
}

void MainWindow::on_actionOpen_triggered()
{
    if (IsFileModified())
    {
        if (QMessageBox::question(this, "Eternity Audio Tool", "Changes on current file haven't been saved. Do you want to open another file anyway?",
                                  QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No), QMessageBox::No) == QMessageBox::No)
        {
            return;
        }
    }

    QString file = QFileDialog::getOpenFileName(this, "Open", config.last_directory_open, "CRIWARE audio containers (*.acb *.csb *.uasset)");

    if (!file.isNull())
    {
        SavePath(file, config.last_directory_open, false);        
        DoLoadFile(file, false);
    }
}

void MainWindow::on_actionOpenAudio_triggered()
{
    if (IsFileModified())
    {
        if (QMessageBox::question(this, "Eternity Audio Tool", "Changes on current file haven't been saved. Do you want to open another file anyway?",
                                  QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No), QMessageBox::No) == QMessageBox::No)
        {
            return;
        }
    }

    QStringList files = QFileDialog::getOpenFileNames(this, "Open", config.last_directory_open, "CRIWARE audio (*.hca *.adx *.aax)");

    if (files.isEmpty())
        return;

    qSort(files);

    QString pseudo_file = PSEUDO_FILE_PREFIX;
    for (int i = 0; i < files.size(); i++)
    {
        if (i != 0)
            pseudo_file += '\r';

        pseudo_file += files[i];
    }

    SavePath(files[0], config.last_directory_open, false);
    DoLoadFile(pseudo_file, false, files.size() > 1);
}

void MainWindow::on_actionSave_triggered()
{
    if (!acb)
        return;

    PseudoCriwareAudioContainer *is_pseudo = dynamic_cast<PseudoCriwareAudioContainer *>(acb);
    bool multiple = is_pseudo && external_awb->GetNumFiles() > 1;

    if (DoSaveFile(acb_path, multiple))
    {        
        DoLoadFile(acb_path, true, multiple);
    }
    else
    {
        exit(-1);
    }
}

void MainWindow::on_actionSave_as_triggered()
{
    assert(acb);

    UassetAcbFile *is_uasset = dynamic_cast<UassetAcbFile *>(acb);
    CsbFile *is_csb = dynamic_cast<CsbFile *>(acb);
    QString filter;

    if (is_uasset)
    {
        filter = "Uasset files (*.uasset)";
    }
    else if (is_csb)
    {
        filter = "CSB files (*.csb)";
    }
    else
    {
        filter = "ACB files (*.acb)";
    }

    QString select = Utils::StdStringToQString(Utils::MakePathString(Utils::QStringToStdString(config.last_directory_save), Utils::GetFileNameString(Utils::QStringToStdString(acb_path))));

    QString file = QFileDialog::getSaveFileName(this, "Save", select, filter);

    if (!file.isNull())
    {
        SavePath(file, config.last_directory_save, false);

        if (DoSaveFile(file))
        {
            DoLoadFile(file, true);
        }
        else
        {
            exit(-1);
        }
    }
}

bool MainWindow::DoLoadFile(const QString &file, bool reload, bool multiple)
{
    if (file.isNull())
        return false;

    if (!reload)
    {
        statusLabel->setText("Loading...");
    }

    StopCurrentTrack();
    job_cancel = false;

    bool error = false;

    ClearTracksListUI();

    WorkerDialog dialog(this);
    dialog.SetLoadMode(file);

    int ret = dialog.exec();

    if (ret < 0)
    {
        statusLabel->setText((multiple) ? "Failed to load files." : "Failed to load file.");
        error = true;
    }
    else if (ret == 0)
    {
        error = true;
    }

    for (int i = 0; i < tracks.size(); i++)
    {
        AddTrackToListUI(i, false);
    }

    if (error)  // If error or cancel, tracks are reloaded from previous file and we leave here
        return false;

    CsbFile *csb = dynamic_cast<CsbFile *>(acb);
    AcbFile *_acb = dynamic_cast<AcbFile *>(acb);

    if (_acb && !csb)
    {
       bool enable_add = (_acb->CanAddTrack());
       Afs2File *external_afs2 = dynamic_cast<Afs2File *>(external_awb);
       Afs2File *internal_afs2 = dynamic_cast<Afs2File *>(internal_awb);

       enable_add = enable_add && (external_afs2 || internal_afs2);
       ui->actionAddTrack->setEnabled(enable_add);
    }

    if (!reload)
    {
        if (multiple)
        {
            statusLabel->setText("Files loaded.");
        }
        else
        {
            if (file.startsWith(PSEUDO_FILE_PREFIX))
                statusLabel->setText("File loaded: " + file.mid((int)strlen(PSEUDO_FILE_PREFIX)));
            else
                statusLabel->setText("File loaded: " + file);
        }
    }

    ui->actionSave->setDisabled(true);
    ui->actionSave_as->setDisabled(true);

    on_tracksList_itemSelectionChanged(); // Added in 0.9
    return true;
}

bool MainWindow::DoSaveFile(const QString &file, bool multiple)
{
    if (file.isNull())
        return false;

    statusLabel->setText("Saving...");

    StopCurrentTrack();
    job_cancel = false;

    WorkerDialog dialog(this);

    dialog.SetSaveMode(file);

    int ret = dialog.exec();

    if (ret < 0)
    {
        if (multiple)
            DPRINTF("Failed to save files. Program will exit.\n");
        else
            DPRINTF("Failed to save file. Program will exit.\n");

        exit(-1);
        return false;
    }
    else if (ret == 0)
    {
        DPRINTF("Cancelled... but we shouldn't be here.\n");
        exit(-1);
        return false;
    }

    if (multiple)
    {
        UPRINTF("Files saved succesfully!\n");
        statusLabel->setText("Files saved.");
    }
    else
    {
        UPRINTF("File saved succesfully!\n");

        if (file.startsWith(PSEUDO_FILE_PREFIX))
            statusLabel->setText("File saved: " + file.mid((int)strlen(PSEUDO_FILE_PREFIX)));
        else
            statusLabel->setText("File saved: " + file);
    }

    return true;
}

bool MainWindow::LoadFile(const QString &file)
{
    CriwareAudioContainer *new_acb;
    AwbFile *new_external_awb = nullptr, *new_internal_awb = nullptr;
    QVector<AudioTrack> new_tracks;
    std::vector<std::string> files_list; // For pseudo

    if (file.isNull())
        return false;

    bool is_uasset = false;
    bool is_csb = false;
    bool is_pseudo = false;
    guessed_ciph_type = -1;    

    if (file.startsWith(PSEUDO_FILE_PREFIX))
    {
        QString list_str = file.mid((int)strlen(PSEUDO_FILE_PREFIX));
        Utils::GetMultipleStrings(Utils::QStringToStdString(list_str), files_list, '\r');

        is_pseudo = true;
        new_acb = new PseudoCriwareAudioContainer(files_list);
    }
    else if (file.endsWith(".uasset", Qt::CaseInsensitive))
    {
        is_uasset = true;

        if (!UassetAcbFile::IsUassetAcb(Utils::QStringToStdString(file)))
        {
            DPRINTF("That .uasset file doesn't contain .acb data.\n");
            return false;
        }

        new_acb = new UassetAcbFile();
    }
    else if (file.endsWith(".csb", Qt::CaseInsensitive))
    {
        is_csb = true;
        new_acb = new CsbFile();
    }
    else
    {
        new_acb = new AcbFile();
    }

    if (!is_pseudo && !new_acb->LoadFromFile(Utils::QStringToStdString(file)))
    {
        DPRINTF("Failed to load file.\n");

        delete new_acb;
        return false;
    }

    if (new_acb->HasAwb())
    {
        uint8_t *awb_buf;
        uint32_t awb_size;

        awb_buf = new_acb->GetAwb(&awb_size); // pointer, not copy
        if (!awb_buf)
        {
            DPRINTF("Failed to get .awb data from .acb\n");

            delete new_acb;
            return false;
        }

        if (is_csb)
        {
            new_internal_awb = new CsbInternalAwbFile();
        }
        else if (awb_size >= 4 && *(uint32_t *)awb_buf == CPK_SIGNATURE)
        {
            new_internal_awb = new CpkFile();
        }
        else
        {
            new_internal_awb = new Afs2File();
        }

        if (!new_internal_awb->Load(awb_buf, awb_size))
        {
            DPRINTF("Failed to load internal .awb data\n");

            delete new_acb;
            delete new_internal_awb;

            return false;
        }
    }
    else
    {
        if (is_uasset)
        {
            DPRINTF("That file contains .acb data but it doesn't contain .awb data inside.\n");

            delete new_acb;
            return false;
        }
    }

    // External awb
    if (!is_uasset)
    {
        if (is_pseudo)
        {
            new_external_awb = new PseudoAwbFile(files_list);
        }
        else
        {
            QString awb_path = GetAwbPath(file, is_csb);

            uint8_t *temp = Utils::ReadFileFrom(Utils::QStringToStdString(awb_path), 0, sizeof(uint32_t), false);
            if (temp)
            {
                if (*(uint32_t *)temp == CPK_SIGNATURE)
                {
                    new_external_awb = new CpkFile();
                }
                else
                {
                    new_external_awb = new Afs2File();
                }

                delete[] temp;

                if (!new_external_awb->LoadFromFile(Utils::QStringToStdString(awb_path)))
                {
                    DPRINTF("Failed to load external .awb.\n");

                    delete new_acb;
                    delete new_external_awb;

                    if (new_internal_awb)
                        delete new_internal_awb;

                    return false;
                }
            }


        }
    }

    if (!new_internal_awb && !new_external_awb)
    {
        DPRINTF("Neither internal nor external awb found for this acb.\n");
        delete new_acb;
        return false;
    }

    new_tracks.resize(new_acb->GetNumTracks());
    emit progressCalculated(new_tracks.size()+1);

    int ciph0_count = 0, ciph1_count = 0, ciph56_count = 0;

    for (int i = 0; i < new_tracks.size(); i++)
    {
        AwbFile *awb;
        uint32_t awb_index;
        bool external;

        if (job_cancel)
        {
            delete new_acb;

            if (new_external_awb)
                delete new_external_awb;

            if (new_internal_awb)
                delete new_internal_awb;

            return false;
        }

        awb_index = new_acb->TrackIndexToAwbIndex(i, &external);

        if (awb_index == (uint32_t)-1)
        {
            DPRINTF("Cannot map track index %d to an awb index.\n", i);
            delete new_acb;

            if (new_external_awb)
                delete new_external_awb;

            if (new_internal_awb)
                delete new_internal_awb;

            return false;
        }

        awb = (external) ? new_external_awb : new_internal_awb;

        if (!awb)
        {
            if (external)
            {
                DPRINTF("The track %d references external awb file %d but no external awb was found.\n", i, awb_index);
            }
            else
            {
                DPRINTF("The track %d references internal awb file %d but acb has not internal awb.\n", i, awb_index);
            }

            delete new_acb;

            if (new_external_awb)
                delete new_external_awb;

            if (new_internal_awb)
                delete new_internal_awb;

            return false;
        }

        uint64_t track_size;
        uint8_t *track_data = awb->ExtractFile(awb_index, &track_size);

        if (!track_data)
        {
            DPRINTF("Failed to get file 0x%x from awb.\n", awb_index);

            delete new_acb;            

            if (new_external_awb)
                delete new_external_awb;

            if (new_internal_awb)
                delete new_internal_awb;

            return false;
        }

        if (HcaFile::CheckHeader(track_data, track_size))
        {
            HcaFile *hca = new HcaFile();

            if (!hca->Load(track_data, track_size))
            {
                DPRINTF("HCA Load failed, on track %d (awb track %d)\n", i, awb_index);

                delete hca;
                delete[] track_data;
                delete new_acb;

                if (new_external_awb)
                    delete new_external_awb;

                if (new_internal_awb)
                    delete new_internal_awb;

                return false;
            }

            uint16_t ciph_type = hca->GetCiphType();
            if (ciph_type == 0)
                ciph0_count++;
            else if (ciph_type == 1)
                ciph1_count++;
            else
                ciph56_count++;

            new_tracks[i].type = AudioTrack::Type::HCA;
            new_tracks[i].file = hca;
        }
        else if (AdxFile::CheckHeader(track_data, track_size))
        {
            AdxFile *adx = new AdxFile();

            if (!adx->Load(track_data, track_size))
            {
                DPRINTF("ADX Load failed.\n");

                delete adx;
                delete[] track_data;
                delete new_acb;

                if (new_external_awb)
                    delete new_external_awb;

                if (new_internal_awb)
                    delete new_internal_awb;

                return false;
            }

            new_tracks[i].type = AudioTrack::Type::ADX;
            new_tracks[i].file = adx;
        }
        else if (AaxFile::CheckHeader(track_data, track_size))
        {
            AaxFile *aax = new AaxFile();

            if (!aax->Load(track_data, track_size))
            {
                DPRINTF("AAX Load failed: %d\n", i);

                delete aax;
                delete[] track_data;
                delete new_acb;

                if (new_external_awb)
                    delete new_external_awb;

                if (new_internal_awb)
                    delete new_internal_awb;

                return false;
            }

            new_tracks[i].type = AudioTrack::Type::AAX;
            new_tracks[i].file = aax;
        }
        else
        {
            new_tracks[i].type = AudioTrack::Type::BIN;
            new_tracks[i].file = nullptr;
        }        

        new_tracks[i].modified = false;
        new_tracks[i].external = false;
        new_tracks[i].index = i;
        new_tracks[i].state = AudioTrack::State::STOPPED;
        delete[] track_data;

        emit progressChanged();
    }

    if (job_cancel)
    {
        delete new_acb;
        delete new_external_awb;

        return false;
    }

    // Point of no return starts here

    if (ciph0_count != 0 && ciph1_count == 0 && ciph56_count == 0)
    {
        guessed_ciph_type = 0;
    }
    else if (ciph1_count != 0 && ciph0_count == 0 && ciph56_count == 0)
    {
        guessed_ciph_type = 1;
    }

    //DPRINTF("Guessed: %d\n", guessed_ciph_type);

    if (acb)
    {
        delete acb;

        if (external_awb)
            delete external_awb;

        if (internal_awb)
            delete internal_awb;
    }

    acb = new_acb;
    external_awb = new_external_awb;
    internal_awb = new_internal_awb;

    for (const AudioTrack &track : tracks)
    {
        if (track.file)
            delete track.file;
    }

    tracks = new_tracks;   
    acb_path = file;

    emit progressChanged();
    return true;
}

bool MainWindow::SaveFile(const QString &file)
{
    if (file.isNull())
        return false;

    QString temp_acb = file + ".___temp_";
    QString awb_path, temp_awb;

    CsbFile *csb = dynamic_cast<CsbFile *>(acb);
    PseudoCriwareAudioContainer *pseudo = dynamic_cast<PseudoCriwareAudioContainer *>(acb);
    AcbFile *acb_acb = dynamic_cast<AcbFile *>(acb);

    if (csb)
        acb_acb = nullptr;

    emit progressCalculated(tracks.size() + 3);

    // Part 1, save tracks back into awb
    for (int i = 0; i < tracks.size(); i++)
    {
        uint32_t awb_index;
        bool external;

        if (job_cancel)
            return false;

        awb_index = acb->TrackIndexToAwbIndex(i, &external);

        if (tracks[i].modified)
        {
            size_t size;
            uint8_t *data;
            AudioFile *audio = tracks[i].file;

            assert(audio);
            assert(tracks[i].type == AudioTrack::Type::HCA || tracks[i].type == AudioTrack::Type::ADX || tracks[i].type == AudioTrack::Type::AAX);

            if (csb)
            {
                if (!csb->SetTrackData(i, (uint8_t)audio->GetNumChannels(), audio->GetSampleRate(), audio->GetNumSamples()))
                {
                    DPRINTF("CsbFile::SetTrackData failed.\n");
                    return false;
                }
            }

            if (acb_acb)
            {
                std::vector<uint32_t> cue_idxs;
                acb_acb->TrackIndexToCueIndexes(i, cue_idxs);

                for (size_t n = 0; n < cue_idxs.size(); n++)
                {
                    acb_acb->SetCueLength(acb_acb->CueIndexToCueId(cue_idxs[n]), (uint32_t)(audio->GetDuration()*1000.0f));
                }

                acb_acb->SetWaveformNumSamples(i, audio->GetNumSamples());
                acb_acb->SetWaveformSamplingRate(i, audio->GetSampleRate());
                acb_acb->SetWaveformNumChannels(i, audio->GetNumChannels());
            }

            data = tracks[i].file->Save(&size);
            if (!data)
            {
                DPRINTF("Failed to save track %d in RAM.\n", i);
                return false;
            }

            AwbFile *awb = (external) ? external_awb : internal_awb;

            {
                if (!awb)
                {
                    DPRINTF("%s: Internal error, awb null.\n", FUNCNAME);
                    return false;
                }

                if (!awb->SetFile(awb_index, data, (uint32_t)size, true))
                {
                    delete[] data;
                    DPRINTF("Failed to set file %d in awb object.\n", awb_index);
                    return false;
                }

                // data now belongs to awb, don't delete!!!!
            }
        }

        // Lets free some RAM
        if (tracks[i].file)
        {
            delete tracks[i].file;
            tracks[i].file = nullptr;
        }

        emit progressChanged();
    }

    if (job_cancel)
        return false;

    tracks.clear();

    // Part 2, save awb (in ram or file)
    CpkFile *is_cpk = dynamic_cast<CpkFile *>(external_awb);

    if (internal_awb)
    {
        // Standalone acb case
        size_t size;
        uint8_t *data;        

        data = internal_awb->Save(&size);
        if (!data)
        {
            DPRINTF("Failed to save awb data (internal)\n");
            return false;
        }

        if (!acb->SetAwb(data, (uint32_t)size, true))
        {
            delete[] data;
            DPRINTF("SetAwb failed (internal).\n");
            return false;
        }

        // data now belongs to acb, don't delete!!!!

        // Time to shut this one down
        delete internal_awb;
        internal_awb = nullptr;
    }

    if (external_awb)
    {
        // External awb case

        if (acb->HasAwbHeader())
        {
            if (is_cpk)
            {
                DPRINTF("CPK awb with header, we haven't seen this format. Header copy won't be saved in .acb.\n\nPlease, send original file to eternity.\n");
            }
            else
            {
                unsigned int header_size;
                uint8_t *awb_header = external_awb->CreateHeader(&header_size);
                if (!awb_header)
                {
                    DPRINTF("Failed to create awb header.\n");
                    return false;
                }

                if (!acb->SetAwbHeader(awb_header, header_size, true))
                {
                    delete[] awb_header;
                    DPRINTF("Failed to set awb header in .acb.\n");
                    return false;
                }

                // awb_header buf now belongs to acb
            }
        }

        awb_path = GetAwbPath(file, csb != nullptr);
        temp_awb = awb_path + ".___temp_";

        if (!pseudo && !external_awb->SaveToFile(Utils::QStringToStdString(temp_awb)))
            return false;

        // Delete this now, which should free the handle of the opened file (this is needed when saving file in same path!)
        delete external_awb;
        external_awb = nullptr;

        uint8_t awb_md5[0x10];

        if (!csb && !pseudo && Utils::FileMd5(Utils::QStringToStdString(temp_awb), awb_md5))
        {
            acb->SetExternalAwbHash(awb_md5);
        }
    }

    emit progressChanged();

    if (job_cancel)
        return false;

    // Part 3 save acb file
    if (!pseudo && !acb->SaveToFile(Utils::QStringToStdString(temp_acb)))
    {
        if (!temp_awb.isNull())
        {
            QFile::remove(temp_awb);
        }

        return false;
    }

    // Delete and free file handle (if any)
    delete acb;
    acb = nullptr;

    emit progressChanged();

    if (job_cancel)
        return false;

    // Point of no return

    if (pseudo)
    {
        emit progressChanged();
        return true;
    }

    // Extra part, delete original files (if any) and rename new ones
    QFile::remove(file);

    if (!QFile::rename(temp_acb, file))
    {
        DPRINTF("Rename of acb failed!!\n");
        return false;
    }

    if (!awb_path.isNull())
    {
        QFile::remove(awb_path);

        if (!QFile::rename(temp_awb, awb_path))
        {
            DPRINTF("Rename of awb failed!!\n");
            return false;
        }
    }

    emit progressChanged();
    return true;
}

bool MainWindow::IsFileModified()
{
    if (!acb)
        return false;

    for (const AudioTrack &track : tracks)
    {
        if (track.modified)
            return true;
    }

    return false;
}

QString MainWindow::DurationToQString(float duration)
{
    int total_seconds = lrint(duration);
    int hours = total_seconds / 3600;
    int rem_minutes = (total_seconds % 3600);
    int minutes = rem_minutes / 60;
    int seconds = rem_minutes % 60;

    return QString("%1:%2:%3").arg(hours, 2, 10, QChar('0')).arg(minutes, 2, 10, QChar('0')).arg(seconds, 2, 10, QChar('0'));
}

void MainWindow::ProcessItemUI(int index, QTreeWidgetItem *item)
{
    if (!item)
        item = GetItem(index);

    QVariant value(index);
    AudioTrack &track = tracks[index];
    QFont font = item->font(COLUMN_TRACK);

    font.setBold(tracks[index].modified);

    if (tracks[index].modified)
    {
        PseudoCriwareAudioContainer *is_pseudo = dynamic_cast<PseudoCriwareAudioContainer *>(acb);

        ui->actionSave->setEnabled(true);
        ui->actionSave_as->setEnabled(!is_pseudo);
    }

    item->setFont(COLUMN_TRACK, font);
    item->setFont(COLUMN_ID, font);
    item->setFont(COLUMN_NAME, font);
    item->setFont(COLUMN_TYPE, font);
    item->setFont(COLUMN_CHANNELS, font);
    item->setFont(COLUMN_FREQUENCY, font);
    item->setFont(COLUMN_DURATION, font);

    item->setText(COLUMN_PLAYING, "");
    item->setText(COLUMN_TRACK, QString("%1").arg(index, 4, 10, QChar('0')));

    if (track.file)
    {
        if (track.type == AudioTrack::Type::HCA)
            item->setText(COLUMN_TYPE, "HCA");
        else if (track.type == AudioTrack::Type::ADX)
            item->setText(COLUMN_TYPE, "ADX");
        else if (track.type == AudioTrack::Type::AAX)
            item->setText(COLUMN_TYPE, "AAX");
        else
            item->setText(COLUMN_TYPE, "???");

        item->setText(COLUMN_CHANNELS, QString("%1").arg(track.file->GetNumChannels()));
        item->setText(COLUMN_FREQUENCY, QString("%1").arg(track.file->GetSampleRate()) + " Hz");

        if (track.index == track_playing && track.state != AudioTrack::State::STOPPED)
        {
            item->setText(COLUMN_DURATION, DurationToQString((float)audio_time/1000.0)+ " / " +DurationToQString(track.file->GetDuration()));
        }
        else
        {
            item->setText(COLUMN_DURATION, DurationToQString(track.file->GetDuration()));
        }
    }
    else
    {
        item->setText(COLUMN_TYPE, "???");
        item->setText(COLUMN_CHANNELS, "???");
        item->setText(COLUMN_FREQUENCY, "???");
        item->setText(COLUMN_DURATION, "???");
    }

    item->setData(COLUMN_TRACK, Qt::UserRole, value);

    if (index == track_playing)
    {
        if (tracks[index].state == AudioTrack::State::PAUSED)
        {
            QLabel *label = new QLabel();
            label->setPixmap(ui->tracksList->style()->standardIcon(QStyle::SP_MediaPause).pixmap(16, 16));

            label->setAlignment(Qt::AlignCenter);
            ui->tracksList->setItemWidget(item, COLUMN_PLAYING, label);
        }
        else if (tracks[index].state == AudioTrack::State::PLAYING)
        {
            QLabel *label = new QLabel();
            label->setPixmap(ui->tracksList->style()->standardIcon(QStyle::SP_MediaPlay).pixmap(16, 16));

            label->setAlignment(Qt::AlignCenter);
            ui->tracksList->setItemWidget(item, COLUMN_PLAYING, label);
        }
        else
        {
            // 0.9: added this to avoid the playing icon from not being removed inmediately when opening another file
            QLabel *label = dynamic_cast<QLabel *>(ui->tracksList->itemWidget(item, COLUMN_PLAYING));
            if (label)
            {
                label->setPixmap(QPixmap());
            }
            // end 0.9

            ui->tracksList->setItemWidget(item, COLUMN_PLAYING, nullptr);
        }
    }
    else
    {
        ui->tracksList->setItemWidget(item, COLUMN_PLAYING, nullptr);
    }

    // Added in 1.0 alpha1
    std::vector<uint32_t> cue_idxs;
    std::vector<uint32_t> cue_ids;
    std::vector<std::string> names;

    acb->TrackIndexToCueIndexes(track.index, cue_idxs);
    cue_ids.resize(cue_idxs.size());
    names.resize(cue_ids.size());

    for (size_t i = 0; i < cue_ids.size(); i++)
    {
        cue_ids[i] = acb->CueIndexToCueId(cue_idxs[i]);
        acb->GetCueName(cue_ids[i], &names[i]);
    }

    //uint32_t cue_idx = acb->TrackIndexToCueIndex(track.index);
    //uint32_t cue_id = acb->CueIndexToCueId(cue_idx);

    item->setText(COLUMN_ID, Utils::StdStringToQString(Utils::ToSingleString(cue_ids)));

    if (names.size() > 0)
    {
        track.name = Utils::StdStringToQString(Utils::ToSingleString(names), false);
        item->setText(COLUMN_NAME, track.name);
    }
}

void MainWindow::AddTrackToListUI(int index, bool select)
{
    if (index < 0 || index >= tracks.size())
        return;

    QTreeWidgetItem *item = new QTreeWidgetItem();

    ui->tracksList->addTopLevelItem(item);

    if (select)
    {
        ui->tracksList->setCurrentItem(item);
    }

    ProcessItemUI(index, item);
}

void MainWindow::UpdateTrackFromListUI(int index, bool select)
{
    if (index < 0 || index >= tracks.size())
        return;

    for (int i = 0; i < ui->tracksList->topLevelItemCount(); i++)
    {
        QTreeWidgetItem *item = ui->tracksList->topLevelItem(i);

        if (item->data(COLUMN_TRACK, Qt::UserRole).toInt() == (int)index)
        {
            ProcessItemUI(index, item);

            if (select)
            {
                ui->tracksList->setCurrentItem(item);
            }

            return;
        }
    }
}

void MainWindow::ClearTracksListUI()
{
    clearing_list = true;

    for (int i = 0; i < ui->tracksList->topLevelItemCount(); )
    {
        QTreeWidgetItem *item = ui->tracksList->topLevelItem(i);
        ui->tracksList->takeTopLevelItem(i);
        delete item;
    }

    clearing_list = false;
    on_tracksList_itemSelectionChanged();
}

AudioTrack *MainWindow::GetTrack(const QTreeWidgetItem *item)
{
    int index = item->data(COLUMN_TRACK, Qt::UserRole).toInt();

    if (index < 0 || index >= tracks.size())
        return nullptr;

    return &tracks[index];
}

QTreeWidgetItem *MainWindow::GetItem(int index)
{
    if (index < 0 || index >= tracks.size())
        return nullptr;

    for (int i = 0; i < ui->tracksList->topLevelItemCount(); i++)
    {
        QTreeWidgetItem *item = ui->tracksList->topLevelItem(i);

        if (item->data(COLUMN_TRACK, Qt::UserRole).toInt() == (int)index)
            return item;
    }

    return nullptr;
}

bool MainWindow::ExtractTrack(const AudioTrack *track, const QString &file)
{
    bool pure_extract = false;

    assert(track && !track->external);

    if (track->type == AudioTrack::Type::BIN)
    {
        assert(!track->modified);
        pure_extract = true;
    }
    else
    {
        if (file.endsWith(".hca", Qt::CaseInsensitive) || file.endsWith(".adx", Qt::CaseInsensitive) || file.endsWith(".aax", Qt::CaseInsensitive))
        {
            if (track->modified)
            {
               if (!track->file->SaveToFile(Utils::QStringToStdString(file)))
               {
                   DPRINTF("Cannot extract file.\n");
                   return false;
               }
            }
            else
            {
                pure_extract = true;
            }
        }
        else
        {
            HcaFile *hca = dynamic_cast<HcaFile *>(track->file);
            AdxFile *adx = dynamic_cast<AdxFile *>(track->file);
            AaxFile *aax = dynamic_cast<AaxFile *>(track->file);
            assert(hca || adx || aax);

            if (hca && !hca->DecodeToWav(Utils::QStringToStdString(file), config.preserve_loop, config.use_float ? AUDIO_FORMAT_FLOAT : config.bit_depth))
            {
                DPRINTF("Decode/save to wav failed.\n");
                return false;
            }
            else if (adx && !adx->DecodeToWav(Utils::QStringToStdString(file), config.preserve_loop, config.use_float ? AUDIO_FORMAT_FLOAT : config.bit_depth))
            {
                DPRINTF("Decode/save to wav failed.\n");
                return false;
            }
            else if (aax && !aax->DecodeToWav(Utils::QStringToStdString(file), config.preserve_loop, config.use_float ? AUDIO_FORMAT_FLOAT : config.bit_depth))
            {
                DPRINTF("Decode/save to wav failed.\n");
                return false;
            }
        }
    }

    if (pure_extract)
    {
        QMutexLocker locker(&main_mutex);

        bool external;
        uint32_t awb_index = acb->TrackIndexToAwbIndex(track->index, &external);
        AwbFile *awb = (external) ? external_awb : internal_awb;

        if (!awb)
        {
            DPRINTF("%s: Internal error, awb null.\n", FUNCNAME);
            return false;
        }

        if (!awb->ExtractFile(awb_index, Utils::QStringToStdString(file)))
        {
            DPRINTF("Cannot extract file (2).\n");
            return false;
        }
    }

    return true;
}

bool MainWindow::ReplaceTrack(int index, const QString &file, bool loop)
{
    AudioFile *audio;
    HcaFile *hca;
    AdxFile *adx;
    AaxFile *aax;

    bool is_adx = (tracks[index].type == AudioTrack::Type::ADX);
    bool is_aax = (tracks[index].type == AudioTrack::Type::AAX);
    bool is_hca = (!is_adx && !is_aax);  // HCA is the default. Meaning, that "BIN" tracks (unknown format) will also be replaced with HCA.

    if (is_adx)
    {
        adx = new AdxFile();
        audio = adx;
    }
    else if (is_aax)
    {
        aax = new AaxFile();
        audio = aax;
    }
    else
    {
        hca = new HcaFile();
        audio = hca;
    }

    if (file.endsWith(".wav", Qt::CaseInsensitive))
    {
        if (is_adx)
        {
            if (!adx->EncodeFromWav(Utils::QStringToStdString(file), true, config.use_input_loop))
            {
                delete adx;
                return false;
            }
        }
        else if (is_aax)
        {
            if (!aax->EncodeFromWav(Utils::QStringToStdString(file), true, config.use_input_loop))
            {
                delete aax;
                return false;
            }
        }
        else
        {
            if (!hca->EncodeFromWav(Utils::QStringToStdString(file), config.quality, config.cutoff_freq, config.use_input_loop))
            {
                delete hca;
                return false;
            }
        }
    }
    else if (file.endsWith(".hca", Qt::CaseInsensitive) || file.endsWith(".adx", Qt::CaseInsensitive) || file.endsWith(".aax", Qt::CaseInsensitive))
    {        
        if (!audio->LoadFromFile(Utils::QStringToStdString(file)))
        {
            delete audio;
            return false;
        }
    }
    else
    {
        FfmpegAudioFile ffmpeg;

        if (!ffmpeg.LoadFromFile(Utils::QStringToStdString(file)))
        {
            delete audio;
            return false;
        }

        std::vector<AudioFile *> list;
        list.push_back(audio);

        if (!ffmpeg.ToFiles(list, ffmpeg.GetNumChannels(), false, config.use_input_loop))
        {
            delete audio;
            return false;
        }
    }

    if (loop)
    {
        if (!config.use_input_loop || !audio->HasLoop())
        {
            audio->SetMaxLoop();
        }
    }
    else
    {
        audio->RemoveLoop();
    }

    if (is_hca)
    {
        uint16_t ciph_type;

        if (config.ciph_type == CIPH_TYPE_AUTO)
        {
            if (guessed_ciph_type >= 0)
                ciph_type = guessed_ciph_type;
            else
                ciph_type = 0;
        }
        else
        {
            ciph_type = config.ciph_type;
        }

        //DPRINTF("Changing from %x to %x\n", hca->GetCiphType(), ciph_type);
        hca->SetCiphType(ciph_type);
    }

    AudioTrack track;

    track.file = audio;
    track.external = true;
    track.modified = true;
    track.index = index;
    track.state = AudioTrack::State::STOPPED;

    if (is_adx)
        track.type = AudioTrack::Type::ADX;
    else if (is_aax)
        track.type = AudioTrack::Type::AAX;
    else
        track.type = AudioTrack::Type::HCA;

    assert(index >= 0 && index < tracks.size());
    if (tracks[index].file)
        delete tracks[index].file;

    tracks[index] = track;    

    return true;
}

bool MainWindow::AddTrack(const QString &file, bool loop)
{
    AudioTrack track;

    track.file = nullptr;
    track.external = true;
    track.modified = true;
    track.index = tracks.size();
    track.state = AudioTrack::State::STOPPED;
    track.type = AudioTrack::Type::HCA;

    tracks.push_back(track);

    if (!ReplaceTrack(track.index, file, loop))
    {
        tracks.removeLast();
        return false;
    }

    return true;
}

bool MainWindow::ExtractJoinTracks(const std::vector<AudioFile *> &tracks, const QString &file)
{
    WavFile wav;

    if (!wav.FromFiles(tracks, (config.use_float) ? AUDIO_FORMAT_FLOAT : config.bit_depth, config.preserve_loop))
    {
        DPRINTF("WavFile::FromFiles failed.\n");
        return false;
    }

    if (!wav.SaveToFile(Utils::QStringToStdString(file)))
        return false;

    return true;
}

bool MainWindow::ReplaceSplitTracks(AudioFile &input, const QVector<AudioTrack *> &tracks, bool loop)
{
    std::vector<AudioFile *> files;

    if (tracks.size() == 0)
        return false;

    int max_threads = 0;

    for (AudioTrack *const &track : tracks)
    {
        AudioFile *file;

        assert(track && track->type != AudioTrack::Type::BIN);

        if (track->type == AudioTrack::Type::HCA)
        {
            file = new HcaFile();
            HcaFile::SetDefaultQuality(config.quality);
            HcaFile::SetDefaultCutoff(config.cutoff_freq);
            max_threads = (Thread::PhisycalCoresCount() == 1) ? 1 : 2;
        }
        else if (track->type == AudioTrack::Type::ADX)
        {
            file = new AdxFile();
        }
        else if (track->type == AudioTrack::Type::AAX)
        {
            file = new AaxFile();
        }
        else
        {
            DPRINTF("%s: Programming error.\n", FUNCNAME);
            exit(-1);
            return false;
        }

        files.push_back(file);
    }

    if (!input.ToFiles(files, tracks[0]->file->GetNumChannels(), true, loop, max_threads))
    {
        for (AudioFile *&audio : files)
        {
            delete audio;
        }

        DPRINTF("AudioFile::ToFiles failed.\n");
        return false;
    }

    for (int i = 0; i < tracks.size(); i++)
    {
        AudioFile *file = files[i];
        HcaFile *hca = dynamic_cast<HcaFile *>(file);

        if (loop)
        {
            if (!config.use_input_loop || !file->HasLoop())
            {
                file->SetMaxLoop();
            }
        }
        else
        {
            file->RemoveLoop();
        }

        if (hca)
        {
            uint16_t ciph_type;

            if (config.ciph_type == CIPH_TYPE_AUTO)
            {
                if (guessed_ciph_type >= 0)
                    ciph_type = guessed_ciph_type;
                else
                    ciph_type = 0;
            }
            else
            {
                ciph_type = config.ciph_type;
            }

            hca->SetCiphType(ciph_type);
        }

        tracks[i]->external = true;
        tracks[i]->modified = true;
        tracks[i]->state = AudioTrack::State::STOPPED;
        tracks[i]->file = file;
    }

    return true;
}

bool MainWindow::ConvertFileToHCA(const QString &input, const QString &output, bool loop)
{
    HcaFile hca;

    if (input.endsWith(".wav", Qt::CaseInsensitive))
    {
        if (!hca.EncodeFromWav(Utils::QStringToStdString(input), config.quality, config.use_input_loop))
            return false;
    }
    else
    {
        FfmpegAudioFile ffmpeg;

        if (!ffmpeg.LoadFromFile(Utils::QStringToStdString(input)))
            return false;

        std::vector<AudioFile *> list;
        list.push_back(&hca);

        if (!ffmpeg.ToFiles(list, ffmpeg.GetNumChannels(), false, config.preserve_loop))
            return false;
    }

    if (loop)
    {
        if (!config.use_input_loop || !hca.HasLoop())
        {
            hca.SetMaxLoop();
        }
    }
    else
    {
        hca.RemoveLoop();
    }

    uint16_t ciph_type;

    if (config.ciph_type == CIPH_TYPE_AUTO)
        ciph_type = 0;
    else
        ciph_type = config.ciph_type;

    hca.SetCiphType(ciph_type);

    return hca.SaveToFile(Utils::QStringToStdString(output));
}

/*bool MainWindow::ExtractTracks(const QVector<AudioTrack *> &tracks, const QString &directory)
{
    bool decode = false;
    bool ask_question = false;

    if (tracks.size() <= 0)
        return false;

    for (const AudioTrack *track : tracks)
    {
        assert(track);

        if (!track->external && track->type != AudioTrack::Type::BIN)
            ask_question = true;
    }

    if (ask_question && QMessageBox::question(this, "Extract as wav?", "Extract as wav?\nIf you press no, files will be extracted as .hca", QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No), QMessageBox::Yes) == QMessageBox::Yes)
    {
        decode = true;
    }

    for (const AudioTrack *track : tracks)
    {
        assert(track);

        if (track->external)
            continue;

        QString file = directory + QString("%1").arg(track->index+1, 4, 10, QChar('0'));

        if (track->type == AudioTrack::Type::HCA)
        {
            file += (decode) ? ".wav" : ".hca";
        }
        else
        {
            file += ".bin";
        }

        if (!ExtractTrack(track, file))
        {
            return false;
        }
    }

    return true;
}*/
int MainWindow::ExtractTracks(const QVector<AudioTrack *> &tracks, const QString &directory)
{
    bool decode = false;
    bool ask_question = false;

    if (tracks.size() <= 0)
        return -1;

    for (const AudioTrack *track : tracks)
    {
        assert(track);

        if (!track->external && track->type != AudioTrack::Type::BIN)
            ask_question = true;
    }

    if (ask_question && QMessageBox::question(this, "Extract as wav?", "Extract as wav?\nIf you press no, files will be extracted as .hca/.adx/.aax", QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No), QMessageBox::Yes) == QMessageBox::Yes)
    {
        decode = true;
    }

    WorkerDialog dialog(this);

    dialog.setExportMode(tracks, directory, decode);
    return dialog.exec();
}

// first_file should be set to true for first_file in multi replace
// track can be nullptr! (case of unknown file type in the awb)
bool MainWindow::CheckCompatibility(const QString &file, const AudioFile *track, bool multi, bool first_file, bool *cancel)
{
    static QString multi_footer = "(\"No\" will abort the import of this file only, \"Cancel\" will abort the import of all files.)";
    static QString proceed_msg = "\nDo you want to continue anyway?";

    static int ask_smpl, ask_channels, ask_channels2;
    uint16_t num_channels = 0;
    uint16_t max_channels = 8;

    if (track)
    {
        const AdxFile *is_adx = dynamic_cast<const AdxFile *>(track);
        const AaxFile *is_aax = dynamic_cast<const AaxFile *>(track);

        if (is_adx || is_aax)
            max_channels = 2;
    }

    if (!multi)
        first_file = true;

    if (first_file)
    {
        ask_smpl = 1;
        ask_channels = 1;
        ask_channels2 = 1;
    }

    *cancel = false;

    if (file.endsWith(".wav", Qt::CaseInsensitive))
    {
        WavFile wav;

        if (wav.LoadFromFile(Utils::QStringToStdString(file), false))
        {
            num_channels = wav.GetNumChannels();

            if (wav.HasLoop())
            {
                uint32_t start, end;
                int count;

                start = end = 0;
                wav.GetLoopSample(&start, &end, &count);

                if (start > wav.GetNumSamples() || end > wav.GetNumSamples())
                {
                    bool fix_file = false;

                    if (ask_smpl == 0)
                        fix_file = true;

                    if (ask_smpl == 1)
                    {
                        QMessageBox::StandardButtons buttons;
                        QMessageBox::StandardButton default_button;

                        QString msg = "File: \"" + file + "\"\n";
                        msg += "WAV SMPL header has wrong data, which will make encoder fail (this usually happens to some files generated by HCA decoder).\n";
                        msg += "\nDo you want me to fix the wav file before the encoding?";

                        if (multi)
                        {
                            buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::YesToAll|QMessageBox::No|QMessageBox::NoToAll|QMessageBox::Cancel);
                        }
                        else
                        {
                            buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No);
                        }

                        default_button = QMessageBox::Yes;

                        int ret = QMessageBox::question(this, "Eternity Audio Tool", msg, buttons, default_button);

                        if (ret == QMessageBox::NoToAll)
                        {
                            ask_smpl = -1;
                        }
                        else if (ret == QMessageBox::Yes)
                        {
                            fix_file = true;
                        }
                        else if (ret == QMessageBox::YesToAll)
                        {
                            fix_file = true;
                            ask_smpl = 0;
                        }
                        else if (ret == QMessageBox::Cancel)
                        {
                            *cancel = true;
                            return false;
                        }
                    } // ask_smpl == 1

                    if (fix_file)
                    {
                        if (!wav.MoveToMemory())
                        {
                            DPRINTF("%s: MoveToMemory failed.\n", FUNCNAME);
                            return false;
                        }

                        if (start > wav.GetNumSamples())
                            start = wav.GetNumSamples();
                        else if (end > wav.GetNumSamples())
                            end = wav.GetNumSamples();

                        wav.SetLoopSample(start, end, count);

                        if (!wav.SaveToFile(Utils::QStringToStdString(file)))
                            return false;
                    }
                }
            }
        }
    }
    else if (ask_channels != 0 || ask_channels2 != 0)
    {
        if (file.endsWith(".adx", Qt::CaseInsensitive))
        {
            uint32_t channels;

            if (AdxFile::CheckHeader(Utils::QStringToStdString(file), &channels))
            {
                num_channels = (uint16_t)channels;
            }
        }
        else if (file.endsWith(".aax", Qt::CaseInsensitive))
        {
            AaxFile aax;

            if (aax.LoadFromFile(Utils::QStringToStdString(file), false))
            {
                num_channels = aax.GetNumChannels();
            }
        }
        else if (file.endsWith(".hca", Qt::CaseInsensitive))
        {
            uint32_t channels;

            if (HcaFile::CheckHeader(Utils::QStringToStdString(file), &channels))
            {
                num_channels = (uint16_t)channels;
            }
        }
        else
        {
            uint16_t channels;

            if (FfmpegAudioFile::CheckHeader(Utils::QStringToStdString(file), &channels))
            {
                num_channels = channels;
            }
        }
    }

    if (num_channels > max_channels)
    {
        if (ask_channels == 0)
            return false;

        if (ask_channels == 1)
        {
            QMessageBox::StandardButtons buttons;
            QMessageBox::StandardButton default_button;

            QString msg = "File: \"" + file + "\"\n";
            msg += "This file has more channels that format support. Encoder will fail.\n";
            msg += proceed_msg;

            if (multi)
            {
                msg += "\n" + multi_footer;
                buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::YesToAll|QMessageBox::No|QMessageBox::NoToAll|QMessageBox::Cancel);
            }
            else
            {
                buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No);
            }

            default_button = QMessageBox::No;

            int ret = QMessageBox::question(this, "Eternity Audio Tool", msg, buttons, default_button);

            if (ret == QMessageBox::No)
            {
                return false;
            }
            else if (ret == QMessageBox::NoToAll)
            {
                ask_channels = 0;
                return false;
            }
            else if (ret == QMessageBox::YesToAll)
            {
                ask_channels = -1;
            }
            else if (ret == QMessageBox::Cancel)
            {
                *cancel = true;
                return false;
            }
        }
    }
    else if (num_channels != 0 && track && num_channels != track->GetNumChannels())
    {
        if (ask_channels2 == 0)
            return false;

        if (ask_channels2 == 1)
        {
            QMessageBox::StandardButtons buttons;
            QMessageBox::StandardButton default_button;

            QString msg = "File: \"" + file + "\"\n";
            msg += "This file has a number of channels different to the one in the track.\n";
            msg += "Channels in the track: " + QString("%1").arg(track->GetNumChannels()) + "\n";
            msg += "Channels in the file: " + QString("%1").arg(num_channels) + "\n";
            msg += proceed_msg;

            if (multi)
            {
                msg += "\n" + multi_footer;
                buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::YesToAll|QMessageBox::No|QMessageBox::NoToAll|QMessageBox::Cancel);
            }
            else
            {
                buttons = QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No);
            }

            default_button = QMessageBox::No;

            int ret = QMessageBox::question(this, "Eternity Audio Tool", msg, buttons, default_button);

            if (ret == QMessageBox::No)
            {
                return false;
            }
            else if (ret == QMessageBox::NoToAll)
            {
                ask_channels2 = 0;
                return false;
            }
            else if (ret == QMessageBox::YesToAll)
            {
                ask_channels2 = -1;
            }
            else if (ret == QMessageBox::Cancel)
            {
                *cancel = true;
                return false;
            }
        }
    }

    return true;
}

int MainWindow::ReplaceTracks(QVector<AudioTrack *> &tracks, const QString &directory)
{
    if (tracks.size() <= 0)
        return -1;

    bool first = true;

    for (int i = 0; i < tracks.size(); i++)
    {
        AudioTrack *track = tracks[i];

        assert(track);

        QString file = directory + GetNameForTrack(*track); // QString("%1").arg(track->index+1, 4, 10, QChar('0'));
        std::string file_std = Utils::QStringToStdString(file);
        bool cancel = false;
        bool remove = false;

        if (Utils::FileExists(file_std + ".wav"))
        {
            file += ".wav";
        }        
        else if (Utils::FileExists(file_std + ".hca") && track->type != AudioTrack::Type::ADX && track->type != AudioTrack::Type::AAX)
        {
            file += ".hca";
        }
        else if (Utils::FileExists(file_std + ".adx") && track->type == AudioTrack::Type::ADX)
        {
            file += ".adx";
        }
        else if (Utils::FileExists(file_std + ".aax") && track->type == AudioTrack::Type::AAX)
        {
            file += ".aax";
        }
        else
        {
            remove = true;

            for (const std::string &ext : supported_ffmpeg_extensions)
            {
                if (Utils::FileExists(file_std + ext))
                {
                    file += Utils::StdStringToQString(ext);
                    remove = false;
                    break;
                }
            }
        }

        if (!remove)
        {
            // NOTE: Don't do i == 0 to replace first variable, because if first track is removed, we will go back to index 0
            bool ret =  CheckCompatibility(file, track->file, true, first, &cancel);
            first = false;            

            if (!ret)
            {
                if (cancel)
                    return 0;

                remove = true;
            }
        }

        if (remove)
        {
            tracks.remove(i);
            i--;
        }
        else
        {
            if (track->index == track_playing)
            {
                StopCurrentTrack();
            }
        }
    }

    // We became empty after checking files
    if (tracks.size() == 0)
        return 1;

    WorkerDialog dialog(this);

    dialog.setImportMode(tracks, directory, LoopQuestion());
    return dialog.exec();
}

void MainWindow::PlayTrack(AudioTrack *track)
{
    QMutexLocker locker(&audio_mutex);

    if (!track->file)
        return;

    int bits;
    size_t size;

    uint8_t *buf = track->file->Decode(&bits, &size);
    if (!buf)
    {
        DPRINTF("Cannot play this file.\n");
        return;
    }

    QAudioFormat format;

    format.setSampleRate(track->file->GetSampleRate());
    format.setChannelCount(track->file->GetNumChannels());
    format.setCodec("audio/pcm");
    format.setByteOrder(QAudioFormat::LittleEndian);

    if (bits == AUDIO_FORMAT_FLOAT)
    {
        format.setSampleSize(32);
        format.setSampleType(QAudioFormat::Float);
    }
    else
    {
        format.setSampleSize(bits);
        format.setSampleType(QAudioFormat::SignedInt);
    }

    QAudioDeviceInfo info(QAudioDeviceInfo::defaultOutputDevice());
    if (!info.isFormatSupported(format))
    {
        DPRINTF("Cannot play that format.\n");
        delete[] buf;
        return;
    }

    if (audio_buffer)
        delete audio_buffer;

    audio_buffer = new QBuffer();
    audio_buffer->setData((const char *)buf, (int)size);
    delete[] buf;

    audio_buffer->open(QIODevice::ReadOnly);

    if (audio_output)
        delete audio_output;

    track_playing = track->index;
    audio_time = 0;

    audio_output = new QAudioOutput(format, this);    
    if (audio_output->notifyInterval() < 1000)
        audio_output->setNotifyInterval(1000);

    connect(audio_output, SIGNAL(notify()), this, SLOT(onPlaybackTimeChanged()));
    connect(audio_output, SIGNAL(stateChanged(QAudio::State)), this, SLOT(onPlaybackStateChanged(QAudio::State)));
    audio_output->start(audio_buffer);
}

void MainWindow::PlaySelectedTrack()
{
     QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

     if (selected.size() == 0)
         return;

     AudioTrack *track = GetTrack(selected[0]);

     if (track_playing != track->index)
     {
         StopCurrentTrack();
     }
     else
     {
         if (track->state == AudioTrack::State::PLAYING)
         {
            assert(audio_output);
            audio_output->suspend();
            return;
         }
         else if (track->state == AudioTrack::State::PAUSED)
         {
             assert(audio_output);
             audio_output->resume();
             return;
         }
     }

     PlayTrack(track);
}

void MainWindow::StopCurrentTrack()
{
    if (track_playing >= 0 && track_playing < tracks.size())
    {
        assert(audio_output);

        {
            QMutexLocker locker(&audio_mutex);

            tracks[track_playing].state = AudioTrack::State::STOPPED;
            UpdateTrackFromListUI(track_playing, false);            
            track_playing = -1;
            on_tracksList_itemSelectionChanged();
        }

        audio_output->reset();
    }
}

void MainWindow::onPlaybackStateChanged(QAudio::State state)
{
    if (state == QAudio::IdleState || state == QAudio::StoppedState)
    {
        QMutexLocker locker(&audio_mutex);

        if (audio_output)
        {
            delete audio_output;
            audio_output = nullptr;
        }

        if (audio_buffer)
        {
            delete audio_buffer;
            audio_buffer = nullptr;
        }

        if (track_playing >= 0 && track_playing < tracks.size())
        {
            tracks[track_playing].state = AudioTrack::State::STOPPED;
            UpdateTrackFromListUI(track_playing, false);
            track_playing = -1;
        }

        on_tracksList_itemSelectionChanged();
    }
    else if (state == QAudio::ActiveState)
    {
        if (track_playing >= 0 && track_playing < tracks.size())
        {
            tracks[track_playing].state = AudioTrack::State::PLAYING;
            UpdateTrackFromListUI(track_playing, false);
        }

        on_tracksList_itemSelectionChanged();
    }
    else if (state == QAudio::SuspendedState)
    {
        if (track_playing >= 0 && track_playing < tracks.size())
        {
            tracks[track_playing].state = AudioTrack::State::PAUSED;
            UpdateTrackFromListUI(track_playing, false);
        }

        on_tracksList_itemSelectionChanged();
    }
}

void MainWindow::onPlaybackTimeChanged()
{
    QMutexLocker locker(&audio_mutex);
    QTreeWidgetItem *item;

    item = GetItem(track_playing);

    if (item)
    {
        AudioTrack *track = GetTrack(item);

        audio_time += audio_output->notifyInterval();
        item->setText(COLUMN_DURATION, DurationToQString((float)audio_time/1000.0) + " / " + DurationToQString(track->file->GetDuration()));
    }
}

bool MainWindow::CanJoin(const QList<QTreeWidgetItem *> &items)
{
    if (items.size() <= 1 || items.size() > 8)
        return false;

    AudioTrack *track = GetTrack(items[0]);
    assert(track);
    AudioFile *file = track->file;

    if (!file)
        return false;

    uint32_t num_samples = file->GetNumSamples();
    uint32_t sample_rate = file->GetSampleRate();
    uint16_t num_channels = file->GetNumChannels();

    for (int i = 1; i < items.size(); i++)
    {
        AudioTrack *track = GetTrack(items[i]);

        if (!track->file)
            return false;

        if (track->file->GetNumSamples() != num_samples || track->file->GetNumChannels() != num_channels ||
            track->file->GetSampleRate() != sample_rate)
        {
            return false;
        }
    }

    return true;
}

QString MainWindow::GetNameForTrack(const AudioTrack &track)
{
    QString name = QString("%1").arg(track.index, 4, 10, QChar('0'));

    if (!track.name.isEmpty())
    {
        name += "_" + track.name;
    }

    return name;
}

void MainWindow::on_tracksList_itemSelectionChanged()
{
    if (clearing_list)
        return;

    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();
    bool disable_extract_all = true;

    for (const AudioTrack &track : tracks)
    {
        if (!track.external)
        {
            disable_extract_all = false;
            break;
        }
    }

    ui->actionStop->setDisabled(track_playing < 0 || track_playing >= tracks.size());

    if (selected.size() == 0)
    {
        ui->actionPlay->setDisabled(true);
        ui->actionPause->setDisabled(true);
        ui->actionExtract->setDisabled(true);
        ui->actionExtractJoin->setDisabled(true);
        ui->actionReplace->setDisabled(true);
        ui->actionReplaceSplit->setDisabled(true);
        ui->actionLoop->setDisabled(true);
    }
    else if (selected.size() == 1)
    {
        AudioTrack *track = GetTrack(selected[0]);
        assert(track);

        ui->actionPlay->setDisabled(track->type == AudioTrack::Type::BIN);
        ui->actionPause->setDisabled(true);
        ui->actionExtract->setDisabled(track->external);
        ui->actionExtractJoin->setDisabled(true);
        ui->actionReplace->setEnabled(true);
        ui->actionReplaceSplit->setDisabled(true);
        ui->actionLoop->setEnabled(track->file);
    }
    else
    {
        AudioTrack *track = GetTrack(selected[0]);
        bool disable_extract = true;
        bool can_join = CanJoin(selected);

        ui->actionPlay->setDisabled(track->type == AudioTrack::Type::BIN);
        ui->actionPause->setDisabled(true);
        ui->actionReplace->setEnabled(true);

        for (QTreeWidgetItem *item : selected)
        {
            AudioTrack *track = GetTrack(item);
            assert(track);

            if (!track->external)
            {
                disable_extract = false;
                break;
            }
        }

        ui->actionExtract->setDisabled(disable_extract);
        ui->actionExtractJoin->setDisabled(disable_extract || !can_join);
        ui->actionReplaceSplit->setDisabled(!can_join);
        ui->actionLoop->setDisabled(!can_join);
    }

    ui->actionExtract_all->setDisabled(disable_extract_all);
    ui->actionReplace_all->setDisabled(tracks.size() == 0);   

    if (selected.size() >= 1)
    {
        AudioTrack *track = GetTrack(selected[0]);

        if (track && track->index == track_playing && track->state == AudioTrack::State::PLAYING)
        {
            ui->actionPlay->setDisabled(true);
            ui->actionPause->setEnabled(true);
        }
        else
        {
            ui->actionPlay->setEnabled(track->type != AudioTrack::Type::BIN);
            ui->actionPause->setDisabled(true);
        }
    }
}

void MainWindow::on_actionExtract_triggered()
{
    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

    if (selected.size() == 0)
        return;

    if (selected.size() == 1)
    {
        AudioTrack *track = GetTrack(selected[0]);
        assert(track && !track->external);

        QString file = GetNameForTrack(*track); // selected[0]->text(COLUMN_TRACK);
        QString filter;

        if (track->type == AudioTrack::Type::HCA)
        {
            file += ".wav";
            filter = "WAV Files (*.wav);;HCA files (*.hca)";
        }
        else if (track->type == AudioTrack::Type::ADX)
        {
            file += ".wav";
            filter = "WAV Files (*.wav);;ADX files (*.adx)";
        }
        else if (track->type == AudioTrack::Type::AAX)
        {
            file += ".wav";
            filter = "WAV Files (*.wav);;AAX files (*.aax)";
        }
        else
        {
            file += ".bin";
            filter = "Binary files (*.bin)";
        }

        if (config.last_directory_export.length() > 0)
        {
            if (config.last_directory_export.endsWith('/') || config.last_directory_export.endsWith('\\'))
            {
                file = config.last_directory_export + file;
            }
            else
            {
                file = config.last_directory_export + '/' + file;
            }
        }

        file = QFileDialog::getSaveFileName(this, "Save", file, filter);

        if (file.isNull())
            return;

        SavePath(file, config.last_directory_export, false);

        if (!ExtractTrack(track, file))
            return;

        UPRINTF("File extracted succesfully.\n");
    }
    else
    {
        QString directory = QFileDialog::getExistingDirectory(this, "Extract", config.last_directory_export);
        if (directory.isNull())
            return;

        SavePath(directory, config.last_directory_export, true);

        if (!directory.endsWith('/') && !directory.endsWith('\\'))
            directory += '/';

        QVector<AudioTrack *> tracks;

        for (QTreeWidgetItem *item : selected)
        {
            AudioTrack *track = GetTrack(item);
            assert(track);

            tracks.push_back(track);
        }

        int troolean = ExtractTracks(tracks, directory);
        if (troolean > 0)
        {
            UPRINTF("Files extracted succesfully.\n");
        }
        else if (troolean == 0)
        {
            DPRINTF("Extract canceled by user.\n");
        }
        else
        {
            DPRINTF("Error extracting.\n");
        }
    }
}

void MainWindow::on_actionExtract_all_triggered()
{
    if (tracks.size() == 0)
        return;

    QString directory = QFileDialog::getExistingDirectory(this, "Extract", config.last_directory_export);
    if (directory.isNull())
        return;

    SavePath(directory, config.last_directory_export, true);

    if (!directory.endsWith('/') && !directory.endsWith('\\'))
        directory += '/';

    QVector<AudioTrack *> all_tracks;

    for (AudioTrack &track : tracks)
    {
        all_tracks.push_back(&track);
    }

    int troolean = ExtractTracks(all_tracks, directory);
    if (troolean > 0)
    {
        UPRINTF("Files extracted succesfully.\n");
    }
    else if (troolean == 0)
    {
        DPRINTF("Extract canceled by user.\n");
    }
    else
    {
        DPRINTF("Error extracting.\n");
    }
}

void MainWindow::on_actionExtractJoin_triggered()
{
    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

    if (selected.size() <= 1 || selected.size() > 8)
        return; // Shouldn't happen

    qSort(selected.begin(), selected.end(), [](const QTreeWidgetItem* a, const QTreeWidgetItem* b) -> bool { return a->text(COLUMN_TRACK) < b->text(COLUMN_TRACK); });

    QString file = selected[0]->text(COLUMN_TRACK) + '_' + selected[selected.size()-1]->text(COLUMN_TRACK) + ".wav";

    if (config.last_directory_export.length() > 0)
    {
        if (config.last_directory_export.endsWith('/') || config.last_directory_export.endsWith('\\'))
        {
            file = config.last_directory_export + file;
        }
        else
        {
            file = config.last_directory_export + '/' + file; // add2
        }
    }

    file = QFileDialog::getSaveFileName(this, "Save", file, "WAV Files (*.wav)");

    if (file.isNull())
        return;

    SavePath(file, config.last_directory_export, false);

    std::vector<AudioFile *> tracks;

    for (QTreeWidgetItem *item : selected)
    {
        AudioTrack *track = GetTrack(item);
        assert(track && track->file);

        tracks.push_back(track->file);
    }

    if (!ExtractJoinTracks(tracks, file))
        return;

    UPRINTF("File extracted succesfully.\n");
}

void MainWindow::on_actionPlay_triggered()
{
    PlaySelectedTrack();
}

void MainWindow::on_actionPause_triggered()
{
    PlaySelectedTrack();
}

void MainWindow::on_actionStop_triggered()
{
    StopCurrentTrack();
}

void MainWindow::on_tracksList_doubleClicked(const QModelIndex &index)
{
    UNUSED(index);
    PlaySelectedTrack();
}

void MainWindow::on_actionExit_triggered()
{
    if (ProcessShutdown())
        qApp->exit();
}

bool MainWindow::LoopQuestion()
{
    LoopQuestionDialog dialog(this, config.last_was_looped, config.use_input_loop);

    int ret = dialog.exec();

    if (ret)
    {
        config.use_input_loop = (ret & USE_INPUT_LOOP);
        config.last_was_looped = true;
        return true;
    }

    config.last_was_looped = false;
    return false;
}

void MainWindow::on_actionReplace_triggered()
{
    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

    if (selected.size() == 0)
        return;

    if (selected.size() == 1)
    {
        AudioTrack *track = GetTrack(selected[0]);
        assert(track);

        QString filter = "Standard audio files (*.wav";

        for (const std::string &ext : supported_ffmpeg_extensions)
        {
            filter += " *" + Utils::StdStringToQString(ext);
        }

        filter += ')';

        if (track->type == AudioTrack::Type::ADX)
        {
            filter += ";;ADX files (*.adx)";
        }
        else if (track->type == AudioTrack::Type::AAX)
        {
            filter += ";;AAX files (*.aax)";
        }
        else
        {
            filter += ";;HCA files (*.hca)";
        }

        QString file = QFileDialog::getOpenFileName(this, "Open", config.last_directory_import, filter);

        if (file.isNull())
            return;

        SavePath(file, config.last_directory_import, false);

        bool cancel;

        if (!CheckCompatibility(file, track->file, false, true, &cancel))
            return;

        if (track->index == track_playing)
            StopCurrentTrack();

        if (ReplaceTrack(track->index, file, LoopQuestion()))
        {
            ProcessItemUI(track->index, nullptr);
            on_tracksList_itemSelectionChanged();
            UPRINTF("File imported succesfully.\n");
        }
    }
    else
    {
        QString directory = QFileDialog::getExistingDirectory(this, "Extract", config.last_directory_import);
        if (directory.isNull())
            return;

        SavePath(directory, config.last_directory_import, true);

        if (!directory.endsWith('/') && !directory.endsWith('\\'))
            directory += '/';

        QVector<AudioTrack *> tracks;

        for (QTreeWidgetItem *item : selected)
        {
            AudioTrack *track = GetTrack(item);
            assert(track);

            tracks.push_back(track);
        }

        int troolean = ReplaceTracks(tracks, directory);
        if (troolean > 0)
        {
            if (tracks.size() == 0)
            {
                UPRINTF("There are no files to import.\n");
            }
            else
            {
                UPRINTF("%d files imported succesfully.\n", tracks.size());
            }
        }
        else if (troolean == 0)
        {
            DPRINTF("Import canceled by user.\n");
        }
        else
        {
            DPRINTF("Error happened in import.\n");
        }

        for (const AudioTrack *track : tracks)
        {
            ProcessItemUI(track->index, nullptr);
        }

        on_tracksList_itemSelectionChanged();
    }
}

void MainWindow::on_actionReplace_all_triggered()
{
    QString directory = QFileDialog::getExistingDirectory(this, "Extract", config.last_directory_import);
    if (directory.isNull())
        return;

    SavePath(directory, config.last_directory_import, true);

    if (!directory.endsWith('/') && !directory.endsWith('\\'))
        directory += '/';

    QVector<AudioTrack *> all_tracks;

    for (AudioTrack &track : tracks)
    {
        all_tracks.push_back(&track);
    }

    int troolean = ReplaceTracks(all_tracks, directory);
    if (troolean > 0)
    {
        if (all_tracks.size() == 0)
        {
            UPRINTF("There are no files to import.\n");
        }
        else
        {
            UPRINTF("%d files imported succesfully.\n", all_tracks.size());
        }
    }
    else if (troolean == 0)
    {
        DPRINTF("Import canceled by user.\n");
    }
    else
    {
        DPRINTF("Error happened in import.\n");
    }

    for (const AudioTrack *track : all_tracks)
    {
        ProcessItemUI(track->index, nullptr);
    }

    on_tracksList_itemSelectionChanged();
}

void MainWindow::on_actionReplaceSplit_triggered()
{
    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

    if (selected.size() <= 1 || selected.size() > 8)
        return; // Shouldn't happen

    qSort(selected.begin(), selected.end(), [](const QTreeWidgetItem* a, const QTreeWidgetItem* b) -> bool { return a->text(COLUMN_TRACK) < b->text(COLUMN_TRACK); });

    QString filter = "Standard audio files (*.wav";

    for (const std::string &ext : supported_ffmpeg_extensions)
    {
        filter += " *" + Utils::StdStringToQString(ext);
    }

    filter += ')';

    QString file = QFileDialog::getOpenFileName(this, "Open", config.last_directory_import, filter);

    if (file.isNull())
        return;

    SavePath(file, config.last_directory_import, false);

    AudioFile *audio;

    if (file.endsWith(".wav", Qt::CaseInsensitive))
    {
        audio = new WavFile();
    }
    else
    {
        audio = new FfmpegAudioFile();
    }

    if (!audio->LoadFromFile(Utils::QStringToStdString(file)))
    {
        delete audio;
        return;
    }

    assert(GetTrack(selected[0]->file));

    uint16_t channels = GetTrack(selected[0])->file->GetNumChannels();
    uint16_t total_channels = channels * (uint16_t)selected.size();
    uint16_t audio_channels = audio->GetNumChannels();

    if (total_channels < audio_channels)
    {
        DPRINTF("Number of channels in input file is greater than those combined in the selected tracks.");
        delete audio;
        return;
    }

    if ((audio_channels % channels) != 0)
    {
        DPRINTF("Number of channels in the input file should be multiple of any of the selected tracks channels.\n"
                "(%d is not multiple of %d)\n", (int)audio_channels, (int)channels);
        delete audio;
        return;
    }

    QVector<AudioTrack *> tracks;

    if (total_channels > audio_channels)
    {
        if (QMessageBox::question(this, "Silence remaining?", "The number of tracks in the wav is smaller than those combined in the selected tracks.\n\n"
                                  "Do you want to silence the remaining tracks? (pressing \"No\" will cancel import)", QMessageBox::StandardButtons(QMessageBox::Yes|QMessageBox::No),
                                  QMessageBox::Yes) == QMessageBox::No)
        {
            delete audio;
            return;
        }

        int num_silence = total_channels-audio_channels;

        QVector<int> silence_select;
        QVector<QString> tracks_names;

        silence_select.resize(selected.size());
        tracks_names.resize(selected.size());

        for (int i = 0; i < selected.size(); i++)
        {
            tracks_names[i] = selected[i]->text(COLUMN_TRACK);
        }

        if (config.last_silence_total == selected.size() && num_silence == config.last_silence_select.size())
        {
            for (int i : config.last_silence_select)
            {
                assert(i >= 0 && i < selected.size());
                silence_select[i] = true;
            }
        }
        else
        {
            for (int i = 0; i < num_silence; i++)
            {
                silence_select[selected.size()-i-1] = true;
            }
        }

        SelectSilenceDialog dialog(this, tracks_names, &silence_select, num_silence);

        if (dialog.exec() <= 0)
        {
            delete audio;
            return;
        }

        config.last_silence_total = selected.size();
        config.last_silence_select.clear();

        // Put first the non-silenced
        for (int i = 0; i < selected.size(); i++)
        {
            QTreeWidgetItem *item = selected[i];

            if (!silence_select[i])
            {
                AudioTrack *track = GetTrack(item);
                assert(track);

                if (track->index == track_playing)
                    StopCurrentTrack();

                tracks.push_back(track);
            }
        }

        // Put the silenced now
        for (int i = 0; i < selected.size(); i++)
        {
            QTreeWidgetItem *item = selected[i];

            if (silence_select[i])
            {
                AudioTrack *track = GetTrack(item);
                assert(track);

                if (track->index == track_playing)
                    StopCurrentTrack();

                tracks.push_back(track);
                config.last_silence_select.push_back(i);
            }
        }
    }
    else
    {
        for (QTreeWidgetItem *item : selected)
        {
            AudioTrack *track = GetTrack(item);
            assert(track);

            if (track->index == track_playing)
                StopCurrentTrack();

            tracks.push_back(track);
        }
    }

    if (ReplaceSplitTracks(*audio, tracks, LoopQuestion()))
    {
        for (AudioTrack *track : tracks)
            ProcessItemUI(track->index, nullptr);

        on_tracksList_itemSelectionChanged();
        UPRINTF("File imported succesfully.\n");
    }

    delete audio;
}

void MainWindow::on_actionOptions_triggered()
{
    OptionsDialog dialog(this);

    if (dialog.exec())
        config.Save();
}

void MainWindow::on_actionLoop_triggered()
{
    QList<QTreeWidgetItem *> selected = ui->tracksList->selectedItems();

    if (selected.size() < 1) // shouldn't happen
        return;

    if (selected.size() > 1)
    {
        qSort(selected.begin(), selected.end(), [](const QTreeWidgetItem* a, const QTreeWidgetItem* b) -> bool { return a->text(COLUMN_TRACK) < b->text(COLUMN_TRACK); });
    }

    QVector<AudioFile *> files;

    for (QTreeWidgetItem *item : selected)
    {
        files.push_back(GetTrack(item)->file);
    }

    LoopDialog dialog(this, files);

    if (dialog.exec())
    {
        for (QTreeWidgetItem *item : selected)
        {
            AudioTrack *track = GetTrack(item);

            track->modified = true;
            ProcessItemUI(track->index, nullptr);
        }
    }
}

void MainWindow::on_actionAddTrack_triggered()
{
    AcbFile *acb_acb = dynamic_cast<AcbFile *>(acb);
    if (!acb_acb)
        return;

    AddTrackDialog dialog(acb_acb, external_awb, internal_awb, this);

    if (!dialog.exec())
        return;

    int loop_sel = dialog.GetLoop();

    if (loop_sel == 1)
        config.use_input_loop = false;
    else if (loop_sel == 2)
        config.use_input_loop = true;

    if (!AddTrack(dialog.GetFile(), loop_sel > 0))
    {
        DPRINTF("AddTrack failed.\n");
        return;
    }

    AwbFile *awb = (dialog.IsExternalAwb()) ? external_awb : internal_awb;
    if (!awb->AddFile("DUMMY", 5))
    {
        tracks.removeLast();
        DPRINTF("AwbFile::AddFile failed.\n");
        return;
    }

    AudioFile *audio = tracks[tracks.size()-1].file;

    if (!acb_acb->AddTrack(Utils::QStringToStdString(dialog.GetName()), awb->GetNumFiles()-1, dialog.IsExternalAwb(),
                           (uint32_t)(audio->GetDuration()*1000.0f), audio->GetNumSamples(), audio->HasLoop(),
                           audio->GetSampleRate(), audio->GetNumChannels(),
                           dialog.GetCmd1(), dialog.GetCmd2(), nullptr, nullptr, nullptr))
    {
        DPRINTF("Cannot add data to the acb. Program cannot continue and will close.\n");
        qApp->exit();
        return;
    }

    AddTrackToListUI(tracks.size()-1, false);
    UPRINTF("Track added succesfully!\n");
}

void MainWindow::on_actionConvertHCA_triggered()
{
    QString filter = "Standard audio files (*.wav";

    for (const std::string &ext : supported_ffmpeg_extensions)
    {
        filter += " *" + Utils::StdStringToQString(ext);
    }

    filter += ')';

    QStringList files = QFileDialog::getOpenFileNames(this, "Select files", config.last_directory_import, filter);
    if (files.isEmpty())
        return;

    SavePath(files[0], config.last_directory_import, false);

    if (files.size() == 1)
    {
        QString output = files[0];
        int ext = output.lastIndexOf('.');

        if (ext == -1)
            output += ".hca";
        else
            output = output.left(ext) + ".hca";

        if (!ConvertFileToHCA(files[0], output, LoopQuestion()))
        {
            DPRINTF("Conversion failed.\n");
        }
        else
        {
            UPRINTF("Conversion done succesfully!\n");
        }
    }
    else
    {
        bool loop = LoopQuestion();

        WorkerDialog dialog(this);
        dialog.SetEncoderMode(files, loop);

        int troolean = dialog.exec();

        if (troolean > 0)
        {
            UPRINTF("Files converted succesfully.\n");
        }
        else if (troolean == 0)
        {
            DPRINTF("Conversion canceled by user.\n");
        }
        else
        {
            DPRINTF("Error in conversion.\n");
        }
    }
}

void MainWindow::on_actionAssociate_acb_extension_triggered()
{
    std::string reg_data;
    std::string program = Utils::QStringToStdString(qApp->applicationFilePath().replace('/', '\\').replace('\\', "\\\\"));

    reg_data = "Windows Registry Editor Version 5.00\r\n\r\n";
    reg_data += "[HKEY_CURRENT_USER\\Software\\Classes\\ACB File\\shell\\open\\command]\r\n";
    reg_data += "@=\"" + program +  " \\\"%1\\\"\"\r\n";
    reg_data += "[HKEY_CURRENT_USER\\Software\\Classes\\.acb]\r\n";
    reg_data += "@=\"ACB File\"\r\n";

    std::string temp_file = Utils::GetTempFile("ete", ".reg");

    if (!Utils::WriteFileBool(temp_file, (uint8_t *)reg_data.c_str(), reg_data.length()))
        return;

    Utils::RunCmd("regedit", { temp_file });
}

void MainWindow::on_actionAbout_triggered()
{
    QMessageBox::information(this, "Eternity Audio Tool", "Eternity Audio Tool 1.0alpha1 by eternity");
}



