#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QLabel>
#include <QTreeWidgetItem>
#include <QMutexLocker>
#include <QAudioOutput>
#include <QBuffer>

#include "AcbFile.h"
#include "UassetAcbFile.h"
#include "CsbFile.h"
#include "PseudoCriwareAudioContainer.h"
#include "Afs2File.h"
#include "CpkFile.h"
#include "CsbInternalAwbFile.h"
#include "PseudoAwbFile.h"
#include "WavFile.h"
#include "FfmpegAudioFile.h"
#include "HcaFile.h"
#include "AdxFile.h"
#include "AaxFile.h"

namespace Ui {
class MainWindow;
}

struct AudioTrack
{
    AudioFile *file;

    enum class Type
    {
        HCA,
        ADX,
        AAX,
        BIN
    } type;

    bool modified;
    bool external;

    enum class State
    {
        STOPPED,
        PLAYING,
        PAUSED
    } state = State::STOPPED;

    int index;
    QString name;
};

extern const std::vector<std::string> supported_ffmpeg_extensions;

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(QWidget *parent = 0);
    ~MainWindow();

    bool LoadFile(const QString &file);
    bool SaveFile(const QString &file);

    bool ExtractTrack(const AudioTrack *track, const QString &file);    
    bool ReplaceTrack(int index, const QString &file, bool loop);
    bool AddTrack(const QString &file, bool loop);

    bool ExtractJoinTracks(const std::vector<AudioFile *> &tracks, const QString &file);
    bool ReplaceSplitTracks(AudioFile &input, const QVector<AudioTrack *>&tracks, bool loop);

    bool ConvertFileToHCA(const QString &input, const QString &output, bool loop);

    QString GetNameForTrack(const AudioTrack &track);

public slots:

    void onJobCancel()
    {
        job_cancel = true;
    }

protected:
    void closeEvent(QCloseEvent *event);

private slots:

    void onPlaybackStateChanged(QAudio::State state);
    void onPlaybackTimeChanged();

    void on_actionOpen_triggered();

    void on_tracksList_itemSelectionChanged();

    void on_actionExtract_triggered();

    void on_actionExtract_all_triggered();

    void on_actionPlay_triggered();

    void on_actionPause_triggered();

    void on_actionStop_triggered();

    void on_tracksList_doubleClicked(const QModelIndex &index);


    void on_actionExit_triggered();

    void on_actionReplace_triggered();

    void on_actionReplace_all_triggered();

    void on_actionAbout_triggered();

    void on_actionSave_triggered();

    void on_actionSave_as_triggered();

    void on_actionOptions_triggered();

    void on_actionAssociate_acb_extension_triggered();

    void on_actionLoop_triggered();

    void on_actionExtractJoin_triggered();

    void on_actionReplaceSplit_triggered();

    void on_actionOpenAudio_triggered();

    void on_actionAddTrack_triggered();

    void on_actionConvertHCA_triggered();

private:
    Ui::MainWindow *ui;

    QLabel *statusLabel;   

    CriwareAudioContainer *acb;
    AwbFile *external_awb;
    AwbFile *internal_awb;
    QString acb_path;

    int guessed_ciph_type;

    QVector<AudioTrack> tracks;

    QAudioOutput *audio_output;
    QBuffer *audio_buffer;

    QMutex main_mutex;
    QMutex audio_mutex;

    int track_playing;
    int audio_time;

    bool job_cancel;
    bool clearing_list = false;

    static QString DurationToQString(float duration);

    bool ProcessShutdown();

    void SavePath(const QString &file_path, QString &directory, bool is_directory);
    static QString GetAwbPath(const QString &file, bool cpk);
    bool DoLoadFile(const QString &file, bool reload, bool multiple=false);
    bool DoSaveFile(const QString &file, bool multiple=false);

    bool IsFileModified();

    void ProcessItemUI(int index, QTreeWidgetItem *item);
    void AddTrackToListUI(int index, bool select);
    void UpdateTrackFromListUI(int index, bool select);
    void ClearTracksListUI();

    AudioTrack *GetTrack(const QTreeWidgetItem *item);    
    QTreeWidgetItem *GetItem(int index);

    int ExtractTracks(const QVector<AudioTrack *> &tracks, const QString &directory);
    int ReplaceTracks(QVector<AudioTrack *> &tracks, const QString &directory);

    void PlayTrack(AudioTrack *track);
    void PlaySelectedTrack();
    void StopCurrentTrack();

    bool LoopQuestion();

    bool CheckCompatibility(const QString &file, const AudioFile *track, bool multi, bool first_file, bool *cancel);    
    bool CanJoin(const QList<QTreeWidgetItem *> &items);    

signals:

    void progressCalculated(int max_progress);
    void progressChanged();
};

#endif // MAINWINDOW_H
