#include "workerdialog.h"
#include "ui_workerdialog.h"

WorkerDialog::WorkerDialog(QWidget *parent) :
    QDialog(parent), ui(new Ui::WorkerDialog)
{
    ui->setupUi(this);
    setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);

    window = dynamic_cast<MainWindow *>(parent);
    assert(window);
}

WorkerDialog::~WorkerDialog()
{
    delete ui;
}

void WorkerDialog::setExportMode(const QVector<AudioTrack *> &tracks, const QString &directory, bool decode)
{
    this->tracks = tracks;
    this->path = directory;
    this->decode = decode;
    this->mode = WorkerDialogMode::EXPORT;
    this->max_jobs = tracks.size();

    ui->label->setText("Extracting files...");
    setWindowTitle("Extracting files...");
    ui->progressBar->setMaximum(tracks.size());
}

void WorkerDialog::setImportMode(const QVector<AudioTrack *> &tracks, const QString &directory, bool loop)
{
    this->tracks = tracks;
    this->path = directory;
    this->loop = loop;
    this->mode = WorkerDialogMode::IMPORT;
    this->max_jobs = tracks.size();

    ui->label->setText("Importing files...");
    setWindowTitle("Importing files...");
    ui->progressBar->setMaximum(tracks.size());
}

void WorkerDialog::SetLoadMode(const QString &file)
{
    path = file;
    mode = WorkerDialogMode::LOAD;

    ui->label->setText("Loading file...");
    setWindowTitle("Loading file...");
}

void WorkerDialog::SetSaveMode(const QString &file)
{
    path = file;
    mode = WorkerDialogMode::SAVE;

    ui->label->setText("Saving file...");
    setWindowTitle("Saving file...");
    ui->cancelButton->setDisabled(true);
}

void WorkerDialog::SetEncoderMode(const QStringList &files, bool loop)
{
    this->files = files;
    this->loop = loop;
    this->mode = WorkerDialogMode::ENCODER;
    this->max_jobs = files.size();

    ui->label->setText("Encoding files...");
    setWindowTitle("Encoding files...");
    ui->progressBar->setMaximum(files.size());
}

int WorkerDialog::exec()
{
    if (mode == WorkerDialogMode::EXPORT)
        DoExport();
    else if (mode == WorkerDialogMode::IMPORT)
        DoImport();
    else if (mode == WorkerDialogMode::LOAD)
        DoLoad();
    else if (mode == WorkerDialogMode::SAVE)
        DoSave();
    else if (mode == WorkerDialogMode::ENCODER)
        DoEncoder();

    return QDialog::exec();
}

void WorkerDialog::DoExport()
{
    assert(tracks.size() > 0);

    QThreadPool *pool = QThreadPool::globalInstance();
    QVector<ExportWork *> works;

    works.resize(tracks.size());

    for (int i = 0; i < tracks.size(); i++)
    {
        QString file = path + window->GetNameForTrack(*tracks[i]); //QString("%1").arg(tracks[i]->index+1, 4, 10, QChar('0'));

        if (tracks[i]->type == AudioTrack::Type::HCA)
        {
            file += (decode) ? ".wav" : ".hca";
        }
        else if (tracks[i]->type == AudioTrack::Type::ADX)
        {
            file += (decode) ? ".wav" : ".adx";
        }
        else if (tracks[i]->type == AudioTrack::Type::AAX)
        {
            file += (decode) ? ".wav" : ".aax";
        }
        else
        {
            file += ".bin";
        }

        works[i] = new ExportWork(window, tracks[i], file);
        connect(works[i], SIGNAL(workFinished()), this, SLOT(onWorkFinished()));
        connect(this, SIGNAL(cancelSignal()), works[i], SLOT(onCancel()));
        connect(works[i], SIGNAL(errorSignal()), this, SLOT(onError()));
    }

    jobs_finished = 0;    

    pool->setMaxThreadCount(QThread::idealThreadCount());

    // The more threads, the more memory usage, and it is big. We need a limit.
    if (pool->maxThreadCount() > 16)
        pool->setMaxThreadCount(16);    

    //pool->setMaxThreadCount(1); // For slower testing

    for (int i = 0; i < tracks.size(); i++)
    {
        pool->start(works[i]);
    }
}

void WorkerDialog::DoImport()
{
    assert(tracks.size() > 0);

    QThreadPool *pool = QThreadPool::globalInstance();
    QVector<ImportWork *> works;

    works.resize(tracks.size());

    for (int i = 0; i < tracks.size(); i++)
    {
        QString file = path + window->GetNameForTrack(*tracks[i]); //      QString("%1").arg(tracks[i]->index+1, 4, 10, QChar('0'));
        std::string file_std = Utils::QStringToStdString(file);

        if (Utils::FileExists(file_std + ".wav"))
        {
            file += ".wav";
        }
        else
        {
            if (tracks[i]->type == AudioTrack::Type::ADX && Utils::FileExists(file_std + ".adx"))
                file += ".adx";
            else if (tracks[i]->type == AudioTrack::Type::AAX && Utils::FileExists(file_std + ".aax"))
                file += ".aax";
            else if (Utils::FileExists(file_std + ".hca"))
                file += ".hca";
            else
            {
                for (const std::string &ext : supported_ffmpeg_extensions)
                {
                    if (Utils::FileExists(file_std + ext))
                    {
                        file += Utils::StdStringToQString(ext);
                    }
                }
            }
        }

        works[i] = new ImportWork(window, tracks[i], file, loop);
        connect(works[i], SIGNAL(workFinished()), this, SLOT(onWorkFinished()));
        connect(this, SIGNAL(cancelSignal()), works[i], SLOT(onCancel()));
        connect(works[i], SIGNAL(errorSignal()), this, SLOT(onError()));
    }

    jobs_finished = 0;

    pool->setMaxThreadCount(QThread::idealThreadCount());

    // The hca encoder program already uses N threads, where N=logical cores.
    // We can still gain some speed on some systems, but not in all of them
    if (tracks[0]->type != AudioTrack::Type::ADX)
    {
        if (pool->maxThreadCount() > 4)
            pool->setMaxThreadCount(2);
        else
            pool->setMaxThreadCount(1);
    }

    for (int i = 0; i < tracks.size(); i++)
    {
        pool->start(works[i]);
    }
}

void WorkerDialog::DoLoad()
{
    QThreadPool *pool = QThreadPool::globalInstance();
    LoadWork *work = new LoadWork(window, path);

    connect(window, SIGNAL(progressCalculated(int)), this, SLOT(onSetWorkSize(int)));
    connect(window, SIGNAL(progressChanged()), this, SLOT(onWorkFinished()));
    connect(this, SIGNAL(cancelSignal()), work, SLOT(onCancel()));
    connect(this, SIGNAL(cancelSignal()), window, SLOT(onJobCancel()));
    connect(work, SIGNAL(errorSignal()), this, SLOT(onError()));

    pool->start(work);
}

void WorkerDialog::DoSave()
{
    QThreadPool *pool = QThreadPool::globalInstance();
    SaveWork *work = new SaveWork(window, path);

    connect(window, SIGNAL(progressCalculated(int)), this, SLOT(onSetWorkSize(int)));
    connect(window, SIGNAL(progressChanged()), this, SLOT(onWorkFinished()));
    //connect(this, SIGNAL(cancelSignal()), work, SLOT(onCancel()));
    //connect(this, SIGNAL(cancelSignal()), window, SLOT(onJobCancel()));
    connect(work, SIGNAL(errorSignal()), this, SLOT(onError()));

    pool->start(work);
}

void WorkerDialog::DoEncoder()
{
    QThreadPool *pool = QThreadPool::globalInstance();
    QVector<EncoderWork *> works;

    works.resize(files.size());

    for (int i = 0; i < files.size(); i++)
    {
        works[i] = new EncoderWork(window, files[i], loop);
        connect(works[i], SIGNAL(workFinished()), this, SLOT(onWorkFinished()));
        connect(this, SIGNAL(cancelSignal()), works[i], SLOT(onCancel()));
        connect(works[i], SIGNAL(errorSignal()), this, SLOT(onError()));
    }

    jobs_finished = 0;

    pool->setMaxThreadCount(QThread::idealThreadCount());

    if (pool->maxThreadCount() > 4)
        pool->setMaxThreadCount(2);
    else
        pool->setMaxThreadCount(1);

    for (int i = 0; i < files.size(); i++)
    {
        pool->start(works[i]);
    }
}

void WorkerDialog::onWorkFinished()
{
    QMutexLocker locker(&mutex);

    jobs_finished++;
    ui->progressBar->setValue(jobs_finished);

    if (jobs_finished == max_jobs)
    {
        QThreadPool::globalInstance()->waitForDone();
        done(1);
    }
}

void WorkerDialog::reject()
{
    QMutexLocker locker(&mutex);

    if (jobs_finished != max_jobs)
    {
        ui->label->setText("Cancelling...");

        QThreadPool::globalInstance()->clear();
        emit cancelSignal();

        QThreadPool::globalInstance()->waitForDone();
        done(0);
    }
}

void WorkerDialog::onError()
{
    QMutexLocker locker(&mutex);

    ui->label->setText("There was an error. Cancelling all jobs...");

    QThreadPool::globalInstance()->clear();
    emit cancelSignal();

    QThreadPool::globalInstance()->waitForDone();
    done(-1);
}

void WorkerDialog::onSetWorkSize(int size)
{
    max_jobs = size;
    ui->progressBar->setMaximum(size);
}

void WorkerDialog::on_cancelButton_clicked()
{
    reject();
}

void ExportWork::run()
{
    assert(window && track && !file.isNull());

    if (cancel)
        return;

    if (!track->external)
    {
        bool success = window->ExtractTrack(track, file);
        // Simulate random error for testing
        /*if (success)
        {
            if (Utils::RandomInt(0, 5) == 3)
                success = false;
        }*/

        if (!success && !cancel)
        {
            emit errorSignal();
            return;
        }
    }

    if (cancel)
        return;

    emit workFinished();
}

void ExportWork::onCancel()
{
    cancel = true;
}

void ImportWork::run()
{
    assert(window && track && !file.isNull());

    if (cancel)
        return;

    bool success = window->ReplaceTrack(track->index, file, loop);

    if (!success && !cancel)
    {
        emit errorSignal();
        return;
    }

    if (cancel)
        return;

    emit workFinished();
}

void ImportWork::onCancel()
{
    cancel = true;
}

void LoadWork::run()
{
    assert(window);

    if (cancel)
        return;

    bool success = window->LoadFile(file);

    if (cancel)
        return;

    if (!success)
        emit errorSignal();
}

void LoadWork::onCancel()
{
    cancel = true;
}

void SaveWork::run()
{
    assert(window);

    if (cancel)
        return;

    bool success = window->SaveFile(file);

    if (cancel)
        return;

    if (!success)
        emit errorSignal();
}

void SaveWork::onCancel()
{
    cancel = true;
}

void EncoderWork::run()
{
    if (cancel)
        return;

    QString output = file;
    int ext = output.lastIndexOf('.');

    if (ext == -1)
        output += ".hca";
    else
        output = output.left(ext) + ".hca";


    bool success = window->ConvertFileToHCA(file, output, loop);

    if (!success && !cancel)
    {
        emit errorSignal();
        return;
    }

    if (cancel)
        return;

    emit workFinished();
}

void EncoderWork::onCancel()
{
    cancel = true;
}
