#ifndef WORKERDIALOG_H
#define WORKERDIALOG_H

#include <QDialog>
#include <QThreadPool>
#include <QMutexLocker>
#include "mainwindow.h"

namespace Ui {
class WorkerDialog;
}

enum class WorkerDialogMode
{
    IMPORT,
    EXPORT,
    LOAD,
    SAVE,
    ENCODER,
};

class ExportWork : public QObject, public QRunnable
{
    Q_OBJECT

public:

    ExportWork(MainWindow *window, const AudioTrack *track, const QString &file) : QRunnable(), window(window), track(track), file(file){ }

    void run();

public slots:

    void onCancel();

signals:

    void workFinished();
    void errorSignal();

private:

    MainWindow *window;
    const AudioTrack *track;
    QString file;

    bool cancel = false;
};

class ImportWork : public QObject, public QRunnable
{
    Q_OBJECT

public:

    ImportWork(MainWindow *window, const AudioTrack *track, const QString &file, bool loop)
        : QRunnable(), window(window), track(track), file(file), loop(loop) { }

    void run();

public slots:

    void onCancel();

signals:

    void workFinished();
    void errorSignal();

private:

    MainWindow *window;
    const AudioTrack *track;
    QString file;

    bool loop;
    bool cancel = false;
};

class LoadWork : public QObject, public QRunnable
{
    Q_OBJECT

public:

    LoadWork(MainWindow *window, const QString &file) : QRunnable(), window(window), file(file) {}

    void run();

public slots:

    void onCancel();

signals:

    void workFinished();
    void errorSignal();


private:

    MainWindow *window;
    QString file;

    bool cancel = false;
};

class SaveWork : public QObject, public QRunnable
{
    Q_OBJECT

public:

    SaveWork(MainWindow *window, const QString &file) : QRunnable(), window(window), file(file) {}

    void run();

public slots:

    void onCancel();

signals:

    void workFinished();
    void errorSignal();


private:

    MainWindow *window;
    QString file;

    bool cancel = false;
};

class EncoderWork : public QObject, public QRunnable
{
    Q_OBJECT

public:

    EncoderWork(MainWindow *window, const QString &file, bool loop) : window(window), file(file), loop(loop) { }

    void run();

public slots:

    void onCancel();

signals:

    void workFinished();
    void errorSignal();

private:

    MainWindow *window;
    QString file;
    bool loop;

    bool cancel = false;
};

class WorkerDialog : public QDialog
{
    Q_OBJECT

public slots:
    void onWorkFinished();    
    void reject() override;
    void onError();
    void onSetWorkSize(int size);

public:
    explicit WorkerDialog(QWidget *parent);
    ~WorkerDialog();

    void setExportMode(const QVector<AudioTrack *> &tracks, const QString &directory, bool decode);
    void setImportMode(const QVector<AudioTrack *> &tracks, const QString &directory, bool loop);
    void SetLoadMode(const QString &file);
    void SetSaveMode(const QString &file);
    void SetEncoderMode(const QStringList &files, bool loop);

    int exec() override;

signals:

    void cancelSignal();

private slots:
    void on_cancelButton_clicked();

private:
    Ui::WorkerDialog *ui;

    MainWindow *window;
    QVector<AudioTrack *> tracks;
    QString path;
    WorkerDialogMode mode;

    QStringList files;

    bool decode;
    bool loop;

    QMutex mutex;
    int jobs_finished = 0;
    int max_jobs;

    void DoExport();
    void DoImport();
    void DoLoad();
    void DoSave();
    void DoEncoder();
};

#endif // WORKERDIALOG_H
