#include <algorithm>
#include "Xenoverse2.h"

#include "CusFile.h"
#include "debug.h"

TiXmlElement *CusSkillSet::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("SkillSet");

    std::string name = Xenoverse2::GetCharaAndCostumeName(char_id, costume_id, skills[9], true);
    if (name.length() != 0)
        Utils::WriteComment(entry_root, name);

    Utils::WriteParamUnsigned(entry_root, "CHAR_ID", char_id, true);
    Utils::WriteParamUnsigned(entry_root, "COSTUME_ID", costume_id, true);
    Utils::WriteParamMultipleUnsigned(entry_root, "SKILLS", std::vector<uint16_t>(skills, skills+10), true);

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool CusSkillSet::Compile(const TiXmlElement *root)
{
    if (!Utils::GetParamUnsigned(root, "CHAR_ID", &char_id))
        return false;

    if (!Utils::GetParamUnsigned(root, "COSTUME_ID", &costume_id))
        return false;

    if (!Utils::GetParamMultipleUnsigned(root, "SKILLS", skills, 10))
        return false;

    return true;
}

TiXmlElement *CusSkill::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("Skill");

    entry_root->SetAttribute("name", name);
    entry_root->SetAttribute("id", Utils::UnsignedToString(id, true));
    entry_root->SetAttribute("name_id", Utils::UnsignedToString(name_id, true));

    Utils::WriteParamUnsigned(entry_root, "RACE_LOCK", race_lock, true);
    Utils::WriteParamUnsigned(entry_root, "TYPE", type, true);
    Utils::WriteParamUnsigned(entry_root, "U_0E", unk_0E, true);
    Utils::WriteParamUnsigned(entry_root, "HAIR", hair, true);
    Utils::WriteParamUnsigned(entry_root, "U_12", unk_12, true);
    Utils::WriteParamMultipleStrings(entry_root, "PATHS", std::vector<std::string>(paths, paths+7));
    Utils::WriteParamUnsigned(entry_root, "U_30", unk_30, true);
    Utils::WriteParamUnsigned(entry_root, "U_32", unk_32, true);
    Utils::WriteParamUnsigned(entry_root, "U_34", unk_34, true);
    Utils::WriteParamUnsigned(entry_root, "U_36", unk_36, true);
    Utils::WriteParamUnsigned(entry_root, "U_38", unk_38, true);
    Utils::WriteParamUnsigned(entry_root, "U_3A", unk_3A, true);
    Utils::WriteParamUnsigned(entry_root, "U_3C", unk_3C, true);
    Utils::WriteParamUnsigned(entry_root, "U_3E", unk_3E, true);
    Utils::WriteParamUnsigned(entry_root, "U_40", unk_40, true);

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool CusSkill::Compile(const TiXmlElement *root)
{
    if (!Utils::ReadAttrString(root, "name", name))
        return false;

    if (name.length() != 3 && name.length() != 4)
    {
        DPRINTF("%s: Name attribute must be exactly 3 or 4 chars: %s\n", FUNCNAME, name.c_str());
        return false;
    }

    if (!Utils::ReadAttrUnsigned(root, "id", &id))
        return false;

    if (!Utils::ReadAttrUnsigned(root, "name_id", &name_id))
        return false;

    if (!Utils::GetParamUnsigned(root, "RACE_LOCK", &race_lock))
        return false;

    if (!Utils::GetParamUnsigned(root, "TYPE", &type))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_0E", &unk_0E))
        return false;

    if (!Utils::GetParamUnsigned(root, "HAIR", &hair))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_12", &unk_12))
        return false;

    std::vector<std::string> paths_temp;

    if (!Utils::GetParamMultipleStrings(root, "PATHS", paths_temp))
        return false;

    if (paths_temp.size() != 7)
    {
        DPRINTF("%s: Number of strings in PATHS field must be 7 (found %Id).\n", FUNCNAME, paths_temp.size());
        return false;
    }

    for (size_t i = 0; i < 7; i++)
        paths[i] = paths_temp[i];

    if (!Utils::GetParamUnsigned(root, "U_30", &unk_30))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_32", &unk_32))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_34", &unk_34))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_36", &unk_36))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_38", &unk_38))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_3A", &unk_3A))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_3C", &unk_3C))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_3E", &unk_3E))
        return false;

    if (!Utils::GetParamUnsigned(root, "U_40", &unk_40))
        return false;

    return true;
}

CusFile::CusFile()
{
    this->big_endian = false;
}

CusFile::~CusFile()
{
}

void CusFile::Reset()
{
    skill_sets.clear();
    super_skills.clear();
    ultimate_skills.clear();
    evasive_skills.clear();
    unk_skills.clear();
    blast_skills.clear();
    awaken_skills.clear();
}

bool CusFile::LoadSkills(const uint8_t *top, const CUSSkill *sets_in, std::vector<CusSkill> &sets_out, uint32_t num)
{
    sets_out.resize(num);

    for (size_t i = 0; i < sets_out.size(); i++)
    {
        CusSkill &skill = sets_out[i];

        if (sets_in[i].unk_04 != 0)
        {
            DPRINTF("%s: a value is not zero as expected (entry 0x%Ix): 0x%x\n", FUNCNAME, i, sets_in[i].unk_04);
            return false;
        }

        skill.name = sets_in[i].name;
        skill.id = sets_in[i].id;
        skill.name_id = sets_in[i].name_id;
        skill.race_lock = sets_in[i].race_lock;
        skill.type = sets_in[i].type;
        skill.unk_0E = sets_in[i].unk_0E;
        skill.hair = sets_in[i].hair;
        skill.unk_12 = sets_in[i].unk_12;

        for (int j = 0; j < 7; j++)
        {
            skill.paths[j] = GetString(top, sets_in[i].paths_offsets[j]);
        }

        skill.unk_30 = sets_in[i].unk_30;
        skill.unk_32 = sets_in[i].unk_32;
        skill.unk_34 = sets_in[i].unk_34;
        skill.unk_36 = sets_in[i].unk_36;
        skill.unk_38 = sets_in[i].unk_38;
        skill.unk_3A = sets_in[i].unk_3A;
        skill.unk_3C = sets_in[i].unk_3C;
        skill.unk_3E = sets_in[i].unk_3E;
        skill.unk_40 = sets_in[i].unk_40;
    }

    return true;
}

bool CusFile::Load(const uint8_t *buf, size_t size)
{
    Reset();

    if (size < sizeof(CUSHeader))
        return false;

    const CUSHeader *hdr = (const CUSHeader *)buf;

    if (hdr->signature != CUS_SIGNATURE && memcmp(buf+1, "CUS", 3) != 0)
        return false;

    if (hdr->endianess_check != 0xFFFE)
        return false;

    const CUSSkillSet *skill_sets_f = (const CUSSkillSet *)GetOffsetPtr(buf, hdr->skillset_offset);
    skill_sets.resize(hdr->num_skillsets);

    for (size_t i = 0; i < skill_sets.size(); i++)
    {
        CusSkillSet &skill_set = skill_sets[i];

        if (skill_sets_f[i].unk_1C != 0)
        {
            DPRINTF("%s: Warning, unk_1C not zero (0x%x) as expected (in CUSSKillSet, near 0x%x).\n", FUNCNAME, skill_sets_f[i].unk_1C, Utils::DifPointer(skill_sets_f+i, buf));
            return false;
        }

        skill_set.char_id = skill_sets_f[i].char_id;
        skill_set.costume_id = skill_sets_f[i].costume_id;
        memcpy(skill_set.skills, skill_sets_f[i].skills, sizeof(skill_set.skills));
    }

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->super_offset), super_skills, hdr->num_super))
        return false;

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->ultimate_offset), ultimate_skills, hdr->num_ultimate))
        return false;

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->evasive_offset), evasive_skills, hdr->num_evasive))
        return false;

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->unk_offset), unk_skills, hdr->num_unk))
        return false;

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->blast_offset), blast_skills, hdr->num_blast))
        return false;

    if (!LoadSkills(buf, (const CUSSkill *)GetOffsetPtr(buf, hdr->awaken_offset), awaken_skills, hdr->num_awaken))
        return false;

    return true;
}

size_t CusFile::CalculateStringsSize(std::unordered_set<std::string> &list, const std::vector<CusSkill> &skills) const
{
    size_t size = 0;

    for (const CusSkill &skill : skills)
    {
        for (size_t i = 0; i < 7; i++)
        {
            size_t len = skill.paths[i].length();

            if (len == 0)
                continue;

            if (list.find(skill.paths[i]) != list.end())
                continue;

            size += len+1;
            list.insert(skill.paths[i]);
        }
    }

    return size;
}

size_t CusFile::CalculateFileSize() const
{
    std::unordered_set<std::string> strings_list;

    size_t size = sizeof(CUSHeader);
    size += skill_sets.size()*sizeof(CUSSkillSet);
    size += (super_skills.size()+ultimate_skills.size()+evasive_skills.size()+unk_skills.size()+blast_skills.size()+awaken_skills.size()) *sizeof(CUSSkill);

    size += CalculateStringsSize(strings_list, super_skills);
    size += CalculateStringsSize(strings_list, ultimate_skills);
    size += CalculateStringsSize(strings_list, evasive_skills);
    size += CalculateStringsSize(strings_list, unk_skills);
    size += CalculateStringsSize(strings_list, blast_skills);
    size += CalculateStringsSize(strings_list, awaken_skills);

    //DPRINTF("size = %x\n", size);

    return size;
}

void CusFile::SaveSkills(uint8_t *top, char *str_top, char **str_current, const std::vector<CusSkill> &sets_in, CUSSkill *sets_out, std::unordered_set<std::string> &strings_list)
{
    for (size_t i = 0; i < sets_in.size(); i++)
    {
        const CusSkill &skill = sets_in[i];

        strcpy(sets_out[i].name, skill.name.c_str());
        sets_out[i].id = skill.id;
        sets_out[i].name_id = skill.name_id;
        sets_out[i].race_lock = skill.race_lock;
        sets_out[i].type = skill.type;
        sets_out[i].unk_0E = skill.unk_0E;
        sets_out[i].hair = skill.hair;
        sets_out[i].unk_12 = skill.unk_12;

        for (size_t j = 0; j < 7; j++)
        {
            if (skill.paths[j].length() == 0)
            {
                sets_out[i].paths_offsets[j] = 0;
                continue;
            }

            if (strings_list.find(skill.paths[j]) == strings_list.end())
            {
                sets_out[i].paths_offsets[j] = Utils::DifPointer(*str_current, top);

                strings_list.insert(skill.paths[j]);
                strcpy(*str_current, skill.paths[j].c_str());
                *str_current += skill.paths[j].length() + 1;
            }
            else
            {
                const char *str = FindString(str_top, skill.paths[j].c_str(), strings_list.size());
                if (!str)
                {
                    DPRINTF("%s: Internal coding error.\n", FUNCNAME);
                    exit(-1);
                }

                sets_out[i].paths_offsets[j] = Utils::DifPointer(str, top);
            }
        }

        sets_out[i].unk_30 = skill.unk_30;
        sets_out[i].unk_32 = skill.unk_32;
        sets_out[i].unk_34 = skill.unk_34;
        sets_out[i].unk_36 = skill.unk_36;
        sets_out[i].unk_38 = skill.unk_38;
        sets_out[i].unk_3A = skill.unk_3A;
        sets_out[i].unk_3C = skill.unk_3C;
        sets_out[i].unk_3E = skill.unk_3E;
        sets_out[i].unk_40 = skill.unk_40;
    }
}

uint8_t *CusFile::Save(size_t *psize)
{
    std::unordered_set<std::string> strings_list;
    size_t size = CalculateFileSize();

    uint8_t *buf = new uint8_t[size];
    memset(buf, 0, size);

    CUSHeader *hdr = (CUSHeader *)buf;
    CUSSkillSet *skill_set_f = (CUSSkillSet *)(hdr+1);
    CUSSkill *super_skills_f = (CUSSkill *)(skill_set_f+skill_sets.size());
    CUSSkill *ultimate_skills_f = (CUSSkill *)(super_skills_f+super_skills.size());
    CUSSkill *evasive_skills_f = (CUSSkill *)(ultimate_skills_f+ultimate_skills.size());
    CUSSkill *unk_skills_f = (CUSSkill *)(evasive_skills_f+evasive_skills.size());
    CUSSkill *blast_skills_f = (CUSSkill *)(unk_skills_f+unk_skills.size());
    CUSSkill *awaken_skills_f = (CUSSkill *)(blast_skills_f+blast_skills.size());
    char *strings = (char *)(awaken_skills_f+awaken_skills.size());
    char *current_str = strings;

    for (size_t i = 0; i < skill_sets.size(); i++)
    {
        const CusSkillSet &skill_set = skill_sets[i];

        skill_set_f[i].char_id = skill_set.char_id;
        skill_set_f[i].costume_id = skill_set.costume_id;
        memcpy(skill_set_f[i].skills, skill_set.skills, sizeof(skill_set.skills));
    }

    SaveSkills(buf, strings, &current_str, super_skills, super_skills_f, strings_list);
    SaveSkills(buf, strings, &current_str, ultimate_skills, ultimate_skills_f, strings_list);
    SaveSkills(buf, strings, &current_str, evasive_skills, evasive_skills_f, strings_list);
    SaveSkills(buf, strings, &current_str, unk_skills, unk_skills_f, strings_list);
    SaveSkills(buf, strings, &current_str, blast_skills, blast_skills_f, strings_list);
    SaveSkills(buf, strings, &current_str, awaken_skills, awaken_skills_f, strings_list);

    hdr->signature = CUS_SIGNATURE;
    hdr->endianess_check = 0xFFFE;
    hdr->num_skillsets = (uint32_t)skill_sets.size();
    hdr->skillset_offset = Utils::DifPointer(skill_set_f, buf);

    hdr->num_super = (uint32_t)super_skills.size();
    hdr->num_ultimate = (uint32_t)ultimate_skills.size();
    hdr->num_evasive = (uint32_t)evasive_skills.size();
    hdr->num_unk = (uint32_t)unk_skills.size();
    hdr->num_blast = (uint32_t)blast_skills.size();
    hdr->num_awaken = (uint32_t)awaken_skills.size();

    hdr->super_offset = Utils::DifPointer(super_skills_f, buf);
    hdr->ultimate_offset = Utils::DifPointer(ultimate_skills_f, buf);
    hdr->evasive_offset = Utils::DifPointer(evasive_skills_f, buf);
    hdr->unk_offset = Utils::DifPointer(unk_skills_f, buf);
    hdr->blast_offset = Utils::DifPointer(blast_skills_f, buf);
    hdr->awaken_offset = Utils::DifPointer(awaken_skills_f, buf);

    *psize = size;
    return buf;
}

TiXmlElement *CusFile::DecompileSkills(const char *name, TiXmlNode *root, const std::vector<CusSkill> &skills) const
{
    TiXmlElement *entry_root = new TiXmlElement(name);

    for (const CusSkill &skill : skills)
        skill.Decompile(entry_root);

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool CusFile::CompileSkills(const TiXmlElement *root, std::vector<CusSkill> &skills)
{
    for (const TiXmlElement *elem = root->FirstChildElement(); elem; elem = elem->NextSiblingElement())
    {
        if (elem->ValueStr() == "Skill")
        {
           CusSkill skill;

           if (!skill.Compile(elem))
               return false;

           skills.push_back(skill);
        }
    }

    return true;
}

TiXmlDocument *CusFile::Decompile() const
{
    TiXmlDocument *doc = new TiXmlDocument();

    TiXmlDeclaration* decl = new TiXmlDeclaration("1.0", "utf-8", "" );
    doc->LinkEndChild(decl);

    TiXmlElement *root = new TiXmlElement("CUS");

    for (const CusSkillSet &ss : skill_sets)
    {
        ss.Decompile(root);
    }

    DecompileSkills("SuperSkills", root, super_skills);
    DecompileSkills("UltimateSkills", root, ultimate_skills);
    DecompileSkills("EvasiveSkills", root, evasive_skills);
    DecompileSkills("UnkSkills", root, unk_skills);
    DecompileSkills("BlastSkills", root, blast_skills);
    DecompileSkills("AwakenSkills", root, awaken_skills);

    doc->LinkEndChild(root);
    return doc;
}

bool CusFile::Compile(TiXmlDocument *doc, bool)
{
    Reset();

    TiXmlHandle handle(doc);
    const TiXmlElement *root = Utils::FindRoot(&handle, "CUS");

    if (!root)
    {
        DPRINTF("Cannot find\"CUS\" in xml.\n");
        return false;
    }

    for (const TiXmlElement *elem = root->FirstChildElement(); elem; elem = elem->NextSiblingElement())
    {
        const std::string &str = elem->ValueStr();

        if (str == "SkillSet")
        {
            CusSkillSet ss;

            if (!ss.Compile(elem))
                return false;

            skill_sets.push_back(ss);
        }
        else if (str == "SuperSkills")
        {
            if (!CompileSkills(elem, super_skills))
                return false;
        }
        else if (str == "UltimateSkills")
        {
            if (!CompileSkills(elem, ultimate_skills))
                return false;
        }
        else if (str == "EvasiveSkills")
        {
            if (!CompileSkills(elem, evasive_skills))
                return false;
        }
        else if (str == "UnkSkills")
        {
            if (!CompileSkills(elem, unk_skills))
                return false;
        }
        else if (str == "BlastSkills")
        {
            if (!CompileSkills(elem, blast_skills))
                return false;
        }
        else if (str == "AwakenSkills")
        {
            if (!CompileSkills(elem, awaken_skills))
                return false;
        }
    }

    return true;
}

size_t CusFile::FindSkillSetsByCharID(uint32_t char_id, std::vector<CusSkillSet *> &sets)
{
    sets.clear();

    for (CusSkillSet &set : skill_sets)
    {
        if (set.char_id == char_id)
        {
            sets.push_back(&set);
        }
    }

    return sets.size();
}

bool CusFile::AddSkillSet(const CusSkillSet &set, bool unique_char_id)
{
    if (unique_char_id)
    {
        std::vector<CusSkillSet *> temp;

        if (FindSkillSetsByCharID(set.char_id, temp) != 0)
            return false;
    }

    skill_sets.push_back(set);
    return true;
}

size_t CusFile::RemoveSkillSets(uint32_t char_id)
{
    size_t removed = 0;

    for (size_t i = 0; i < skill_sets.size(); i++)
    {
        if (skill_sets[i].char_id == char_id)
        {
            skill_sets.erase(skill_sets.begin()+i);
            i--;
            removed++;
        }
    }

    return removed;
}

CusSkill *CusFile::FindSuperSkillByID(uint32_t id)
{
    for (CusSkill &skill : super_skills)
    {
        if (skill.id == id)
            return &skill;
    }

    return nullptr;
}

CusSkill *CusFile::FindUltimateSkillByID(uint32_t id)
{
    for (CusSkill &skill : ultimate_skills)
    {
        if (skill.id == id)
            return &skill;
    }

    return nullptr;
}

CusSkill *CusFile::FindEvasiveSkillByID(uint32_t id)
{
    for (CusSkill &skill : evasive_skills)
    {
        if (skill.id == id)
            return &skill;
    }

    return nullptr;
}

CusSkill *CusFile::FindBlastSkillByID(uint32_t id)
{
    for (CusSkill &skill : blast_skills)
    {
        if (skill.id == id)
            return &skill;
    }

    return nullptr;
}

CusSkill *CusFile::FindAwakenSkillByID(uint32_t id)
{
    for (CusSkill &skill : awaken_skills)
    {
        if (skill.id == id)
            return &skill;
    }

    return nullptr;
}
