#ifndef __CUSFILE_H__
#define __CUSFILE_H__

#include <unordered_set>
#include "BaseFile.h"

#define CUS_SIGNATURE 0x53554323 // #CUS

#ifdef _MSC_VER
#pragma pack(push,1)
#endif

typedef struct
{
    uint32_t signature; // 0
    uint16_t endianess_check; // 4
    uint16_t unk_06; // 6 pad
    uint32_t num_skillsets; // 8
    uint32_t skillset_offset; // 0xC
    uint32_t num_super; // 0x10
    uint32_t num_ultimate; // 0x14;
    uint32_t num_evasive; // 0x18
    uint32_t num_unk; // 0x1C
    uint32_t num_blast; // 0x20
    uint32_t num_awaken; // 0x24
    uint32_t super_offset; // 0x28
    uint32_t ultimate_offset; // 0x2C
    uint32_t evasive_offset; // 0x30
    uint32_t unk_offset; // 0x34
    uint32_t blast_offset; // 0x38
    uint32_t awaken_offset; // 0x3C
    uint64_t unk_40; // 0x40  zero
} PACKED CUSHeader;

STATIC_ASSERT_STRUCT(CUSHeader, 0x48);

typedef struct
{
    uint32_t char_id; // 0
    uint32_t costume_id; // 4
    uint16_t skills[10]; // 8
    uint32_t unk_1C; // 0x1C check for zero
} PACKED CUSSkillSet;

STATIC_ASSERT_STRUCT(CUSSkillSet, 0x20);

typedef struct
{
    char name[4];
    uint32_t unk_04; // Check for 0
    uint16_t id; // 8
    uint16_t name_id; // 0xA
    uint8_t race_lock; // 0xC
    uint8_t type; // 0xD
    uint16_t unk_0E;
    uint16_t hair; // 0x10
    uint16_t unk_12; // 0x12
    uint32_t paths_offsets[7]; // 0x14
    uint16_t unk_30;
    uint16_t unk_32;
    uint16_t unk_34;
    uint16_t unk_36;
    uint16_t unk_38;
    uint16_t unk_3A;
    uint16_t unk_3C;
    uint16_t unk_3E;
    uint32_t unk_40;
} PACKED CUSSkill;

STATIC_ASSERT_STRUCT(CUSSkill, 0x44);

#ifdef _MSC_VER
#pragma pack(pop)
#endif

struct CusSkillSet
{
    uint32_t char_id;
    uint32_t costume_id;
    uint16_t skills[10];

    CusSkillSet()
    {
        char_id = 0;
        costume_id = 0;
        memset(skills, 0xFF, sizeof(skills));
        skills[9] = 0;
    }

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct CusSkill
{
    std::string name;
    uint16_t id;
    uint16_t name_id;
    uint8_t race_lock;
    uint8_t type;
    uint16_t unk_0E;
    uint16_t hair;
    uint16_t unk_12;
    // 0 -> ean
    // 1 -> cam.ean
    // 2 -> eepk
    // 3 -> acb (SE)
    // 4 -> acb (VOX)
    std::string paths[7];
    uint16_t unk_30;
    uint16_t unk_32;
    uint16_t unk_34;
    uint16_t unk_36;
    uint16_t unk_38;
    uint16_t unk_3A;
    uint16_t unk_3C;
    uint16_t unk_3E;
    uint32_t unk_40;

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

class CusFile : public BaseFile
{
private:

    std::vector<CusSkillSet> skill_sets;
    std::vector<CusSkill> super_skills;
    std::vector<CusSkill> ultimate_skills;
    std::vector<CusSkill> evasive_skills;
    std::vector<CusSkill> unk_skills;
    std::vector<CusSkill> blast_skills;
    std::vector<CusSkill> awaken_skills;

    bool LoadSkills(const uint8_t *top, const CUSSkill *sets_in, std::vector<CusSkill> &sets_out, uint32_t num);
    void SaveSkills(uint8_t *top, char *str_top, char **str_current, const std::vector<CusSkill> &sets_in, CUSSkill *sets_out, std::unordered_set<std::string> &strings_list);

    TiXmlElement *DecompileSkills(const char *name, TiXmlNode *root, const std::vector<CusSkill> &skills) const;
    bool CompileSkills(const TiXmlElement *root, std::vector<CusSkill> &skills);

protected:

    void Reset();
    size_t CalculateStringsSize(std::unordered_set<std::string> &list, const std::vector<CusSkill> &skills) const;
    size_t CalculateFileSize() const;

public:

    CusFile();
    virtual ~CusFile();

    virtual bool Load(const uint8_t *buf, size_t size) override;
    virtual uint8_t *Save(size_t *psize) override;

    virtual TiXmlDocument *Decompile() const override;
    virtual bool Compile(TiXmlDocument *doc, bool big_endian=false) override;

    inline size_t GetNumeSkillSets() const { return skill_sets.size(); }

    inline const std::vector<CusSkillSet> &GetSkillSets() const { return skill_sets; }
    inline std::vector<CusSkillSet> &GetSkillSets() { return skill_sets; }

    inline const CusSkillSet &GetSkillSet(size_t idx) const { return skill_sets[idx]; }
    inline CusSkillSet &GetSkillSet(size_t idx) { return skill_sets[idx]; }

    inline const std::vector<CusSkill> &GetSuperSkills() const { return super_skills; }
    inline std::vector<CusSkill> &GetSuperSkills() { return super_skills; }

    inline const std::vector<CusSkill> &GetUltimateSkills() const { return ultimate_skills; }
    inline std::vector<CusSkill> &GetUltimateSkills() { return ultimate_skills; }

    inline const std::vector<CusSkill> &GetEvasiveSkills() const { return evasive_skills; }
    inline std::vector<CusSkill> &GetEvasiveSkills() { return evasive_skills; }

    inline const std::vector<CusSkill> &GetBlastSkills() const { return blast_skills; }
    inline std::vector<CusSkill> &GetBlastSkills() { return blast_skills; }

    inline const std::vector<CusSkill> &GetAwakenSkills() const { return awaken_skills; }
    inline std::vector<CusSkill> &GetAwakenSkills() { return awaken_skills; }

    size_t FindSkillSetsByCharID(uint32_t char_id, std::vector<CusSkillSet *> &sets);
    bool AddSkillSet(const CusSkillSet &set, bool unique_char_id);
    size_t RemoveSkillSets(uint32_t char_id);

    CusSkill *FindSuperSkillByID(uint32_t id);
    CusSkill *FindUltimateSkillByID(uint32_t id);
    CusSkill *FindEvasiveSkillByID(uint32_t id);
    CusSkill *FindBlastSkillByID(uint32_t id);
    CusSkill *FindAwakenSkillByID(uint32_t id);
};

#endif // __CUSFILE_H__
