#ifndef __SEVFILE_H__
#define __SEVFILE_H__

#include "BaseFile.h"

#define SEV_SIGNATURE   0x56455323

#ifdef _MSC_VER
#pragma pack(push,1)
#endif

typedef struct
{
    uint32_t signature; // 0
    uint16_t endianess_check; // 4
    uint16_t header_size; // 6
    uint32_t num_entries; // 8
    uint32_t unk_0C; // zero
} PACKED SEVHeader;

STATIC_ASSERT_STRUCT(SEVHeader, 0x10);

typedef struct
{
    uint32_t char_id; // 0
    uint32_t costume_id; // 4
    uint32_t num_cev_entries; // 8
    uint32_t unk_0C; // zero
} PACKED SEVEntry;

STATIC_ASSERT_STRUCT(SEVEntry, 0x10);

typedef struct
{
    uint32_t char_id; // 0
    uint32_t costume_id;// 4
    uint32_t num_events; // 8
    uint32_t unk_0C; // 0
} PACKED SEVCharEvents;

STATIC_ASSERT_STRUCT(SEVCharEvents, 0x10);

typedef struct
{
    uint32_t unk_00;
    uint32_t num_entries;
    uint32_t unk_08; // Always zero
} PACKED SEVEvent;

STATIC_ASSERT_STRUCT(SEVEvent, 0xC);

typedef struct
{
    uint32_t unk_00;
    uint32_t unk_04;
    uint32_t unk_08;
    uint32_t unk_0C;
    uint32_t unk_10;
} PACKED SEVEventEntry;

STATIC_ASSERT_STRUCT(SEVEventEntry, 0x14);

#ifdef _MSC_VER
#pragma pack(pop)
#endif


struct SevEventEntry
{
    uint32_t unk_00;
    uint32_t unk_04;
    uint32_t unk_08;
    uint32_t unk_0C;
    uint32_t unk_10;

    SevEventEntry()
    {
        unk_00 = 0;
        unk_04 = 0;
        unk_08 = 0;
        unk_0C = 0;
        unk_10 = 0;
    };

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevEvent
{
    uint32_t unk_00;
    std::vector<SevEventEntry> entries;

    SevEvent()
    {
        unk_00 = 0;
    }

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevCharEvents
{
    uint32_t char_id;
    uint32_t costume_id;
    std::vector<SevEvent> events;

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevEntry
{
    uint32_t char_id;
    uint32_t costume_id;
    std::vector<SevCharEvents> chars_events;

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

class SevFile : public BaseFile
{
private:

    std::vector<SevEntry> entries;

protected:

    void Reset();
    size_t CalculateFileLayout(uint32_t *events_offset, uint32_t *ev_entries_offset) const;

public:

    SevFile();
    virtual ~SevFile();

    virtual bool Load(const uint8_t *buf, size_t size) override;
    virtual uint8_t *Save(size_t *psize) override;

    virtual TiXmlDocument *Decompile() const override;
    virtual bool Compile(TiXmlDocument *doc, bool big_endian=false) override;

    inline size_t GetNumEntries() const { return entries.size(); }
    SevEntry *FindEntry(uint32_t char_id, uint32_t costume_id);

    size_t RemoveAllReferencesToChar(uint32_t char_id);

    inline const std::vector<SevEntry> &GetEntries() const { return entries; }
    inline std::vector<SevEntry> &GetEntries() { return entries; }

    inline const SevEntry &operator[](size_t n) const { return entries[n]; }
    inline const SevEntry &operator[](size_t n) { return entries[n]; }

    inline std::vector<SevEntry>::const_iterator begin() const { return entries.begin(); }
    inline std::vector<SevEntry>::const_iterator end() const { return entries.end(); }

    inline std::vector<SevEntry>::iterator begin() { return entries.begin(); }
    inline std::vector<SevEntry>::iterator end() { return entries.end(); }
};

#endif // _SEVFILE_H__
