#include <algorithm>
#include "X2mFile.h"
#include "debug.h"

void X2mSlotEntry::CopyFrom(const CharaListSlotEntry &entry, bool name)
{
    costume_index = entry.costume_index;
    model_preset = entry.model_preset;
    flag_gk2 = entry.flag_gk2;
    voices_id_list[0] = entry.voices_id_list[0];
    voices_id_list[1] = entry.voices_id_list[1];
    audio_files[0].clear();
    audio_files[1].clear();

    if (name)
    {
        for (int i = 0; i < XV2_LANG_NUM; i++)
        {
            if (entry.code.length() != 5 || entry.code[0] != '"' || entry.code[4] != '"')
            {
                costume_name[i].clear();
            }
            else
            {
                std::string code = entry.code.substr(1, 3);

                if (!Xenoverse2::GetCharaCostumeName(code, costume_index, model_preset, costume_name[i], i))
                    costume_name[i].clear();
            }
        }
    }
}

void X2mSlotEntry::CopyTo(CharaListSlotEntry &entry, const std::string &code) const
{
    if (voices_id_list[0] == (int)X2M_DUMMY_ID || voices_id_list[1] == (int)X2M_DUMMY_ID)
    {
        DPRINTF("%s: PROGRAM BUG, audio should have been assigned before calling this function.\n", FUNCNAME);
        exit(-1);
        return;
    }

    entry.with_undefined = false;
    entry.code = std::string("\"") + code + std::string("\"");
    entry.costume_index = costume_index;
    entry.model_preset = model_preset;
    entry.unlock_index = 0;
    entry.flag_gk2 = flag_gk2;
    entry.voices_id_list[0] = voices_id_list[0];
    entry.voices_id_list[1] = voices_id_list[1];
    entry.dlc = "Dlc_Def";
}

TiXmlElement *X2mSlotEntry::Decompile(TiXmlNode *root, bool new_format) const
{
    TiXmlElement *entry_root = new TiXmlElement("SlotEntry");

    entry_root->SetAttribute("costume_index", costume_index);
    Utils::WriteParamUnsigned(entry_root, "MODEL_PRESET", (int64_t)model_preset, true);
    Utils::WriteParamString(entry_root, "FLAG_GK2", (flag_gk2) ? "true" : "false");

    std::vector<uint32_t> voices = { (uint32_t)voices_id_list[0], (uint32_t)voices_id_list[1] };
    Utils::WriteParamMultipleUnsigned(entry_root, "VOICES_ID_LIST", voices, true);

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        if (costume_name[i].length() == 0)
        {
            if (i == XV2_LANG_ENGLISH)
                return nullptr;
        }
        else
        {
            std::string name = "COSTUME_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);
            Utils::WriteParamString(entry_root, name.c_str(), costume_name[i]);
        }
    }

    if (new_format)
    {
        if (audio_files[0].length() > 0)
            Utils::WriteParamString(entry_root, "AUDIO0", audio_files[0]);

        if (audio_files[1].length() > 0)
            Utils::WriteParamString(entry_root, "AUDIO1", audio_files[1]);
    }

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool X2mSlotEntry::Compile(const TiXmlElement *root, bool new_format)
{
    if (!Utils::ReadAttrUnsigned(root, "costume_index", (uint32_t *)&costume_index))
    {
        DPRINTF("%s: costume_index attribute is not optional.\n", FUNCNAME);
        return false;
    }

    if (!Utils::GetParamUnsigned(root, "MODEL_PRESET", (uint32_t *)&model_preset))
        return false;

    std::string flag;

    if (!Utils::GetParamString(root, "FLAG_GK2", flag))
        return false;

    flag = Utils::ToLowerCase(flag);
    if (flag == "true" || flag == "1")
    {
        flag_gk2 = true;
    }
    else if (flag == "false" || flag == "0")
    {
        flag_gk2 = false;
    }
    else
    {
        DPRINTF("%s: Cannot parse FLAG_GK2 param (%s).\n", FUNCNAME, flag.c_str());
        return false;
    }

    if (!Utils::GetParamMultipleUnsigned(root, "VOICES_ID_LIST", (uint32_t *)voices_id_list, 2))
        return false;

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        std::string name = "COSTUME_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);

        if (i == XV2_LANG_ENGLISH)
        {
            if (!Utils::GetParamString(root, name.c_str(), costume_name[i]))
                return false;
        }
        else
        {
            Utils::ReadParamString(root, name.c_str(), costume_name[i]);
        }
    }

    audio_files[0].clear();
    audio_files[1].clear();

    if (new_format)
    {
        if (Utils::ReadParamString(root, "AUDIO0", audio_files[0]))
            voices_id_list[0] = X2M_DUMMY_ID;

        if (Utils::ReadParamString(root, "AUDIO1", audio_files[1]))
            voices_id_list[1] = X2M_DUMMY_ID;
    }

    return true;
}

TiXmlElement *SevEntryHL::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("SevEntryHL");

    entry_root->SetAttribute("costume_id", Utils::UnsignedToString(costume_id, true));
    entry_root->SetAttribute("copy_char", copy_char);
    entry_root->SetAttribute("copy_costume", Utils::UnsignedToString(copy_costume, true));

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool SevEntryHL::Compile(const TiXmlElement *root)
{
    if (!Utils::ReadAttrString(root, "copy_char", copy_char))
    {
        DPRINTF("%s: copy_char attribute is not optional.\n", FUNCNAME);
        return false;
    }

    if (!Utils::ReadAttrUnsigned(root, "costume_id", &costume_id))
        costume_id = 0xFFFFFFFF;

    if (!Utils::ReadAttrUnsigned(root, "copy_costume", &copy_costume))
        copy_costume = 0xFFFFFFFF;

    return true;
}

TiXmlElement *SevEventsLL::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("SubEntry");

    entry_root->SetAttribute("im_source", (im_source) ? "true" : "false");
    entry_root->SetAttribute("char_code", char_code);
    entry_root->SetAttribute("costume_id", Utils::UnsignedToString(costume_id, true));

    for (const SevEvent &event : events)
        event.Decompile(entry_root);

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool SevEventsLL::Compile(const TiXmlElement *root)
{
    std::string temp;

    if (!Utils::ReadAttrString(root, "im_source", temp))
    {
        DPRINTF("%s: im_source is not optional.\n", FUNCNAME);
        return false;
    }

    temp = Utils::ToLowerCase(temp);
    im_source = (temp == "true" || temp == "1");

    if (!Utils::ReadAttrString(root, "char_code", char_code))
    {
        if (!im_source)
        {
            DPRINTF("%s: char_code is not optional when im_source=false.\n", FUNCNAME);
            return false;
        }
    }

    if (!im_source && char_code.length() != 3)
    {
        DPRINTF("%s: char_code must be a 3-letter code when im_source=false.\n", FUNCNAME);
        return false;
    }

    if (!Utils::ReadAttrUnsigned(root, "costume_id", &costume_id))
    {
        DPRINTF("%s: costume_id is not optional.\n", FUNCNAME);
        return false;
    }

    events.clear();

    for (const TiXmlElement *elem = root->FirstChildElement(); elem; elem = elem->NextSiblingElement())
    {
        if (elem->ValueStr() == "Event")
        {
           SevEvent event;

           if (!event.Compile(elem))
               return false;

           events.push_back(event);
        }
    }

    return true;
}

TiXmlElement *SevEntryLL::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("SevEntryLL");

    entry_root->SetAttribute("costume_id", Utils::UnsignedToString(costume_id, true));

    for (const SevEventsLL &events : sub_entries)
        events.Decompile(entry_root);

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool SevEntryLL::Compile(const TiXmlElement *root)
{
    if (!Utils::ReadAttrUnsigned(root, "costume_id", &costume_id))
    {
        DPRINTF("%s: costume_id is not optional\n", FUNCNAME);
        return false;
    }

    for (const TiXmlElement *elem = root->FirstChildElement(); elem; elem = elem->NextSiblingElement())
    {
        if (elem->ValueStr() == "SubEntry")
        {
           SevEventsLL events;

           if (!events.Compile(elem))
               return false;

           sub_entries.push_back(events);
        }       
    }

    return true;
}

TiXmlElement *X2mCharaName::Decompile(TiXmlNode *root) const
{
    TiXmlElement *entry_root = new TiXmlElement("X2mCharaName");
    entry_root->SetAttribute("costume_index", costume_index);

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        const std::string &name = chara_name[i];

        if (name.length() == 0)
        {
            if (i == XV2_LANG_ENGLISH)
            {
                DPRINTF("%s: Name for language english is not optional.\n", FUNCNAME);
                return nullptr;
            }

            continue;
        }

        const std::string temp_str = "CHARA_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);
        Utils::WriteParamString(entry_root, temp_str.c_str(), name);
    }

    root->LinkEndChild(entry_root);
    return entry_root;
}

bool X2mCharaName::Compile(const TiXmlElement *root)
{
    if (!Utils::ReadAttrUnsigned(root, "costume_index", &costume_index))
    {
       DPRINTF("%s: attribute costume_index is not optional.\n", FUNCNAME);
       return false;
    }

    for (size_t i = 0; i < XV2_LANG_NUM; i++)
    {
        chara_name[i].clear();
        const std::string temp_str = "CHARA_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);

        Utils::ReadParamString(root, temp_str.c_str(), chara_name[i]);

        if (i == XV2_LANG_ENGLISH && chara_name[i].length() == 0)
        {
            DPRINTF("%s: Name for language english is not optional.\n", FUNCNAME);
            return false;
        }
    }

    return true;
}

X2mFile::X2mFile()
{
    Reset();
}

X2mFile::~X2mFile()
{

}

void X2mFile::Reset()
{
    ZipFile::DeleteAll();

    type = X2mType::REPLACER;
    format_version = X2M_CURRENT_VERSION;

    mod_name.clear();
    mod_author.clear();
    mod_version = 1.0f;
    Utils::GetRandomData(mod_guid, sizeof(mod_guid));

    entry_name.clear();
    chara_name.clear();
    chara_name.resize(XV2_LANG_NUM);
    mult_chara_names.clear();
    is_ozaru = false;

    slot_entries.clear();
    cms_entry.id = X2M_INVALID_ID;    
    skill_sets.clear();
    cso_entries.clear();
    psc_entries.clear();
    aur_entries.clear();
    sev_hl_entries.clear();
    sev_ll_entries.clear();
    cml_entries.clear();
    hci_entries.clear();
}

bool X2mFile::Validate()
{
    if (format_version > X2M_CURRENT_VERSION)
    {
        DPRINTF("This x2m requires a newer version of the installer/creator.\n");
        return false;
    }

    if (mod_name.length() == 0)
    {
        DPRINTF("%s: MOD_NAME cannot be empty.\n", FUNCNAME);
        return false;
    }

    if (mod_author.length() == 0)
    {
        DPRINTF("%s: MOD_AUTHOR cannot be empty.\n", FUNCNAME);
        return false;
    }

    if (type == X2mType::NEW_CHARACTER)
    {
        if (entry_name.length() != 3)
        {
            DPRINTF("%s: ENTRY_NAME must be 3 characters long!\n", FUNCNAME);
            return false;
        }

        if (Xenoverse2::IsOriginalChara(entry_name))
        {
            DPRINTF("%s: ENTRY_NAME cannot be an original character for the game.\n", FUNCNAME);
            return false;
        }

        if (!DirExists(entry_name))
        {
            DPRINTF("%s: Directory \"%s\" must exist in the package.\n", FUNCNAME, entry_name.c_str());
            return false;
        }

        if (chara_name[XV2_LANG_ENGLISH].length() == 0)
        {
            DPRINTF("%s: CHARA_NAME_EN cannot be empty.\n", FUNCNAME);
            return false;
        }

        if (!SelPortraitExists())
        {
            DPRINTF("%s: %s doesn't exist.\n", FUNCNAME, X2M_SEL_PORTRAIT);
            return false;
        }

        if (!BtlPortraitExists())
        {
            DPRINTF("%s: %s doesn't exist.\n", FUNCNAME, X2M_BTL_PORTRAIT);
            return false;
        }

        if (slot_entries.size() == 0)
        {
            DPRINTF("%s: There must be at least one SlotEntry.\n", FUNCNAME);
            return false;
        }

        if (slot_entries.size() > XV2_MAX_SUBSLOTS)
        {
            DPRINTF("%s: Number of slots entries cannot be bigger than %d\n", FUNCNAME, XV2_MAX_SUBSLOTS);
            return false;
        }

        for (const X2mSlotEntry &entry: slot_entries)
        {
            if (entry.costume_name[XV2_LANG_ENGLISH].length() == 0)
            {
                DPRINTF("%s: COSTUME_NAME_EN is not optional.\n", FUNCNAME);
                return false;
            }            
        }        

        if (cms_entry.id == X2M_INVALID_ID)
        {
            DPRINTF("%s: There is no cms entry!\n", FUNCNAME);
            return false;
        }

        if (skill_sets.size() == 0)
        {
            DPRINTF("%s: There must be at least one SkillSet.\n", FUNCNAME);
            return false;
        }

        if (skill_sets.size() != slot_entries.size())
        {
            DPRINTF("%s: Number of skill sets must be same as number of slot entries.\n", FUNCNAME);
            return false;
        }

        if (slot_entries.size() > 1)
        {
            struct Combination
            {
                int costume_index;
                int model_preset;

                inline bool operator==(const Combination &rhs) const
                {
                    return (this->costume_index == rhs.costume_index && this->model_preset == rhs.model_preset);
                }

                inline bool operator!=(const Combination &rhs) const
                {
                    return !(*this == rhs);
                }
            };

            std::vector<Combination> combinations;

            for (size_t i = 0; i < slot_entries.size(); i++)
            {
                Combination comb;

                comb.costume_index = slot_entries[i].costume_index;
                comb.model_preset = slot_entries[i].model_preset;

                if (std::find(combinations.begin(), combinations.end(), comb) == combinations.end())
                {
                    combinations.push_back(comb);
                }
                else
                {
                    DPRINTF("%s: Every combination of costume_index+model_preset must be unique.\n", FUNCNAME);
                    return false;
                }
            }

            size_t num_costumes = GetNumCostumes();

            if (HasCso())
            {
                if (cso_entries.size() != num_costumes)
                {
                    DPRINTF("%s: The number of cso entries must be equal to the number of different costume_index.\n", FUNCNAME);
                    return false;
                }
            }

            if (HasPsc())
            {
                if (psc_entries.size() != num_costumes)
                {
                    DPRINTF("%s: The number of psc entries must be equal to the number of different costume_index.\n", FUNCNAME);
                    return false;
                }
            }

            if (HasAur())
            {
                if (aur_entries.size() != num_costumes)
                {
                    DPRINTF("%s: The number of aur entries must be equal to the number of different costume_index.\n", FUNCNAME);
                    return false;
                }
            }

            if (HasSevHL() && HasSevLL())
            {
                DPRINTF("%s: A mod cannot have SevHL and SevLL at the same time.\n", FUNCNAME);
                return false;
            }

            if (HasSevHL())
            {
                std::unordered_set<uint32_t> costumes;

                for (const SevEntryHL &entry : sev_hl_entries)
                {
                    if (entry.copy_char == entry_name)
                    {
                        DPRINTF("%s: Recursive prevention unit, copy_char cannot be the entry_name of the mod.\n", FUNCNAME);
                        return false;
                    }

                    costumes.insert(entry.costume_id);
                }

                if (costumes.size() != sev_hl_entries.size())
                {
                    DPRINTF("%s: Each costume_id in each sev_hl entry must be unique.\n", FUNCNAME);
                    return false;
                }
            }
            else if (HasSevLL())
            {
                std::unordered_set<uint32_t> costumes;

                for (const SevEntryLL &entry : sev_ll_entries)
                {
                    for (const SevEventsLL &sub_entry : entry.sub_entries)
                    {
                        if (!sub_entry.im_source && sub_entry.char_code.length() == 0)
                        {
                            DPRINTF("%s: In SevEventsLL: if im_source is false, char_code cannot be empty.\n", FUNCNAME);
                            return false;
                        }
                    }

                    costumes.insert(entry.costume_id);
                }

                if (costumes.size() != sev_ll_entries.size())
                {
                    DPRINTF("%s: Each costume_id in each sev_ll entry must be unique.\n", FUNCNAME);
                    return false;
                }
            }

            if (HasCml())
            {
                if (cml_entries.size() != num_costumes)
                {
                    DPRINTF("%s: The number of cml entries must be equal to the number of different costume_index.\n", FUNCNAME);
                    return false;
                }
            }

            if (format_version >= X2M_MIN_VERSION_CSS_AUDIO)
            {
                for (size_t i = 0; i < slot_entries.size(); i++)
                {
                    const X2mSlotEntry &slot = slot_entries[i];

                    for (size_t j = 0; j < 2; j++)
                    {
                        if (slot.audio_files[j].length() > 0)
                        {
                            if (!CssAudioFileExists(slot.audio_files[j]))
                            {
                                DPRINTF("%s: The audio file with base name \"%s\" doesn't exist!\n", FUNCNAME, slot.audio_files[j].c_str());
                                return false;
                            }
                        }
                    }
                }
            }
        }
    }
    else if (type == X2mType::REPLACER)
    {
        if (!JungleExists())
        {
            DPRINTF("%s: Jungle must exist for a replacer mod.\n", FUNCNAME);
            return false;
        }
    }

    return true;
}

bool X2mFile::Decompile()
{
    if (!Validate())
        return false;

    TiXmlDocument doc;

    TiXmlDeclaration* decl = new TiXmlDeclaration("1.0", "utf-8", "" );
    doc.LinkEndChild(decl);

    TiXmlElement *root = new TiXmlElement("X2M");
    std::string temp_str;

    if (type == X2mType::REPLACER)
        temp_str = "REPLACER";
    else
        temp_str = "NEW_CHARACTER";

    root->SetAttribute("type", temp_str);
    Utils::WriteParamFloat(root, "X2M_FORMAT_VERSION", format_version);

    Utils::WriteParamString(root, "MOD_NAME", mod_name);
    Utils::WriteParamString(root, "MOD_AUTHOR", mod_author);
    Utils::WriteParamFloat(root, "MOD_VERSION", mod_version);
    Utils::WriteParamGUID(root, "MOD_GUID", mod_guid);

    if (type == X2mType::NEW_CHARACTER)
    {
        Utils::WriteParamString(root, "ENTRY_NAME", entry_name);

        if (is_ozaru)
        {
            Utils::WriteParamString(root, "IS_OZARU", "true");
        }

        for (size_t i = 0; i < XV2_LANG_NUM; i++)
        {
            const std::string &name = chara_name[i];

            if (name.length() == 0)
                continue;

            temp_str = "CHARA_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);
            Utils::WriteParamString(root, temp_str.c_str(), name);
        }

        if (HasMultNames())
        {
            for (const X2mCharaName &name : mult_chara_names)
            {
                if (!name.Decompile(root))
                    return false;
            }
        }

        for (const X2mSlotEntry &entry : slot_entries)
        {
            if (!entry.Decompile(root, format_version >= X2M_MIN_VERSION_CSS_AUDIO))
                return false;
        }

        cms_entry.id = X2M_DUMMY_ID;

        if (!cms_entry.Decompile(root))
            return false;

        for (CusSkillSet &entry : skill_sets)
        {
            entry.char_id = entry.costume_id = X2M_DUMMY_ID;
            entry.skills[9] = X2M_DUMMY_ID16;

            if (!entry.Decompile(root))
                return false;
        }

        for (CsoEntry &entry : cso_entries)
        {
            entry.char_id = entry.costume_id = X2M_DUMMY_ID;

            if (!entry.Decompile(root))
                return false;
        }

        for (PscSpecEntry &entry : psc_entries)
        {
            entry.costume_id = entry.costume_id2 = X2M_DUMMY_ID;

            if (!entry.Decompile(root))
                return false;
        }

        for (AurCharaLink &entry : aur_entries)
        {
            entry.char_id = entry.costume_id = X2M_DUMMY_ID;

            if (!entry.Decompile(root))
                return false;
        }

        for (const SevEntryHL &entry : sev_hl_entries)
        {
            if (!entry.Decompile(root))
                return false;
        }

        for (const SevEntryLL &entry : sev_ll_entries)
        {
            if (!entry.Decompile(root))
                return false;
        }

        for (CmlEntry &entry : cml_entries)
        {
            entry.char_id = entry.costume_id = X2M_DUMMY_ID16;

            if (!entry.Decompile(root))
                return false;
        }

        std::vector<HciEntry> used_hci_entries;

        for (HciEntry &entry : hci_entries)
        {
            entry.char_id = X2M_DUMMY_ID16;

            if (std::find(used_hci_entries.begin(), used_hci_entries.end(), entry) == used_hci_entries.end())
            {
                used_hci_entries.push_back(entry);

                if (!entry.Decompile(root))
                    return false;
            }
        }
    }


    doc.LinkEndChild(root);

    TiXmlPrinter printer;
    doc.Accept(&printer);

    const char *xml_buf = printer.CStr();
    return WriteFile("x2m.xml", xml_buf, strlen(xml_buf));
}

bool X2mFile::Compile()
{
    TiXmlDocument doc;
    char *xml_buf;

    xml_buf = ReadTextFile("x2m.xml");
    if (!xml_buf)
    {
        DPRINTF("%s: x2m.xml not found.\n", FUNCNAME);
        return false;
    }

    doc.Parse(xml_buf);
    delete[] xml_buf;

    if (doc.ErrorId() != 0)
    {
        DPRINTF("%s: cannot parse xml (wrong formatted xml?)\n\n"
                "This is what tinyxml says: %s (%d). Line=%d, col=%d\n", FUNCNAME, doc.ErrorDesc(), doc.ErrorId(), doc.ErrorRow(), doc.ErrorCol());

        return false;
    }

    TiXmlHandle handle(&doc);
    TiXmlElement *root = Utils::FindRoot(&handle, "X2M");
    if (!root)
    {
        DPRINTF("%s: Cannot find \"X2M\" in xml.\n", FUNCNAME);
        return false;
    }

    std::string temp_str;

    if (!Utils::ReadAttrString(root, "type", temp_str))
    {
        DPRINTF("%s: type attribute is not optional.\n", FUNCNAME);
        return false;
    }

    if (temp_str == "REPLACER")
    {
        type = X2mType::REPLACER;
    }
    else if (temp_str == "NEW_CHARACTER")
    {
        type = X2mType::NEW_CHARACTER;
    }
    else
    {
        DPRINTF("%s: Unrecognized type: \"%s\"\n", FUNCNAME, temp_str.c_str());
    }

    if (!Utils::GetParamFloat(root, "X2M_FORMAT_VERSION", &format_version))
        return false;

    if (!Utils::GetParamString(root, "MOD_NAME", mod_name))
        return false;

    if (!Utils::GetParamString(root, "MOD_AUTHOR", mod_author))
        return false;

    if (!Utils::GetParamFloat(root, "MOD_VERSION", &mod_version))
        return false;

    if (!Utils::GetParamGUID(root, "MOD_GUID", mod_guid))
        return false;

    if (type == X2mType::NEW_CHARACTER)
    {
        if (!Utils::GetParamString(root, "ENTRY_NAME", entry_name))
            return false;

        entry_name = Utils::ToUpperCase(entry_name);

        for (int i = 0; i < XV2_LANG_NUM; i++)
        {
            temp_str = "CHARA_NAME_" + Utils::ToUpperCase(xv2_lang_codes[i]);
            Utils::ReadParamString(root, temp_str.c_str(), chara_name[i]);
        }

        std::string ozaru_str;

        if (format_version >= X2M_MIN_VERSION_OZARU && Utils::ReadParamString(root, "IS_OZARU", ozaru_str))
        {
            is_ozaru = (Utils::ToLowerCase(ozaru_str) == "true" || ozaru_str == "1");
        }
        else
        {
            is_ozaru = false; // Not really needed, done at Reset
        }
    }

    for (const TiXmlElement *elem = root->FirstChildElement(); elem; elem = elem->NextSiblingElement())
    {
        const std::string &param_name = elem->ValueStr();

        if (type == X2mType::NEW_CHARACTER)
        {
            if (param_name == "SlotEntry")
            {
                X2mSlotEntry entry;

                if (!entry.Compile(elem, format_version >= X2M_MIN_VERSION_CSS_AUDIO))
                    return false;

                slot_entries.push_back(entry);
            }
            else if (param_name == "X2mCharaName" && format_version >= X2M_MIN_VERSION_MULTIPLE_CHAR_NAMES)
            {
                X2mCharaName name;

                if (!name.Compile(elem))
                    return false;

                mult_chara_names.push_back(name);
            }
            else if (param_name == "Entry")
            {
                if (!cms_entry.Compile(elem))
                {
                    cms_entry.id = X2M_INVALID_ID;
                    return false;
                }
            }
            else if (param_name == "SkillSet")
            {
                CusSkillSet entry;

                if (!entry.Compile(elem))
                    return false;

                skill_sets.push_back(entry);
            }
            else if (param_name == "CsoEntry" && format_version >= X2M_MIN_VERSION_CSO)
            {
                CsoEntry entry;

                if (!entry.Compile(elem))
                    return false;

                cso_entries.push_back(entry);
            }
            else if (param_name == "PscSpecEntry" && format_version >= X2M_MIN_VERSION_PSC)
            {
                PscSpecEntry entry;

                if (!entry.Compile(elem))
                    return false;

                psc_entries.push_back(entry);
            }
            else if (param_name == "CharacLink" && format_version >= X2M_MIN_VERSION_AUR)
            {
                AurCharaLink entry;

                if (!entry.Compile(elem))
                    return false;

                aur_entries.push_back(entry);
            }
            else if (param_name == "SevEntryHL" && format_version >= X2M_MIN_VERSION_SEV)
            {
                SevEntryHL entry;

                if (!entry.Compile(elem))
                    return false;

                sev_hl_entries.push_back(entry);
            }
            else if (param_name == "SevEntryLL" && format_version >= X2M_MIN_VERSION_SEV)
            {
                SevEntryLL entry;

                if (!entry.Compile(elem))
                    return false;

                sev_ll_entries.push_back(entry);
            }
            else if (param_name == "CmlEntry" && format_version >= X2M_MIN_VERSION_CML)
            {
                CmlEntry entry;

                if (!entry.Compile(elem))
                    return false;

                cml_entries.push_back(entry);
            }
            else if (param_name == "HciEntry" && format_version >= X2M_MIN_VERSION_HCI)
            {
                HciEntry entry;

                if (!entry.Compile(elem))
                    return false;

                entry.char_id = X2M_DUMMY_ID16;

                if (std::find(hci_entries.begin(), hci_entries.end(), entry) == hci_entries.end())
                {
                    hci_entries.push_back(entry);
                }
            }
        }
    }

    return Validate();
}

bool X2mFile::Load(const uint8_t *buf, size_t size)
{
    Reset();

    if (!ZipFile::Load(buf, size))
        return false;

    return Compile();
}

uint8_t *X2mFile::Save(size_t *psize)
{
    if (!Decompile())
        return nullptr;

    return ZipFile::Save(psize);
}

bool X2mFile::CharaDirectoryExists() const
{
    if (entry_name.length() == 0)
        return false;

    return DirExists(entry_name);
}

void X2mFile::DeleteCharaDirectory()
{
    if (entry_name.length() == 0)
        return;

    DeleteDir(entry_name);
}

size_t X2mFile::GetNumCostumes() const
{
    std::unordered_set<int32_t> costumes;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        costumes.insert(slot.costume_index);
    }

    return costumes.size();
}

bool X2mFile::CostumeExists(int32_t costume_id) const
{
    std::unordered_set<int32_t> costumes;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        costumes.insert(slot.costume_index);
    }

    return (costumes.find(costume_id) != costumes.end());
}

X2mFile *X2mFile::CreateDummyPackage() const
{
    X2mFile *dummy = new X2mFile();

    dummy->type = type;
    dummy->format_version = format_version;

    dummy->mod_name = mod_name;
    dummy->mod_author = mod_author;
    dummy->mod_version = mod_version;
    memcpy(dummy->mod_guid, mod_guid, sizeof(mod_guid));

    dummy->entry_name = entry_name;
    dummy->chara_name = chara_name;
    dummy->mult_chara_names = mult_chara_names;
    dummy->is_ozaru = is_ozaru;

    dummy->slot_entries = slot_entries;
    dummy->cms_entry = cms_entry;
    dummy->skill_sets = skill_sets;
    dummy->cso_entries = cso_entries;
    dummy->psc_entries = psc_entries;
    dummy->aur_entries = aur_entries;
    dummy->sev_hl_entries = sev_hl_entries;
    dummy->sev_ll_entries = sev_ll_entries;
    dummy->cml_entries = cml_entries;
    dummy->hci_entries = hci_entries;

    size_t num_entries = GetNumEntries();
    const std::string dummy_content = "DUMMY";

    for (size_t i = 0; i < num_entries; i++)
    {
        zip_stat_t zstat;

        if (zip_stat_index(archive, i, 0, &zstat) == -1)
            continue;

        if (!(zstat.valid & ZIP_STAT_SIZE))
            continue;

        if (!(zstat.valid & ZIP_STAT_NAME))
            continue;

        std::string entry_path = Utils::NormalizePath(zstat.name);

        if (entry_path.back() == '/') // Stupid explicit directory entry added by winrar and others
            continue;

        if (!dummy->WriteFile(entry_path, dummy_content.c_str(), dummy_content.length()))
        {
            delete dummy;
            return nullptr;
        }
    }

    return dummy;
}

bool X2mFile::HasDangerousJungle() const
{
    if (!JungleExists())
        return false;

    static const std::vector<std::string> dangerous_paths =
    {
        std::string(X2M_JUNGLE) + "data/system/",
        std::string(X2M_JUNGLE) + "XV2PATCHER/",
        std::string(X2M_JUNGLE) + "bin/"
    };

    static const std::vector<std::string> dangerous_files =
    {
        std::string(X2M_JUNGLE) + "data/ui/iggy/CHARASELE.iggy",
        std::string(X2M_JUNGLE) + "data/ui/texture/CHARA01.emb",
        std::string(X2M_JUNGLE) + "data/ui/CharaImage/chara_image.hci",
        std::string(X2M_JUNGLE) + "data/sound/VOX/Sys/CRT_CS_vox.acb",
        std::string(X2M_JUNGLE) + "data/sound/VOX/Sys/CRT_CS_vox.awb",
        std::string(X2M_JUNGLE) + "data/sound/VOX/Sys/en/CRT_CS_vox.acb",
        std::string(X2M_JUNGLE) + "data/sound/VOX/Sys/en/CRT_CS_vox.awb"
    };

    for (const std::string &path : dangerous_paths)
    {
        if (DirExists(path))
            return true;
    }

    for (const std::string &file : dangerous_files)
    {
        if (FileExists(file))
            return true;
    }

    return false;
}

bool X2mFile::CssAudioFileExists(const std::string &name, bool english) const
{
    if (english)
        return FileExists(std::string(X2M_CSS_AUDIO) + name + "_en.hca");

    return FileExists(std::string(X2M_CSS_AUDIO) + name + "_jp.hca");
}

bool X2mFile::CssAudioFileExists(const std::string &name) const
{
    return (CssAudioFileExists(name, false) || CssAudioFileExists(name, true));
}

HcaFile *X2mFile::LoadCssAudioFile(const std::string &name, bool english, bool fallback)
{
    const std::string jp_path = std::string(X2M_CSS_AUDIO) + name + "_jp.hca";
    const std::string en_path = std::string(X2M_CSS_AUDIO) + name + "_en.hca";

    size_t size;
    uint8_t *buf = ReadFile((english) ? en_path : jp_path, &size);

    if (!buf)
    {
        if (!fallback)
            return nullptr;

        // Load from the other language
        buf = ReadFile((english) ? jp_path : en_path, &size);
        if (!buf)
            return nullptr;
    }

    HcaFile *hca = new HcaFile();
    if (!hca->Load(buf, size))
    {
        delete hca;
        return nullptr;
    }

    hca->SetCiphType(0);
    return hca;
}

size_t X2mFile::GetCssAudioFiles(std::vector<std::string> &files)
{
    files.clear();

    size_t num_entries = GetNumEntries();

    for (size_t i = 0; i < num_entries; i++)
    {
        zip_stat_t zstat;

        if (zip_stat_index(archive, i, 0, &zstat) == -1)
            continue;

        if (!(zstat.valid & ZIP_STAT_NAME))
            continue;

        std::string entry_path = Utils::NormalizePath(zstat.name);

        if (Utils::BeginsWith(entry_path, X2M_CSS_AUDIO, false))
        {
            if (Utils::EndsWith(entry_path, "_jp.hca", false) || Utils::EndsWith(entry_path, "_en.hca", false))
            {
                std::string base = Utils::GetFileNameString(entry_path);
                if (base.length() == strlen("_jp.hca"))
                    continue;

                base = Utils::ToUpperCase(base.substr(0, base.length()-strlen("_jp.hca")));

                if (std::find(files.begin(), files.end(), base) == files.end())
                    files.push_back(base);
            }
        }
    }

    return files.size();
}

static bool css_audio_visitor(const std::string &path, bool, void *param)
{
    X2mFile *pthis = (X2mFile *)param;

    if (Utils::EndsWith(path, "_jp.hca", false) || Utils::EndsWith(path, "_en.hca", false))
    {
        size_t size;
        uint8_t *buf = Utils::ReadFile(path, &size);
        if (buf)
        {
            std::string base = Utils::GetFileNameString(path);
            if (base.length() <= strlen("_jp.hca"))
            {
                delete[] buf;
                return true;
            }

            pthis->WriteFile(X2M_CSS_AUDIO + base, buf, size);
            delete[] buf;
        }
    }

    return true;
}

size_t X2mFile::SetCssAudioDir(const std::string &dir_path)
{
    Utils::VisitDirectory(dir_path, true, false, false, css_audio_visitor, this);

    std::vector<std::string> files;
    return GetCssAudioFiles(files);
}

void X2mFile::DeleteCssAudio()
{
    DeleteDir(X2M_CSS_AUDIO);
}

bool X2mFile::InstallCharaName()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        std::string *name;

        Xenoverse2::RemoveAllCharaName(entry_name, i);

        if (i == XV2_LANG_ENGLISH)
        {
            if (chara_name[i].length() == 0)
                return false;

            name = &chara_name[i];
        }
        else
        {
            if (chara_name[i].length() == 0)
            {
                name = &chara_name[XV2_LANG_ENGLISH];
            }
            else
            {
                name = &chara_name[i];
            }
        }

        if (!Xenoverse2::SetCharaName(entry_name, *name, i))
            return false;

        if (HasMultNames())
        {
            for (size_t j = 0; j < slot_entries.size(); j++)
            {
                std::string specific_name = *name;
                const X2mSlotEntry &slot = slot_entries[j];

                for (size_t k = 0; k < mult_chara_names.size(); k++)
                {
                    const X2mCharaName &m_chara_name = mult_chara_names[k];

                    if (m_chara_name.costume_index == slot.costume_index)
                    {
                        if (i == XV2_LANG_ENGLISH)
                        {
                            if (m_chara_name.chara_name[i].length() == 0)
                                return false;

                            specific_name = m_chara_name.chara_name[i];
                        }
                        else
                        {
                            if (m_chara_name.chara_name[i].length() == 0)
                            {
                                specific_name = m_chara_name.chara_name[XV2_LANG_ENGLISH];
                            }
                            else
                            {
                                specific_name = m_chara_name.chara_name[i];
                            }
                        }

                        break;
                    }
                } // end k loop

                if (!Xenoverse2::SetCharaName(entry_name, specific_name, i, (int)j))
                    return false;
            }
        }
    }

    return true;
}

bool X2mFile::InstallCostumeNames()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (slot_entries.size() == 0)
        return false;

    for (size_t i = 0; i < slot_entries.size(); i++)
    {
        const X2mSlotEntry &entry = slot_entries[i];

        for (int j = 0; j < XV2_LANG_NUM; j++)
        {
            const std::string *name;

            if (j == XV2_LANG_ENGLISH)
            {
                if (entry.costume_name[j].length() == 0)
                    return false;

                name = &entry.costume_name[j];
            }
            else
            {
                if (entry.costume_name[j].length() == 0)
                {
                    name = &entry.costume_name[XV2_LANG_ENGLISH];
                }
                else
                {
                    name = &entry.costume_name[j];
                }
            }

            if (!Xenoverse2::SetCharaCostumeName(entry_name, entry.costume_index, entry.model_preset, *name, j))
                return false;
        }

    }

    return true;
}

void X2mFile::UninstallCssAudio()
{
    for (int i = 0; i < 100; i++)
    {
        const std::string name = std::string("X2M_") + entry_name + "_" + Utils::ToStringAndPad(i, 2);
        Xenoverse2::FreeCssSound(name, false);
        Xenoverse2::FreeCssSound(name, true);
    }
}

bool X2mFile::InstallCssAudio(X2mSlotEntry &entry)
{
    if (format_version < X2M_MIN_VERSION_CSS_AUDIO)
        return true;

    for (int i = 0 ; i < 2; i++)
    {
        if (entry.voices_id_list[i] != (int)X2M_DUMMY_ID)
            continue;

        if (entry.audio_files[i].length() == 0)
        {
            DPRINTF("%s: Internal error, audio file path is empty and voices id list is dummy.\n", FUNCNAME);
            return false;
        }

        auto it = std::find(installed_css_audio.begin(), installed_css_audio.end(), entry.audio_files[i]);
        if (it != installed_css_audio.end())
        {
            size_t index = it - installed_css_audio.begin();
            entry.voices_id_list[i] = installed_css_cue[index];
            continue;
        }

        if (installed_css_audio.size() == 100)
        {
            DPRINTF("%s: Audio overflow, 100 or more audio entries in the mod.\n", FUNCNAME);
            return false;
        }

        if (!CssAudioFileExists(entry.audio_files[i]))
        {
            DPRINTF("%s: Internal error, audio file with base name \"%s\" doesn't exist in package.\n", FUNCNAME, entry.audio_files[i].c_str());
            return false;
        }

        HcaFile *hca_jp = LoadCssAudioFile(entry.audio_files[i], false, true);
        if (!hca_jp)
        {
            DPRINTF("%s: Cannot load css audio sound \"%s\". Most likely reasons is that it is an invalid HCA file.\n", FUNCNAME, entry.audio_files[i].c_str());
            return false;
        }

        HcaFile *hca_en = LoadCssAudioFile(entry.audio_files[i], true, true);
        if (!hca_en)
        {
            DPRINTF("%s: Cannot load css audio sound \"%s\" (2). Most likely reasons is that it is an invalid HCA file.\n", FUNCNAME, entry.audio_files[i].c_str());
            delete hca_jp;
            return false;
        }

        uint32_t cue_id_jp, cue_id_en;
        const std::string cue_name = "X2M_" + entry_name + "_" + Utils::ToStringAndPad((int)installed_css_audio.size(), 2);

        cue_id_jp = Xenoverse2::SetCssSound(cue_name, false, *hca_jp);
        delete hca_jp;

        if (cue_id_jp == (uint32_t)-1)
        {
            DPRINTF("%s: SetCssSound failed when trying to set japanese track.\n", FUNCNAME);
            delete hca_en;
            return false;
        }

        cue_id_en = Xenoverse2::SetCssSound(cue_name, true, *hca_en);
        delete hca_en;

        if (cue_id_en == (uint32_t)-1)
        {
            DPRINTF("%s: SetCssSound failed when trying to set english track.\n", FUNCNAME);
            return false;
        }

        if (cue_id_jp != cue_id_en)
        {
            DPRINTF("%s: English and Japanese CRT_CS_vox.acb are unsynchronized and cannot assign a common cue_id.\n\nPossible reasons:\n"
                    "- One of them was deleted by user\n."
                    "- One of them was edited by the user to add a track and the other one wasn't.\n"
                    "- One of them was set by the user to a previous version and the other wasn't.\n", FUNCNAME);
            return false;
        }

        entry.voices_id_list[i] = cue_id_jp;
        installed_css_audio.push_back(entry.audio_files[i]);
        installed_css_cue.push_back(cue_id_jp);
    }

    return true;
}

bool X2mFile::InstallSlots(bool update)
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    UninstallCssAudio();
    installed_css_audio.clear();
    installed_css_cue.clear();

    if (update)
    {
        std::vector<CharaListSlotEntry *> l_entries;
        chara_list->FindSlotsByCode(std::string("\"") + entry_name + std::string("\""), l_entries);

        if (l_entries.size() != slot_entries.size())
        {
            DPRINTF("%s: Update mode failed because there are different num of slots in the list than in the mod. Try uninstalling the mod and install it again.\n", FUNCNAME);
            return false;
        }

        for (size_t i = 0; i < l_entries.size(); i++)
        {
            X2mSlotEntry &x_entry = slot_entries[i];

            if (!InstallCssAudio(x_entry))
                return false;

            x_entry.CopyTo(*(l_entries[i]), entry_name);
        }

        return true;
    }

    std::vector<CharaListSlot> &chara_slots = chara_list->GetSlots();

    if (chara_slots.size() >= XV2_MAX_SLOTS)
        return false;

    CharaListSlot slot;

    slot.entries.resize(slot_entries.size());

    for (size_t i = 0; i < slot_entries.size(); i++)
    {
        X2mSlotEntry &x_entry = slot_entries[i];
        CharaListSlotEntry &l_entry = slot.entries[i];

        if (!InstallCssAudio(x_entry))
            return false;

        x_entry.CopyTo(l_entry, entry_name);
    }

    chara_slots.push_back(slot);
    return true;
}

bool X2mFile::InstallSlots(const std::vector<size_t> &positions)
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    UninstallCssAudio();
    installed_css_audio.clear();
    installed_css_cue.clear();

    if (positions.size() != slot_entries.size())
        return false;

    std::vector<CharaListSlot> &chara_slots = chara_list->GetSlots();

    // Check first
    for (size_t pos : positions)
    {
        if (pos >= chara_slots.size())
            return false;

        if (chara_slots[pos].entries.size() >= XV2_MAX_SUBSLOTS)
            return false;
    }

    for (size_t i = 0; i < slot_entries.size(); i++)
    {
        X2mSlotEntry &x_entry = slot_entries[i];
        CharaListSlotEntry l_entry;

        if (!InstallCssAudio(x_entry))
            return false;

        x_entry.CopyTo(l_entry, entry_name);
        chara_slots[positions[i]].entries.push_back(l_entry);

        if (chara_slots[positions[i]].entries.size() >= XV2_MAX_SUBSLOTS)
            return false;
    }

    return true;
}

bool X2mFile::InstallCms(bool update)
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    cms_entry.name = entry_name;

    if (update)
    {
        CmsEntryXV2 *entry = dynamic_cast<CmsEntryXV2 *>(game_cms->FindEntryByName(cms_entry.name));

        if (!entry)
        {
            DPRINTF("%s: Update failed because cms entry of mod doesn't exist. A fix may be to uninstall mod and install it again.\n", FUNCNAME);
            return false;
        }

        cms_entry.id = entry->id;
        *entry = cms_entry;
        return true;
    }

    return game_cms->AddEntryXV2(cms_entry, true);
}

bool X2mFile::InstallCus(bool update)
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    if (skill_sets.size() != slot_entries.size())
    {
        DPRINTF("%s: Internal error.\n", FUNCNAME);
        return false;
    }

    if (update)
    {
        std::vector<CusSkillSet *> sets;

        game_cus->FindSkillSetsByCharID(cms_entry.id, sets);

        if (sets.size() != skill_sets.size())
        {
            DPRINTF("%s: Update mode failed because there are different num of skillset in the cus than in the mod (for this char). Try uninstalling the mod and install it again.\n", FUNCNAME);
            return false;
        }

        for (size_t i = 0; i < sets.size(); i++)
        {
            CusSkillSet &x_set = skill_sets[i];

            x_set.char_id = cms_entry.id;
            x_set.costume_id = slot_entries[i].costume_index;
            x_set.skills[9] = slot_entries[i].model_preset;
            *(sets[i]) = x_set;
        }

        return true;
    }

    for (size_t i = 0; i < skill_sets.size(); i++)
    {
        CusSkillSet &x_set = skill_sets[i];

        x_set.char_id = cms_entry.id;
        x_set.costume_id = slot_entries[i].costume_index;
        x_set.skills[9] = slot_entries[i].model_preset;

        if (!game_cus->AddSkillSet(x_set, (i == 0)))
        {
            DPRINTF("%s: AddSkillSet %d  failed. You probably had a dead id in the cus file, caused by manual editing or a replacer mod.\n", FUNCNAME, i);
            return false;
        }
    }

    return true;
}

bool X2mFile::FillCsoEntries()
{
    std::unordered_set<int32_t> costumes;
    size_t i = 0;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        if (costumes.find(slot.costume_index) == costumes.end())
        {
            costumes.insert(slot.costume_index);

            if (i >= cso_entries.size())
                return false;

            cso_entries[i].char_id = cms_entry.id;
            cso_entries[i].costume_id = slot.costume_index;
            i++;
        }
    }

    return (costumes.size() == cso_entries.size());
}

bool X2mFile::InstallCso(bool update)
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    std::vector<CsoEntry *> entries;
    game_cso->FindEntriesByCharID(cms_entry.id, entries);

    if (!HasCso())
    {
        if (entries.size() != 0)
        {
            DPRINTF("%s: Update mode failed because there were cso entry(es) in the cso, but 0 here. Uninstall the mod and reinstall it will fix this.\n", FUNCNAME);
            return false;
        }

        return true; // Yes, true!
    }

    if (cso_entries.size() != GetNumCostumes())
    {
        DPRINTF("%s: Internal error.\n", FUNCNAME);
        return false;
    }

    if (update)
    {
        if (entries.size() == 0)
        {
            goto regular_install;
        }
        else if (entries.size() != cso_entries.size())
        {
            DPRINTF("%s: Update mode failed because there are different num of cso entries in the cso than in the mod (for this char). Try uninstalling the mod and install it again.\n", FUNCNAME);
            return false;
        }

        if (!FillCsoEntries())
        {
            DPRINTF("%s: FillCsoEntries failed.\n", FUNCNAME);
            return false;
        }

        for (size_t i = 0; i < cso_entries.size(); i++)
        {
            const CsoEntry &entry = cso_entries[i];
            *(entries[i]) = entry;
        }

        return true;
    }

regular_install:

    if (!FillCsoEntries())
    {
        DPRINTF("%s: FillCsoEntries failed.\n", FUNCNAME);
        return false;
    }

    for (size_t i = 0; i < cso_entries.size(); i++)
    {
        const CsoEntry &entry = cso_entries[i];

        if (!game_cso->AddEntry(entry, (i == 0)))
            return false;
    }

    return true;
}

bool X2mFile::FillPscEntries()
{
    std::unordered_set<int32_t> costumes;
    size_t i = 0;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        if (costumes.find(slot.costume_index) == costumes.end())
        {
            costumes.insert(slot.costume_index);

            if (i >= psc_entries.size())
                return false;

            psc_entries[i].costume_id = psc_entries[i].costume_id2 = slot.costume_index;
            i++;
        }
    }

    return (costumes.size() == psc_entries.size());
}

bool X2mFile::InstallPsc()
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    PscEntry *entry;
    PscEntry new_entry;
    bool add_entry;

    entry = game_psc->FindEntry(cms_entry.id);

    if (!HasPsc())
    {
        if (entry)
        {
            DPRINTF("%s: Update mode failed because there was a psc entry in the psc, but 0 here. Uninstalling the mod and reinstall it will fix this.\n", FUNCNAME);
            return false;
        }

        return true; // Yes, true!
    }

    if (entry)
    {
        add_entry = false;
    }
    else
    {
        add_entry = true;
        entry = &new_entry;
    }

    entry->char_id = cms_entry.id;
    entry->specs.clear();

    if (!FillPscEntries())
    {
        DPRINTF("%s: FillPscEntries failed.\n", FUNCNAME);
        return false;
    }

    for (const PscSpecEntry &spec : psc_entries)
    {
        entry->specs.push_back(spec);
    }

    if (add_entry)
    {
        game_psc->AddEntry(new_entry);
    }

    return true;
}

bool X2mFile::FillAurEntries()
{
    std::unordered_set<int32_t> costumes;
    size_t i = 0;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        if (costumes.find(slot.costume_index) == costumes.end())
        {
            costumes.insert(slot.costume_index);

            if (i >= aur_entries.size())
                return false;

            aur_entries[i].char_id = cms_entry.id;
            aur_entries[i].costume_id = slot.costume_index;
            i++;
        }
    }

    return (costumes.size() == aur_entries.size());
}

bool X2mFile::InstallAur(bool update)
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    std::vector<AurCharaLink *> entries;
    game_aur->FindCharaLinks(cms_entry.id, entries);

    if (!HasAur())
    {
        if (entries.size() != 0)
        {
            DPRINTF("%s: Update mode failed because there were aur entry(es) in the aur, but 0 here. Uninstalling the mod and reinstall it will fix this.\n", FUNCNAME);
            return false;
        }

        return true; // Yes, true!
    }

    if (aur_entries.size() != GetNumCostumes())
    {
        DPRINTF("%s: Internal error.\n", FUNCNAME);
        return false;
    }

    if (update)
    {
        if (entries.size() == 0)
        {
            goto regular_install;
        }
        else if (entries.size() != aur_entries.size())
        {
            DPRINTF("%s: Update mode failed because there are different num of aur entries in the aur than in the mod (for this char). Try uninstalling the mod and install it again.\n", FUNCNAME);
            return false;
        }

        if (!FillAurEntries())
        {
            DPRINTF("%s: FillAurEntries failed.\n", FUNCNAME);
            return false;
        }

        for (size_t i = 0; i < aur_entries.size(); i++)
        {
            const AurCharaLink &entry = aur_entries[i];
            *(entries[i]) = entry;
        }

        return true;
    }

regular_install:

    if (!FillAurEntries())
    {
        DPRINTF("%s: FillAurEntries failed.\n", FUNCNAME);
        return false;
    }

    for (size_t i = 0; i < aur_entries.size(); i++)
    {
        const AurCharaLink &entry = aur_entries[i];

        if (!game_aur->AddCharaLink(entry, (i == 0)))
            return false;
    }

    return true;
}

void X2mFile::RebuildSevHL(const std::vector<SevTableEntryHL> &hl_table)
{
    if (hl_table.size() == 0)
        return;

    for (const SevTableEntryHL &hl_entry : hl_table)
    {
        CmsEntry *cms_entry = game_cms->FindEntryByName(hl_entry.char_id);
        if (!cms_entry)
            continue;

        game_sev->RemoveAllReferencesToChar(cms_entry->id);
    }

    std::vector<SevEntry> &sev_entries = game_sev->GetEntries();

    for (size_t i = 0; i < sev_entries.size(); i++)
    {
        const SevEntry &sev_entry = sev_entries[i];
        CmsEntry *copy_cms_entry = game_cms->FindEntryByID(sev_entry.char_id);

        if (!copy_cms_entry)
            continue;

        for (const SevTableEntryHL &hl_entry : hl_table)
        {
            if (hl_entry.copy_char == copy_cms_entry->name && hl_entry.copy_costume == sev_entry.costume_id)
            {
                CmsEntry *cms_entry = game_cms->FindEntryByName(hl_entry.char_id);
                if (!cms_entry)
                    continue;

                SevEntry new_sev_entry = sev_entry;
                new_sev_entry.char_id = cms_entry->id;
                new_sev_entry.costume_id = hl_entry.costume_id;

                sev_entries.push_back(new_sev_entry);
            }
        }
    }

    for (SevEntry &sev_entry : sev_entries)
    {
        for (size_t i = 0; i < sev_entry.chars_events.size(); i++)
        {
            const SevCharEvents &cevents = sev_entry.chars_events[i];
            CmsEntry *copy_cms_entry = game_cms->FindEntryByID(cevents.char_id);

            if (!copy_cms_entry)
                continue;

            for (const SevTableEntryHL &hl_entry : hl_table)
            {
                if (hl_entry.copy_char == copy_cms_entry->name && hl_entry.copy_costume == cevents.costume_id)
                {
                    CmsEntry *cms_entry = game_cms->FindEntryByName(hl_entry.char_id);
                    if (!cms_entry)
                        continue;

                    SevCharEvents new_cevents = cevents;
                    new_cevents.char_id = cms_entry->id;
                    new_cevents.costume_id = hl_entry.costume_id;

                    sev_entry.chars_events.push_back(new_cevents);
                }
            }
        }
    }
}

void X2mFile::RebuildSevLL(const std::vector<SevTableEntryLL> &ll_table)
{
    if (ll_table.size() == 0)
        return;

    for (const SevTableEntryLL &ll_entry : ll_table)
    {
        CmsEntry *cms_entry = game_cms->FindEntryByName(ll_entry.char_id);
        if (!cms_entry)
            continue;

        game_sev->RemoveAllReferencesToChar(cms_entry->id);
    }

    std::vector<SevEntry> &sev_entries = game_sev->GetEntries();

    // Let's go first for the we are source ones.
    for (const SevTableEntryLL &ll_entry : ll_table)
    {
        CmsEntry *cms_entry = game_cms->FindEntryByName(ll_entry.char_id);
        if (!cms_entry)
            continue;

        SevEntry new_entry;

        new_entry.char_id = cms_entry->id;
        new_entry.costume_id = ll_entry.costume_id;

        for (const SevEventsLL &sub_entry : ll_entry.sub_entries)
        {
            if (!sub_entry.im_source)
                continue;

            CmsEntry *other_cms_entry = nullptr;

            if (sub_entry.char_code.length() != 0)
            {
                other_cms_entry = game_cms->FindEntryByName(sub_entry.char_code);
                if (!other_cms_entry)
                    continue;
            }

            SevCharEvents cevents;

            cevents.char_id = (other_cms_entry) ? other_cms_entry->id : 0xFFFFFFFF;
            cevents.costume_id = sub_entry.costume_id;
            cevents.events = sub_entry.events;

            new_entry.chars_events.push_back(cevents);
        }

        if (new_entry.chars_events.size() != 0)
        {
            sev_entries.push_back(new_entry);
        }
    }

    // Let's go now we the destinations
    for (const SevTableEntryLL &ll_entry : ll_table)
    {
        CmsEntry *cms_entry = game_cms->FindEntryByName(ll_entry.char_id);
        if (!cms_entry)
            continue;

        for (const SevEventsLL &sub_entry : ll_entry.sub_entries)
        {
            if (sub_entry.im_source)
                continue;

            CmsEntry *other_cms_entry = game_cms->FindEntryByName(sub_entry.char_code);
            if (!other_cms_entry)
                continue;

            SevEntry *other_entry = game_sev->FindEntry(other_cms_entry->id, sub_entry.costume_id);
            if (!other_entry)
                continue;

            SevCharEvents new_cevents;

            new_cevents.char_id = cms_entry->id;
            new_cevents.costume_id = ll_entry.costume_id;
            new_cevents.events = sub_entry.events;

            other_entry->chars_events.push_back(new_cevents);
        }
    }
}

bool X2mFile::InstallSev(std::vector<SevTableEntryHL> &hl_table, std::vector<SevTableEntryLL> &ll_table)
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    // Remove ourselves from tables if found (for updating, since costumes or copy data may not match)
    for (size_t i = 0; i < hl_table.size(); i++)
    {
        const SevTableEntryHL &entry_hl = hl_table[i];

        if (entry_hl.char_id == entry_name)
        {
            hl_table.erase(hl_table.begin()+i);
            i--;
        }
    }

    for (size_t i = 0; i < ll_table.size(); i++)
    {
        const SevTableEntryLL &entry_ll = ll_table[i];

        if (entry_ll.char_id == entry_name)
        {
            ll_table.erase(ll_table.begin()+i);
            i--;
        }
    }

    if (HasSevHL())
    {
        for (const SevEntryHL &entry : sev_hl_entries)
        {
            SevTableEntryHL new_hl_entry;

            new_hl_entry.char_id = entry_name;
            new_hl_entry.costume_id = entry.costume_id;
            new_hl_entry.copy_char = entry.copy_char;
            new_hl_entry.copy_costume = entry.copy_costume;

            hl_table.push_back(new_hl_entry);
        }
    }
    else if (HasSevLL())
    {
        for (const SevEntryLL &entry : sev_ll_entries)
        {
            SevTableEntryLL new_ll_entry;

            new_ll_entry.char_id = entry_name;
            new_ll_entry.costume_id = entry.costume_id;
            new_ll_entry.sub_entries = entry.sub_entries;

            ll_table.push_back(new_ll_entry);
        }
    }

    game_sev->RemoveAllReferencesToChar(cms_entry.id); // Needed in case update doesn't have sev

    // Tables must always rebuild, regardless of if the mod has sev or not
    RebuildSevLL(ll_table); // LL table must go first
    RebuildSevHL(hl_table);

    return true;
}

bool X2mFile::FillCmlEntries()
{
    std::unordered_set<int32_t> costumes;
    size_t i = 0;

    for (const X2mSlotEntry &slot : slot_entries)
    {
        if (costumes.find(slot.costume_index) == costumes.end())
        {
            costumes.insert(slot.costume_index);

            if (i >= cml_entries.size())
                return false;

            cml_entries[i].char_id = cms_entry.id;
            cml_entries[i].costume_id = slot.costume_index;
            i++;
        }
    }

    return (costumes.size() == cml_entries.size());
}

bool X2mFile::InstallCml(bool update)
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    std::vector<CmlEntry *> entries;
    game_cml->FindEntriesByCharID(cms_entry.id, entries);

    if (!HasCml())
    {
        if (entries.size() != 0)
        {
            DPRINTF("%s: Update mode failed because there were cml entry(es) in the cml, but 0 here. Uninstalling the mod and reinstall it will fix this.\n", FUNCNAME);
            return false;
        }

        return true; // Yes, true!
    }

    if (cml_entries.size() != GetNumCostumes())
    {
        DPRINTF("%s: Internal error.\n", FUNCNAME);
        return false;
    }

    if (update)
    {
        if (entries.size() == 0)
        {
            goto regular_install;
        }
        else if (entries.size() != cml_entries.size())
        {
            DPRINTF("%s: Update mode failed because there are different num of cml entries in the cml than in the mod (for this char). Try uninstalling the mod and install it again.\n", FUNCNAME);
            return false;
        }

        if (!FillCmlEntries())
        {
            DPRINTF("%s: FillCmlEntries failed.\n", FUNCNAME);
            return false;
        }

        for (size_t i = 0; i < cml_entries.size(); i++)
        {
            const CmlEntry &entry = cml_entries[i];
            *(entries[i]) = entry;
        }

        return true;
    }

regular_install:

    if (!FillCmlEntries())
    {
        DPRINTF("%s: FillCmlEntries failed.\n", FUNCNAME);
        return false;
    }

    for (size_t i = 0; i < cml_entries.size(); i++)
    {
        const CmlEntry &entry = cml_entries[i];

        if (!game_cml->AddEntry(entry, (i == 0)))
            return false;
    }

    return true;
}

bool X2mFile::InstallHci()
{
    if (cms_entry.id == X2M_DUMMY_ID || cms_entry.id == X2M_INVALID_ID)
    {
        DPRINTF("%s: Internal error. InstallCms must be called before this!!!\n", FUNCNAME);
        return false;
    }

    game_hci->RemoveEntries(cms_entry.id); // Remove from previous version (if any)

    if (!HasHci())
        return true; // Yes, true

    for (HciEntry &entry : hci_entries)
    {
        entry.char_id = cms_entry.id;

        if (!game_hci->AddEntry(entry, false))
            return false;
    }

    return true;
}

bool X2mFile::InstallSelPortrait()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (!SelPortraitExists())
        return false;

    size_t size;
    uint8_t *buf = ReadFile(X2M_SEL_PORTRAIT, &size);
    if (!buf)
        return false;

    std::string name = entry_name + "_000.dds";
    bool ret = Xenoverse2::SetSelPortrait(name, buf, size);

    delete[] buf;
    return ret;
}

bool X2mFile::InstallPreBaked()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (is_ozaru)
    {
        game_prebaked->AddOzaru(entry_name);
    }
    else
    {
        game_prebaked->RemoveOzaru(entry_name);
    }

    return true;
}

bool X2mFile::InstallBtlPortrait()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (!BtlPortraitExists())
        return false;

    size_t size;
    uint8_t *buf = ReadFile(X2M_BTL_PORTRAIT, &size);
    if (!buf)
        return false;

    std::string emb_path = std::string("data/ui/texture/") + entry_name + ".emb";
    bool ret = xv2fs->WriteFile(emb_path, buf, size);

    delete[] buf;
    return ret;
}

bool X2mFile::InstallDir(const std::string &in_path, const std::string &out_path)
{
    size_t num_entries = GetNumEntries();

    std::string proper_in_path = in_path;
    std::string proper_out_path = out_path;

    if (proper_in_path.back() != '/')
        proper_in_path += '/';

    if (proper_out_path.length() != 0 && proper_out_path.back() != '/')
        proper_out_path += '/';

    for (size_t i = 0; i < num_entries; i++)
    {
        zip_stat_t zstat;

        if (zip_stat_index(archive, i, 0, &zstat) == -1)
            continue;

        if (!(zstat.valid & ZIP_STAT_NAME))
            continue;

        if (!(zstat.valid & ZIP_STAT_SIZE))
            continue;

        std::string entry_path = Utils::NormalizePath(zstat.name);

        if (entry_path.back() == '/')
            continue;

        if (!Utils::BeginsWith(entry_path, proper_in_path, false))
            continue;

        size_t size;
        uint8_t *buf = ReadFile(zstat.name, &size);
        if (!buf)
            return false;

        std::string out_file = proper_out_path + entry_path.substr(proper_in_path.length());
        bool ret = xv2fs->WriteFile(out_file, buf, size);
        delete[] buf;

        if (!ret)
            return false;
    }

    return true;
}

bool X2mFile::InstallCharaFiles()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (!CharaDirectoryExists())
        return false;

    return InstallDir(entry_name, "data/chara/" + entry_name);
}

bool X2mFile::InstallJungle()
{
    if (!JungleExists())
        return false;

    return InstallDir(X2M_JUNGLE, "");
}

bool X2mFile::UninstallCharaName()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    for (int i = 0; i < XV2_LANG_NUM; i++)
    {
        if (!Xenoverse2::RemoveAllCharaName(entry_name, i))
            return false;
    }

    return true;
}

bool X2mFile::UninstallCostumeNames()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (slot_entries.size() == 0)
        return false;

    for (size_t i = 0; i < slot_entries.size(); i++)
    {
        const X2mSlotEntry &slot = slot_entries[i];

        for (int j = 0; j < XV2_LANG_NUM; j++)
        {
            if (!Xenoverse2::RemoveCharaCostumeName(entry_name, slot.costume_index, slot.model_preset, j))
                return false;
        }
    }

    return true;
}

bool X2mFile::UninstallSlots()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (slot_entries.size() == 0)
        return false;

    UninstallCssAudio();
    chara_list->RemoveSlots(std::string("\"") + entry_name + std::string("\""));
    return true;
}

bool X2mFile::UninstallCms()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    cms_entry.name = entry_name;

    CmsEntryXV2 *entry = dynamic_cast<CmsEntryXV2 *>(game_cms->FindEntryByName(cms_entry.name));
    if (entry)
    {
        cms_entry.id = entry->id;
    }
    else
    {
        return true;
    }

    return game_cms->RemoveEntry(entry_name);
}

bool X2mFile::UninstallCus()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_cus->RemoveSkillSets(cms_entry.id);
    return true;
}

bool X2mFile::UninstallCso()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_cso->RemoveEntries(cms_entry.id);
    return true;
}

bool X2mFile::UninstallPsc()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_psc->RemoveEntry(cms_entry.id);
    return true;
}

bool X2mFile::UninstallAur()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_aur->RemoveCharaLinks(cms_entry.id);
    return true;
}

bool X2mFile::UninstallSev(std::vector<SevTableEntryHL> &hl_table, std::vector<SevTableEntryLL> &ll_table)
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    // Remove ourselves from tables if found
    for (size_t i = 0; i < hl_table.size(); i++)
    {
        const SevTableEntryHL &entry_hl = hl_table[i];

        if (entry_hl.char_id == entry_name)
        {
            hl_table.erase(hl_table.begin()+i);
            i--;
        }
    }

    for (size_t i = 0; i < ll_table.size(); i++)
    {
        const SevTableEntryLL &entry_ll = ll_table[i];

        if (entry_ll.char_id == entry_name)
        {
            ll_table.erase(ll_table.begin()+i);
            i--;
        }
    }

    game_sev->RemoveAllReferencesToChar(cms_entry.id);
    RebuildSevLL(ll_table); // LL must go first
    RebuildSevHL(hl_table);

    return true;
}

bool X2mFile::UninstallCml()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_cml->RemoveEntries(cms_entry.id);
    return true;
}

bool X2mFile::UninstallHci()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (cms_entry.id == X2M_INVALID_ID)
        return false;

    if (cms_entry.id == X2M_DUMMY_ID)
        return true; // Yes, true

    game_hci->RemoveEntries(cms_entry.id);
    return true;
}

bool X2mFile::UninstallSelPortrait()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    std::string name = entry_name + "_000.dds";
    return Xenoverse2::RemoveSelPortrait(name);
}

bool X2mFile::UninstallPreBaked()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    game_prebaked->RemoveOzaru(entry_name);
    return true;
}

bool X2mFile::UninstallBtlPortrait()
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (!BtlPortraitExists())
        return false;

    std::string emb_path = std::string("data/ui/texture/") + entry_name + ".emb";    
    return xv2fs->RemoveFile(emb_path);
}

bool X2mFile::UninstallDir(const std::string &in_path, const std::string &out_path)
{
    size_t num_entries = GetNumEntries();

    std::string proper_in_path = in_path;
    std::string proper_out_path = out_path;

    if (proper_in_path.back() != '/')
        proper_in_path += '/';

    if (proper_out_path.length() != 0 && proper_out_path.back() != '/')
        proper_out_path += '/';

    for (size_t i = 0; i < num_entries; i++)
    {
        zip_stat_t zstat;

        if (zip_stat_index(archive, i, 0, &zstat) == -1)
            continue;

        if (!(zstat.valid & ZIP_STAT_NAME))
            continue;

        if (!(zstat.valid & ZIP_STAT_SIZE))
            continue;

        std::string entry_path = Utils::NormalizePath(zstat.name);

        if (entry_path.back() == '/')
            continue;

        if (!Utils::BeginsWith(entry_path, proper_in_path, false))
            continue;

        std::string out_file = proper_out_path + entry_path.substr(proper_in_path.length());
        bool ret = xv2fs->RemoveFile(out_file);

        if (!ret)
            return false;
    }

    return true;
}

bool X2mFile::UninstallCharaFiles(bool remove_empty)
{
    if (type != X2mType::NEW_CHARACTER)
        return false;

    if (!CharaDirectoryExists())
        return false;

    const std::string path = "data/chara/" + entry_name;

    if (!UninstallDir(entry_name, path))
        return false;

    if (remove_empty)
        xv2fs->RemoveEmptyDir(path);

    return true;
}

bool X2mFile::UninstallJungle()
{
    if (!JungleExists())
        return false;

    return UninstallDir(X2M_JUNGLE, "");
}


