#ifndef __X2MFILE_H__
#define __X2MFILE_H__

#include "ZipFile.h"
#include "Xenoverse2.h"

#define X2M_SEL_PORTRAIT    "UI/SEL.DDS"
#define X2M_BTL_PORTRAIT    "UI/BTL.EMB"
#define X2M_JUNGLE          "JUNGLE/"
#define X2M_CSS_AUDIO       "CSS_AUDIO/"

#define X2M_INVALID_ID  0xCACACACA
#define X2M_DUMMY_ID   0xBACABACA
#define X2M_DUMMY_ID16  0xBACA

enum class X2mType
{
    REPLACER,
    NEW_CHARACTER
};

struct X2mSlotEntry
{
    int32_t costume_index;
    int32_t model_preset;
    bool flag_gk2;
    int32_t voices_id_list[2];
    //
    std::vector<std::string> costume_name;

    // New field 0.5
    std::string audio_files[2];

    X2mSlotEntry()
    {
        costume_index = 0;
        model_preset = 0;
        flag_gk2 = false;
        voices_id_list[0] = voices_id_list[1] = -1;

        costume_name.resize(XV2_LANG_NUM);
        audio_files[0].clear();
        audio_files[1].clear();
    }    

    void CopyFrom(const CharaListSlotEntry &entry, bool name);
    void CopyTo(CharaListSlotEntry &entry, const std::string &code) const;

    TiXmlElement *Decompile(TiXmlNode *root, bool new_format) const;
    bool Compile(const TiXmlElement *root, bool new_format);
};

struct SevEntryHL
{
    uint32_t costume_id;
    std::string copy_char;
    uint32_t copy_costume;

    SevEntryHL()
    {
        costume_id = copy_costume = 0xFFFFFFFF;        
    }

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevEventsLL
{
    bool im_source;

    std::string char_code; // always the char code of the other char
    uint32_t costume_id; // Always the costume id of the other char
    std::vector<SevEvent> events;

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevEntryLL
{
    uint32_t costume_id; // Always our costume id

    std::vector<SevEventsLL> sub_entries;

    SevEntryLL(bool default_entries=false)
    {
        costume_id = 0xFFFFFFFF;

        if (!default_entries)
            return;

        sub_entries.resize(1);

        SevEventsLL &sub_entry = sub_entries.front();
        sub_entry.im_source = true;
        sub_entry.costume_id = 0;
        sub_entry.events.resize(1);

        SevEvent &event = sub_entry.events.front();
        event.entries.resize(1);
    }

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

struct SevTableEntryHL
{
    std::string char_id;
    uint32_t costume_id;

    std::string copy_char;
    uint32_t copy_costume;
};

struct SevTableEntryLL
{
    std::string char_id; // Always our char_id
    uint32_t costume_id; // Always our costume id

    std::vector<SevEventsLL> sub_entries;
};

struct X2mCharaName
{
    uint32_t costume_index;
    std::vector<std::string> chara_name;

    X2mCharaName()
    {
        chara_name.resize(XV2_LANG_NUM);
    }

    TiXmlElement *Decompile(TiXmlNode *root) const;
    bool Compile(const TiXmlElement *root);
};

class X2mFile : public ZipFile
{
private:

    // Require for all types
    X2mType type;
    float format_version;

    std::string mod_name;
    std::string mod_author;
    float mod_version;
    uint8_t mod_guid[16];

    // For new character
    std::string entry_name;
    std::vector<std::string> chara_name;
    std::vector<X2mCharaName> mult_chara_names;

    bool is_ozaru;

    std::vector<X2mSlotEntry> slot_entries;
    CmsEntryXV2 cms_entry;
    std::vector<CusSkillSet> skill_sets;
    std::vector<CsoEntry> cso_entries;
    std::vector<PscSpecEntry> psc_entries;
    std::vector<AurCharaLink> aur_entries;
    std::vector<SevEntryHL> sev_hl_entries;
    std::vector<SevEntryLL> sev_ll_entries;
    std::vector<CmlEntry> cml_entries;
    std::vector<HciEntry> hci_entries;

    // ///////////////////////////////
    // temp variables, do not copy to DUMMY x2d
    std::vector<std::string> installed_css_audio;
    std::vector<uint32_t> installed_css_cue;

    // ///////////////////////////////
    bool Validate();

    bool Decompile();
    bool Compile();

    bool FillCsoEntries();
    bool FillPscEntries();
    bool FillAurEntries();
    bool FillCmlEntries();

    bool InstallDir(const std::string &in_path, const std::string &out_path);
    bool UninstallDir(const std::string &in_path, const std::string &out_path);

    void RebuildSevHL(const std::vector<SevTableEntryHL> &hl_table);
    void RebuildSevLL(const std::vector<SevTableEntryLL> &ll_table);

    void UninstallCssAudio();
    bool InstallCssAudio(X2mSlotEntry &entry);

protected:

    void Reset();

public:

    const float X2M_CURRENT_VERSION = 5.0f;

    const float X2M_MIN_VERSION_CSO = 2.0f;
    const float X2M_MIN_VERSION_PSC = 3.0f;
    const float X2M_MIN_VERSION_AUR = 3.0f;
    const float X2M_MIN_VERSION_SEV = 4.0f;
    const float X2M_MIN_VERSION_CML = 4.0f;
    const float X2M_MIN_VERSION_HCI = 4.0f;
    const float X2M_MIN_VERSION_OZARU = 4.0f;
    const float X2M_MIN_VERSION_CSS_AUDIO = 5.0f;
    const float X2M_MIN_VERSION_MULTIPLE_CHAR_NAMES = 5.0f;

    X2mFile();
    virtual ~X2mFile();

    virtual bool Load(const uint8_t *buf, size_t size) override;
    virtual uint8_t *Save(size_t *psize) override;

    inline X2mType GetType() const { return type; }
    inline void SetType(X2mType type) { this->type = type; }

    inline float GetFormatVersion() const { return format_version; }
    inline void SetFormatVersion(float format_version){ this->format_version = format_version; }

    inline std::string GetModName() const { return mod_name; }
    inline void SetModName(const std::string &mod_name) { this->mod_name = mod_name; }

    inline std::string GetModAuthor() const { return mod_author; }
    inline void SetModAuthor(const std::string &mod_author) { this->mod_author = mod_author; }

    inline float GetModVersion() const { return mod_version; }
    inline void SetModVersion(float mod_version) { this->mod_version = mod_version; }

    inline std::string GetModGuid() const { return Utils::GUID2String(mod_guid); }
    inline bool SetModGuid(const std::string &mod_guid)
    {
        return Utils::String2GUID(this->mod_guid, mod_guid);
    }

    inline std::string GetEntryName() const { return entry_name; }
    inline void SetEntryName(const std::string &entry_name) { this->entry_name = Utils::ToUpperCase(entry_name); }

    inline std::string GetCharaName(int lang) const
    {
        if (lang >= XV2_LANG_NUM)
            return std::string();

        return chara_name[lang];
    }

    inline bool SetCharaName(const std::string &name, int lang)
    {
        if (lang >= XV2_LANG_NUM)
            return false;

        chara_name[lang] = name;
        return true;
    }

    inline size_t GetNumMultNames() const { return mult_chara_names.size(); }
    inline const X2mCharaName &GetMultName(size_t idx) const { return mult_chara_names[idx]; }
    inline X2mCharaName &GetMultName(size_t idx) { return mult_chara_names[idx]; }

    inline size_t AddMultName(const X2mCharaName &entry) { mult_chara_names.push_back(entry); return (mult_chara_names.size()-1); }
    inline void RemoveMultName(size_t idx) { mult_chara_names.erase(mult_chara_names.begin()+idx); }

    inline bool HasMultNames() const { return (format_version >= X2M_MIN_VERSION_MULTIPLE_CHAR_NAMES && GetNumMultNames() > 0); }

    inline bool IsOzaru() const { return is_ozaru; }
    inline void SetOzaru(bool ozaru_flag) { is_ozaru = ozaru_flag; }

    bool CharaDirectoryExists() const;
    void DeleteCharaDirectory();

    inline bool SelPortraitExists() const { return FileExists(X2M_SEL_PORTRAIT); }
    inline bool BtlPortraitExists() const { return FileExists(X2M_BTL_PORTRAIT); }

    inline bool JungleExists() const { return DirExists(X2M_JUNGLE); }
    void DeleteJungle() { DeleteDir(X2M_JUNGLE); }

    inline size_t GetNumSlotEntries() const { return slot_entries.size(); }
    inline const X2mSlotEntry &GetSlotEntry(size_t idx) const { return slot_entries[idx]; }
    inline X2mSlotEntry &GetSlotEntry(size_t idx) { return slot_entries[idx]; }

    size_t GetNumCostumes() const;
    bool CostumeExists(int32_t costume_id) const;

    inline size_t AddSlotEntry(const X2mSlotEntry &entry) { slot_entries.push_back(entry); return (slot_entries.size()-1); }
    inline void RemoveSlotEntry(size_t idx) { slot_entries.erase(slot_entries.begin()+idx); }

    inline bool HasValidCmsEntry() const { return (cms_entry.id != X2M_INVALID_ID); }
    inline const CmsEntryXV2 &GetCmsEntry() const { return cms_entry; }
    inline CmsEntryXV2 &GetCmsEntry() { return cms_entry; }

    inline size_t GetNumSkillSets()  { return skill_sets.size(); }
    inline const CusSkillSet &GetSkillSet(size_t idx) const { return skill_sets[idx]; }
    inline CusSkillSet &GetSkillSet(size_t idx) { return skill_sets[idx]; }

    inline size_t AddSkillSet(const CusSkillSet &entry) { skill_sets.push_back(entry); return (skill_sets.size()-1); }
    inline void RemoveSkillSet(size_t idx) { skill_sets.erase(skill_sets.begin()+idx); }

    X2mFile *CreateDummyPackage() const;
    bool HasDangerousJungle() const;

    inline size_t GetNumCsoEntries() const { return cso_entries.size(); }
    inline const CsoEntry &GetCsoEntry(size_t idx) const { return cso_entries[idx]; }
    inline CsoEntry &GetCsoEntry(size_t idx) { return cso_entries[idx]; }

    inline size_t AddCsoEntry(const CsoEntry &entry) { cso_entries.push_back(entry); return (cso_entries.size()-1); }
    inline void RemoveCsoEntry(size_t idx) { cso_entries.erase(cso_entries.begin()+idx); }

    inline bool HasCso() const { return (format_version >= X2M_MIN_VERSION_CSO && GetNumCsoEntries() > 0); }

    inline size_t GetNumPscEntries() const { return psc_entries.size(); }
    inline const PscSpecEntry &GetPscEntry(size_t idx) const { return psc_entries[idx]; }
    inline PscSpecEntry &GetPscEntry(size_t idx) { return psc_entries[idx]; }

    inline size_t AddPscEntry(const PscSpecEntry &entry) { psc_entries.push_back(entry); return (psc_entries.size()-1); }
    inline void RemovePscEntry(size_t idx) { psc_entries.erase(psc_entries.begin()+idx); }

    inline bool HasPsc() const { return (format_version >= X2M_MIN_VERSION_PSC && GetNumPscEntries() > 0); }

    inline size_t GetNumAurEntries() const { return aur_entries.size(); }
    inline const AurCharaLink &GetAurEntry(size_t idx) const { return aur_entries[idx]; }
    inline AurCharaLink &GetAurEntry(size_t idx) { return aur_entries[idx]; }

    inline size_t AddAurEntry(const AurCharaLink &entry) { aur_entries.push_back(entry); return (aur_entries.size()-1); }
    inline void RemoveAurEntry(size_t idx) { aur_entries.erase(aur_entries.begin()+idx); }

    inline bool HasAur() const { return (format_version >= X2M_MIN_VERSION_AUR && GetNumAurEntries() > 0); }

    inline size_t GetNumSevHLEntries() const { return sev_hl_entries.size(); }
    inline const SevEntryHL &GetSevHLEntry(size_t idx) const { return sev_hl_entries[idx]; }
    inline SevEntryHL &GetSevHLEntry(size_t idx) { return sev_hl_entries[idx]; }

    inline size_t AddSevHLEntry(const SevEntryHL &entry) { sev_hl_entries.push_back(entry); return (sev_hl_entries.size()-1); }
    inline void RemoveSevHLEntry(size_t idx) { sev_hl_entries.erase(sev_hl_entries.begin()+idx); }
    inline void RemoveAllSevHL() { sev_hl_entries.clear(); }

    inline bool HasSevHL() const { return (format_version >= X2M_MIN_VERSION_SEV && GetNumSevHLEntries() > 0); }

    inline size_t GetNumSevLLEntries() const { return sev_ll_entries.size(); }
    inline const SevEntryLL &GetSevLLEntry(size_t idx) const { return sev_ll_entries[idx]; }
    inline SevEntryLL &GetSevLLEntry(size_t idx) { return sev_ll_entries[idx]; }

    inline size_t AddSevLLEntry(const SevEntryLL &entry) { sev_ll_entries.push_back(entry); return (sev_ll_entries.size()-1); }
    inline void RemoveSevLLEntry(size_t idx) { sev_ll_entries.erase(sev_ll_entries.begin()+idx); }
    inline void RemoveAllSevLL() { sev_ll_entries.clear(); }

    inline bool HasSevLL() const { return (format_version >= X2M_MIN_VERSION_SEV && GetNumSevLLEntries() > 0); }

    inline bool HasSev() const { return (HasSevHL() || HasSevLL()); }

    inline size_t GetNumCmlEntries() const { return cml_entries.size(); }
    inline const CmlEntry &GetCmlEntry(size_t idx) const { return cml_entries[idx]; }
    inline CmlEntry &GetCmlEntry(size_t idx) { return cml_entries[idx]; }

    inline size_t AddCmlEntry(const CmlEntry &entry) { cml_entries.push_back(entry); return (cml_entries.size()-1); }
    inline void RemoveCmlEntry(size_t idx) { cml_entries.erase(cml_entries.begin()+idx); }

    inline bool HasCml() const { return (format_version >= X2M_MIN_VERSION_CML && GetNumCmlEntries() > 0); }

    inline size_t GetNumHciEntries() const { return hci_entries.size(); }
    inline const HciEntry &GetHciEntry(size_t idx) const { return hci_entries[idx]; }
    inline HciEntry &GetHciEntry(size_t idx) { return hci_entries[idx]; }

    inline size_t AddHciEntry(const HciEntry &entry) { hci_entries.push_back(entry); return (hci_entries.size()-1); }
    inline void RemoveHciEntry(size_t idx) { hci_entries.erase(hci_entries.begin()+idx); }

    inline bool HasHci() const { return (format_version >= X2M_MIN_VERSION_HCI && GetNumHciEntries() > 0); }

    bool CssAudioFileExists(const std::string &name, bool english) const;
    bool CssAudioFileExists(const std::string &name) const;
    HcaFile *LoadCssAudioFile(const std::string &name, bool english, bool fallback);
    size_t GetCssAudioFiles(std::vector<std::string> &files);
    size_t SetCssAudioDir(const std::string &dir_path);
    void DeleteCssAudio();

    // Ram installs
    bool InstallCharaName();
    bool InstallCostumeNames();
    bool InstallSlots(bool update);
    bool InstallSlots(const std::vector<size_t> &positions);
    bool InstallCms(bool update);
    bool InstallCus(bool update);
    bool InstallCso(bool update);
    bool InstallPsc();
    bool InstallAur(bool update);    
    bool InstallSev(std::vector<SevTableEntryHL> &hl_table, std::vector<SevTableEntryLL> &ll_table);
    bool InstallCml(bool update);
    bool InstallHci();
    bool InstallSelPortrait();
    bool InstallPreBaked();

    // File installs
    bool InstallBtlPortrait();
    bool InstallCharaFiles();
    bool InstallJungle();

    // Ram uninstalls
    bool UninstallCharaName();
    bool UninstallCostumeNames();
    bool UninstallSlots();
    bool UninstallCms();
    bool UninstallCus();
    bool UninstallCso();
    bool UninstallPsc();
    bool UninstallAur();
    bool UninstallSev(std::vector<SevTableEntryHL> &hl_table, std::vector<SevTableEntryLL> &ll_table);
    bool UninstallCml();
    bool UninstallHci();
    bool UninstallSelPortrait();
    bool UninstallPreBaked();

    // File installs
    bool UninstallBtlPortrait();
    bool UninstallCharaFiles(bool remove_empty);
    bool UninstallJungle();

};

#endif // X2MFILE_H
